package bar_multi_klausurplaner_cnf;

import static bar_multi_klausurplaner_cnf.Static.*;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.nio.charset.StandardCharsets;
import java.util.LinkedList;
import java.util.TreeMap;

/**
 * @author Benjamin A. Bartsch
 */
public class D_Ausgabe {

    private final int id;

    private final LinkedList<String> html;
    private String html_string;

    private final TreeMap<String, LinkedList<String>> mapXML;
    private final TreeMap<String, String> mapXML_string;

    private final LinkedList<String> csv_schueler;
    private String csv_schueler_string;

    private final LinkedList<String> csv_kurse_1;
    private String csv_kurse_1_string;

    private final LinkedList<String> csv_kurse_2;
    private String csv_kurse_2_string;

    public D_Ausgabe(int id) {
        this.id = id;
        this.html = new LinkedList<>();
        html_string = "";
        this.mapXML = new TreeMap<>();
        this.mapXML_string = new TreeMap<>();
        this.csv_schueler = new LinkedList<>();
        this.csv_schueler_string = "";
        this.csv_kurse_1 = new LinkedList<>();
        this.csv_kurse_1_string = "";
        this.csv_kurse_2 = new LinkedList<>();
        this.csv_kurse_2_string = "";
    }

    @Override
    public String toString() {
        return id + "";
    }

    int get_id() {
        return id;
    }

    void save(File folder) throws FileNotFoundException, UnsupportedEncodingException, IOException {
        // HTML
        {
            File file = new File(folder, "ausgabe.html");
            try (BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), StandardCharsets.UTF_8))) {
                bw.write(html_string);
                bw.close();
            }
        }
        // XML
        for (String stufe : mapXML_string.keySet()) {
            String xml = mapXML_string.get(stufe);
            File file = new File(folder, stufe + ".xml");
            try (BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), StandardCharsets.UTF_8))) {
                bw.write(xml);
                bw.close();
            }
        }
        // CVS - Schüler
        {
            File file = new File(folder, "Statistik_Schueler.csv");
            try (BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), StandardCharsets.UTF_8))) {
                bw.write(csv_schueler_string);
                bw.close();
            }
        }
        // CVS - Kurse 1
        {
            File file = new File(folder, "Statistik_Kurse_nach_Kursnamen.csv");
            try (BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), StandardCharsets.UTF_8))) {
                bw.write(csv_kurse_1_string);
                bw.close();
            }
        }
        // CVS - Kurse 2
        {
            File file = new File(folder, "Statistik_Kurse_nach_Datum.csv");
            try (BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), StandardCharsets.UTF_8))) {
                bw.write(csv_kurse_2_string);
                bw.close();
            }
        }
    }

    String html_get_string() {
        return html_string;
    }

    void html_append(String line) {
        html.addLast(line);
    }

    LinkedList<String> html_get_list() {
        return html;
    }

    void xml_set(String stufe, LinkedList<String> xml) {
        mapXML.put(stufe, xml);
    }

    String xml_get_string(String stufe) {
        return mapXML_string.get(stufe);
    }

    void csv_set_schueler(LinkedList<String> csv) {
        this.csv_schueler.addAll(csv);
    }

    void csv_set_kurse_nach_kursnamen(LinkedList<String> csv) {
        this.csv_kurse_1.addAll(csv);
    }

    void csv_set_kurse_nach_datum(LinkedList<String> csv) {
        this.csv_kurse_2.addAll(csv);
    }

    void save_to_map(D_Map mapA) {
        mapA.get(id).clear();

        // HTML
        mapA.get(id).get(ID_CONTENT).get(html_string);

        // XML
        for (String stufe : mapXML_string.keySet()) {
            String xml = mapXML_string.get(stufe);
            mapA.get(id).get(ID_STUFEN).get(stufe).get(ID_XML).get(xml);
        }

        // CVS - Schüler
        mapA.get(id).get(ID_CVS_SCHUELER).get(csv_schueler_string);

        // CVS - Kurse 1
        mapA.get(id).get(ID_CVS_KURSE_1).get(csv_kurse_1_string);

        // CVS - Kurse 2
        mapA.get(id).get(ID_CVS_KURSE_2).get(csv_kurse_2_string);
    }

    void load_from_map(D_Map mapA) {
        // HTML
        html_string = mapA.get(id).get(ID_CONTENT).first_key_string();

        // XML
        mapXML_string.clear();
        for (String stufe : mapA.get(id).get(ID_STUFEN).keySetSortedString()) {
            String xml = mapA.get(id).get(ID_STUFEN).get(stufe).get(ID_XML).first_key_string();
            mapXML_string.put(stufe, xml);
        }

        // CVS - Schüler
        csv_schueler_string = mapA.get(id).get(ID_CVS_SCHUELER).first_key_string();

        // CVS - Kurse 1
        csv_kurse_1_string = mapA.get(id).get(ID_CVS_KURSE_1).first_key_string();

        // CVS - Kurse 2
        csv_kurse_2_string = mapA.get(id).get(ID_CVS_KURSE_2).first_key_string();
    }

    void abschliessen() {
        // html_string
        LinkedList<String> temp = new LinkedList<>();
        temp.addLast("<!DOCTYPE html>");
        temp.addLast("<html>");
        temp.addLast("<body>");
        temp.addAll(html);
        temp.addLast("</body>");
        temp.addLast("</html>");
        temp = Static.html_escape(temp);

        StringBuilder sb1 = new StringBuilder();
        for (String line : temp) {
            sb1.append(line);
        }
        html_string = sb1.toString();

        // mapXML_string
        for (String stufe : mapXML.keySet()) {
            StringBuilder sb2 = new StringBuilder();
            for (String line : mapXML.get(stufe)) {
                sb2.append(line);
            }
            mapXML_string.put(stufe, sb2.toString());
        }

        // CSV - Schüler
        StringBuilder sb3 = new StringBuilder();
        for (String line : csv_schueler) {
            sb3.append(line);
        }
        csv_schueler_string = sb3.toString();

        // CSV - Kurse 1
        StringBuilder sb4 = new StringBuilder();
        for (String line : csv_kurse_1) {
            sb4.append(line);
        }
        csv_kurse_1_string = sb4.toString();

        // CSV - Kurse 2
        StringBuilder sb5 = new StringBuilder();
        for (String line : csv_kurse_2) {
            sb5.append(line);
        }
        csv_kurse_2_string = sb5.toString();
    }

}
