package bar_multi_klausurplaner_cnf;

import java.util.SortedMap;
import java.util.TreeMap;
import javax.swing.SwingUtilities;
import static bar_multi_klausurplaner_cnf.Static.*;
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.text.ParseException;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.Random;
import java.util.TreeSet;

/**
 * @author Benjamin A. Bartsch
 */
public class D_Berechnung extends Thread {

    private final int id;
    private final Data data;
    private final D_Map map;
    private boolean user_aborted;
    private final D_Main_SAT main;
    private final D_Ausgabe ausgabe;

    // SOLVER RELATED
    Process process;
    private final ByteArrayOutputStream baos = new ByteArrayOutputStream();
    private final Random rnd = new Random();
    private final LinkedList<Integer> listRND = new LinkedList<>();
    private final LinkedList<String> commands = new LinkedList<>();
    private boolean[] state = null;
    private boolean is_UNSAT;

    private boolean c_simulate;
    private int c_vars;
    private int c_clauses;
    private int c_TRUE;
    private int c_FALSE;

    public D_Berechnung(int id, Data data) {
        this.id = id;
        this.data = data;
        this.map = data.co_map_get_copy();
        this.user_aborted = false;
        this.ausgabe = new D_Ausgabe(id);
        this.main = new D_Main_SAT();
        this.is_UNSAT = false;
    }

    private void loesche_alles_ausser_loesung() throws IOException {
        listRND.clear();
        commands.clear();
        state = null;
        map.clear();
        if (baos != null) {
            baos.close();
        }
    }

    @Override
    public String toString() {
        return "" + id;
    }

    int get_id() {
        return id;
    }

    D_Ausgabe get_ergebnis() {
        return ausgabe;
    }

    boolean get_user_aborted() {
        return user_aborted;
    }

    void set_user_abort() {
        user_aborted = true;
        if (process != null) {
            process.destroy();
        }
    }

    @Override
    public void run() {
        data.co_berechnung_start();

        try {

            long t1 = System.currentTimeMillis();
            // Erzeuge Objekte
            main.create();

            // Konfiguriere Objekte
            main.regeln_pre();

            // Simuliere eine Erstellung der Formel
            c_simulate = true;
            c_reset();
            main.cnf_pre();
            main.cnf_post();

            // Starte den Solver
            this.commands.addLast(Static.get_path_crypto());
            this.commands.addLast("--threads=1");
            ProcessBuilder pb = new ProcessBuilder(commands);
            pb.redirectErrorStream(true);
            process = pb.start();

            // Sende dem Solver den Header
            write_bytes("p cnf " + c_vars + " " + c_clauses + "\n", true);
            //System.out.println("HEADER: " + "p cnf " + c_vars + " " + c_clauses + "\n");

            // Sende dem Solver die Klauseln
            c_simulate = false;
            c_reset();
            main.cnf_pre();
            main.cnf_post();

            // Sende dem Solver das Startsignal
            write_bytes(null, true);
            process.getOutputStream().close();

            // Lese Daten vom Solver (abbrechen falls 'user_aborted')
            InputStream out = process.getInputStream();
            BufferedReader br = new BufferedReader(new InputStreamReader(out, "UTF-8"));
            while (process.isAlive() || br.ready()) {
                if (user_aborted) {
                    break;
                }
                String line = br.readLine();
                if (line == null) {
                    break;
                }
                if (line.startsWith("s UNSATISFIABLE")) {
                    is_UNSAT = true;
                }
                if (line.startsWith("v")) {
                    if (state == null) {
                        state = new boolean[c_vars + 1];
                    }
                    String[] split = line.split("\\s+");
                    for (int i = 1; i < split.length; i++) {
                        int k = Integer.parseInt(split[i].trim());
                        if (k > 0) {
                            state[k] = true;
                        }
                    }
                } else {
                    // System.out.println("SOLVER: " + line);
                }
            }
            process.waitFor();

            // Ausgabe des Zeitbedarfs
            ausgabe.html_append("<br>");
            long t2 = System.currentTimeMillis();
            long tSek = ((t2 - t1) / (1000)) % 60;
            long tMin = ((t2 - t1) / (1000 * 60)) % 60;
            long tHour = ((t2 - t1) / (1000 * 60 * 60));
            String sHour = "";
            if (tHour > 0) {
                sHour = "" + tHour + " Std.";
            }
            String sMin = "";
            if (tMin > 0) {
                sMin = "" + tMin + " Min.";
            }
            String sSek = "";
            if (tSek > 0) {
                sSek = "" + tSek + " Sek.";
            }
            ausgabe.html_append("Zeitbedarf = " + sHour + " " + sMin + " " + sSek);
            ausgabe.html_append("<br>");

            // Erzeuge die Ausgabe
            if (state == null) {
                if (is_UNSAT) {
                    ausgabe.html_append("Keine Lösung gefunden. Solver antwortet: UNSATISFIABLE (nicht lösbar)");
                } else {
                    ausgabe.html_append("Keine Lösung gefunden. Solver antwortet: UNKNOWN (Fehler?)");
                }
            } else {
                main.erzeuge_ausgabe();
            }

            ausgabe.abschliessen();

            loesche_alles_ausser_loesung();
        } catch (UserException | IOException | InterruptedException | ParseException | ArrayIndexOutOfBoundsException ex) {
            if (!user_aborted) {
                data.handle_exception(ex);
                if (ex instanceof UserException == false) {
                    ex.printStackTrace();
                }
                user_aborted = true;
            }
        }

        // Informiere GUI
        SwingUtilities.invokeLater(new Runnable() {
            @Override
            public void run() {
                data.co_berechnung_done(id);
            }
        });
    }

    private final class D_Main_SAT {

        final TreeMap<String, D_Jahrgang_SAT> mapJ;
        int wochen;
        int tage;
        int stunden;
        int max_klausuren_pro_schueler_pro_woche;

        String[][] planS;
        boolean[][] planB;
        int[][] wt; // [Woche, Tag]
        String regeln;
        String[] regeln_split;

        public D_Main_SAT() {
            mapJ = new TreeMap<>();
            planS = map.get(ID_PLAN_S).first_key_string_2d();
            planB = map.get(ID_PLAN_B).first_key_boolean_2d();
            regeln = map.get(ID_REGELN).first_key_string();
            regeln_split = regeln.split("[\\r\\n]+");
            wochen = planB.length;
            tage = 7;
            stunden = map.get(ID_MAX_WOCHEN_STUNDEN).first_key_int();
            max_klausuren_pro_schueler_pro_woche = map.get(ID_MAX_KLAUSUREN_PRO_WOCHE).first_key_int();
            String wochenplan_ab = map.get(ID_WOCHENPLAN_DATUM_AB).first_key_string();
            String wochenplan_bi = map.get(ID_WOCHENPLAN_DATUM_BI).first_key_string();

            ausgabe.html_append("<h1>Einstellungen</h1>");
            ausgabe.html_append("<tt>");
            ausgabe.html_append("Wochenplan - Datum (ab)  = " + wochenplan_ab + "<br>");
            ausgabe.html_append("Wochenplan - Datum (bis) = " + wochenplan_bi + "<br>");
            ausgabe.html_append("Max. Klausuren pro SchülerIn pro Woche = " + max_klausuren_pro_schueler_pro_woche + "<br>");
            ausgabe.html_append("Letzte Stunde = " + stunden + "<br>");
            ausgabe.html_append("<br>");
            ausgabe.html_append("" + Static.html_escape_regeln(regeln));
            ausgabe.html_append("</tt>");

        }

        private void create() throws UserException {
            // Erzeuge alles außer Klausuren
            for (String sJahrgang : map.get(ID_STUFEN).keySetSortedString()) {
                D_Jahrgang_SAT ja = new D_Jahrgang_SAT(this, sJahrgang);
                mapJ.put(sJahrgang, ja);
                ja.create(map.get(ID_STUFEN).get(sJahrgang));
            }

            // REGELN - KLAUSURENANZAHL
            for (String line : regeln_split) {
                String[] split2 = line.split("\\#");
                if (split2.length == 2) {
                    try {
                        int nr = Integer.parseInt(split2[0]);
                        regeln_klausurenanzahl(nr, split2[1]);
                    } catch (UserException | NumberFormatException ex) {
                        if (ex instanceof UserException) {
                            throw ex;
                        }
                    }
                }
            }

            // Erzeuge Klausuren
            for (D_Jahrgang_SAT ja : mapJ.values()) {
                ja.create_klausuren();
            }

        }

        private void regeln_pre() throws UserException {
            // REGELN PRE
            for (String line : regeln_split) {
                String[] split2 = line.split("\\#");
                if (split2.length == 2) {
                    try {
                        int nr = Integer.parseInt(split2[0]);
                        regeln_pre(nr, split2[1]);
                    } catch (UserException | NumberFormatException ex) {
                        if (ex instanceof UserException) {
                            throw ex;
                        }
                    }
                }
            }

        }

        private void cnf_pre() throws IOException, UserException {
            wt = get_array_wt();

            // PRE rekursiv
            for (D_Jahrgang_SAT ja : mapJ.values()) {
                ja.cnf_pre();
            }
        }

        private void cnf_post() throws IOException, UserException, ParseException {
            // Verbiete bestimmte Tage
            for (int w = 0; w < wochen; w++) {
                for (int t = 0; t < tage; t++) {
                    if (planB[w][t] == false) {
                        c_1(-wt[w][t]);
                    }
                }
            }

            // OR der Sub-Pläne
            for (int w = 0; w < wochen; w++) {
                for (int t = 0; t < tage; t++) {
                    LinkedList<Integer> listWT = new LinkedList<>();
                    for (D_Jahrgang_SAT ja : mapJ.values()) {
                        listWT.addLast(ja.wt[w][t]);
                    }
                    int varOR = c_OR_tree(listWT);
                    c_equal(varOR, wt[w][t]);
                }
            }

            // POST rekursiv
            for (D_Jahrgang_SAT ja : mapJ.values()) {
                ja.cnf_post();
            }

            // REGELN POST
            for (String line : regeln_split) {
                String[] split2 = line.split("\\#");
                if (split2.length == 2) {
                    try {
                        int nr = Integer.parseInt(split2[0]);
                        regeln_post(nr, split2[1]);
                    } catch (UserException | IOException | NumberFormatException | ParseException ex) {
                        if (ex instanceof UserException) {
                            throw ex;
                        }
                    }
                }
            }

        }

        public int[][] get_array_wt() {
            int[][] temp = new int[wochen][tage];
            for (int w = 0; w < wochen; w++) {
                for (int t = 0; t < tage; t++) {
                    if (planB[w][t] == true) {
                        temp[w][t] = c_new_var();
                    } else {
                        temp[w][t] = c_FALSE;
                    }
                }
            }
            return temp;
        }

        private int[][][] get_array_wts() {
            int[][][] temp = new int[wochen][tage][stunden];
            for (int w = 0; w < wochen; w++) {
                for (int t = 0; t < tage; t++) {
                    for (int s = 0; s < stunden; s++) {
                        if (planB[w][t] == true) {
                            temp[w][t][s] = c_new_var();
                        } else {
                            temp[w][t][s] = c_FALSE;
                        }
                    }
                }
            }
            return temp;
        }

        private void erzeuge_ausgabe() {
            erzeuge_ausgabe_statistik();
            erzeuge_ausgabe_klausuren();
            erzeuge_ausgabe_zellen();
            erzeuge_ausgabe_plan();

            for (D_Jahrgang_SAT ja : mapJ.values()) {
                ja.erzeuge_ausgabe_schueler_matrix();
            }

            for (D_Jahrgang_SAT ja : mapJ.values()) {
                ja.erzeuge_xml();
            }

            erzeuge_csv_schueler();
            erzeuge_csv_kurse_nach_kursnamen();
            erzeuge_csv_kurse_nach_datum();

        }

        private void erzeuge_ausgabe_statistik() {
            ausgabe.html_append("<h1>Statistik</h1>");
            for (D_Jahrgang_SAT ja : mapJ.values()) {
                ausgabe.html_append(ja.kuerzel + "<br>");
                ausgabe.html_append(Static.html_add_space(4, "") + "Klausurtage = " + ja.get_klausurtage() + "<br>");
                ausgabe.html_append(Static.html_add_space(4, "") + "Klausurstunden laut Plan = " + ja.get_klausurstunden_laut_plan() + " von " + ja.get_klausurstunden_insgesamt() + "<br>");
                ausgabe.html_append(Static.html_add_space(4, "") + "Unterrichtsstunden mit fehlenden SuS = " + ja.get_unterrichtsstunden_mit_fehlenden_sus() + "<br>");
                D_Kurs_SAT ku = ja.get_schlimmsten_kurs_mit_fehlenden_sus();
                ausgabe.html_append(Static.html_add_space(6, "") + "... davon am schlimmsten betroffen = " + ku.name + " mit " + ku.get_unterrichtsstunden_mit_fehlenden_sus() + " Stunden + "
                        + ku.get_unterrichtsstunden_mit_fehlenden_sus_im_eigenen_kurs() + " eigene Klausurstunden während der eigenen Unterrichtszeit.<br>");
            }
            ausgabe.html_append("Schule" + "<br>");
            ausgabe.html_append(Static.html_add_space(4, "") + "Klausurtage = " + this.get_klausurtage() + "<br>");
            ausgabe.html_append(Static.html_add_space(4, "") + "Klausurstunden laut Plan = " + this.get_klausurstunden_laut_plan() + " von " + this.get_klausurstunden_insgesamt() + "<br>");
            ausgabe.html_append(Static.html_add_space(4, "") + "Unterrichtsstunden mit fehlenden SuS = " + this.get_unterrichtsstunden_mit_fehlenden_sus() + "<br>");
            D_Kurs_SAT ku = this.get_schlimmsten_kurs_mit_fehlenden_sus();
            ausgabe.html_append(Static.html_add_space(6, "") + "... davon am schlimmsten betroffen = " +ku.ja.kuerzel+": "+ ku.name + " mit " + ku.get_unterrichtsstunden_mit_fehlenden_sus() + " Stunden + " 
                    + ku.get_unterrichtsstunden_mit_fehlenden_sus_im_eigenen_kurs() + " eigene Klausurstunden während der eigenen Unterrichtszeit.<br>");
        }

        private void erzeuge_ausgabe_klausuren() {
            ausgabe.html_append("<h1>Klausuren pro Tag " + mapJ.keySet() + "</h1>");
            int n = mapJ.size() + 1;
            String sStyleTable = "style=\"width:100%; border: 1px solid black; border-collapse: collapse;\"";
            String sStyleSpan1 = "style=\"text-align:center; border: 1px solid black; border-collapse: collapse;\" colspan=\"1\"";
            String sStyleSpanN = "style=\"text-align:center; border: 1px solid black; border-collapse: collapse;\" colspan=\"" + n + "\"";

            String[] header = new String[WOCHENTAGE_VISUELL.length + 1];
            for (int i = 0; i < header.length; i++) {
                if (i < WOCHENTAGE_VISUELL.length) {
                    header[i] = WOCHENTAGE_VISUELL[i];
                } else {
                    header[i] = "Summe";
                }
            }

            // Table...
            ausgabe.html_append("<table " + sStyleTable + ">");

            // Überschrift 1
            ausgabe.html_append("<tr>");
            ausgabe.html_append("<th " + sStyleSpan1 + ">" + "Woche" + "</th>");
            for (int c = 0; c < header.length; c++) {
                ausgabe.html_append("<th " + sStyleSpanN + ">" + header[c] + "</th>");
            }
            ausgabe.html_append("</tr>");

            // Überschrift 2
            ausgabe.html_append("<tr>");
            ausgabe.html_append("<th " + sStyleSpan1 + ">" + "" + "</th>");
            for (int c = 0; c < header.length; c++) {
                for (D_Jahrgang_SAT ja : mapJ.values()) {
                    ausgabe.html_append("<th " + sStyleSpan1 + ">" + ja.kuerzel + "</th>");
                }
                ausgabe.html_append("<th " + sStyleSpan1 + ">" + "&sum;" + "</th>");
            }
            ausgabe.html_append("</tr>");

            // Zeilen
            ausgabe.html_append("</tr>");
            for (int w = 0; w < wochen; w++) {
                ausgabe.html_append("<tr>");
                // Erste Spalte
                String sWoche = "KW " + Static.datum_kalenderwoche(planS[w][0]) + " (" + planS[w][0] + ")";
                ausgabe.html_append("<th " + sStyleSpan1 + ">" + sWoche + "</th>");
                // Mittlere Spalten
                for (int t = 0; t < WOCHENTAGE_VISUELL.length; t++) {
                    if (planB[w][t] == false) {
                        ausgabe.html_append("<td " + sStyleSpanN + ">" + "---" + "</td>");
                    } else {
                        int summe = 0;
                        for (D_Jahrgang_SAT ja : mapJ.values()) {
                            int nKlausuren = ja.get_klausuren_am(w, t);
                            summe += nKlausuren;
                            ausgabe.html_append("<td " + sStyleSpan1 + ">" + convertZero(nKlausuren) + "</td>");
                        }
                        ausgabe.html_append("<td " + sStyleSpan1 + ">" + "<b>" + convertZero(summe) + "</b>" + "</td>");
                    }
                }
                // Letzte Spalte
                int summe = 0;
                for (D_Jahrgang_SAT ja : mapJ.values()) {
                    int nKlausuren = ja.get_klausuren_in_woche(w);
                    summe += nKlausuren;
                    ausgabe.html_append("<td " + sStyleSpan1 + ">" + convertZero(nKlausuren) + "</td>");
                }
                ausgabe.html_append("<td " + sStyleSpan1 + ">" + "<b>" + convertZero(summe) + "</b>" + "</td>");
                // ... Zeilenende
                ausgabe.html_append("</tr>");
            }
            // Letzte Zeile
            ausgabe.html_append("<tr>");
            // Erste Spalte
            ausgabe.html_append("<th " + sStyleSpan1 + ">" + "&sum;" + "</th>");
            // Mittlere Spalten
            for (int t = 0; t < WOCHENTAGE_VISUELL.length; t++) {
                int summe = 0;
                for (D_Jahrgang_SAT ja : mapJ.values()) {
                    summe += ja.get_klausuren_am_tag(t);
                }
                if (summe == 0) {
                    ausgabe.html_append("<td " + sStyleSpanN + ">" + "" + "</td>");
                } else {
                    for (D_Jahrgang_SAT ja : mapJ.values()) {
                        int nKlausuren = ja.get_klausuren_am_tag(t);
                        ausgabe.html_append("<td " + sStyleSpan1 + ">" + convertZero(nKlausuren) + "</td>");
                    }
                    ausgabe.html_append("<td " + sStyleSpan1 + ">" + "<b>" + convertZero(summe) + "</b>" + "</td>");
                }
            }
            // Letzte Spalte
            int summe = 0;
            for (D_Jahrgang_SAT ja : mapJ.values()) {
                int nKlausuren = ja.get_klausuren_gesamt();
                summe += nKlausuren;
                ausgabe.html_append("<td " + sStyleSpan1 + ">" + convertZero(nKlausuren) + "</td>");
            }
            ausgabe.html_append("<td " + sStyleSpan1 + ">" + "<b>" + convertZero(summe) + "</b>" + "</td>");
            // ... Zeilenende
            ausgabe.html_append("</tr>");

            ausgabe.html_append("</table>");
        }

        private void erzeuge_ausgabe_zellen() {
            ausgabe.html_append("<h1>Zellenübersicht " + mapJ.keySet() + "</h1>");
            ausgabe.html_append("Anzahl an Wochen, an denen eine Zelle von mindestens einer Klausur betroffen ist, ohne mündliche Prüfungen." + "<br>");

            int n = mapJ.size() + 1;
            String sStyleTable = "style=\"width:100%; border: 1px solid black; border-collapse: collapse;\"";
            String sStyleSpan1 = "style=\"text-align:center; border: 1px solid black; border-collapse: collapse;\" colspan=\"1\"";
            String sStyleSpanN = "style=\"text-align:center; border: 1px solid black; border-collapse: collapse;\" colspan=\"" + n + "\"";

            String[] header = WOCHENTAGE_VISUELL;

            // Table...
            ausgabe.html_append("<table " + sStyleTable + ">");

            // Überschrift 1
            ausgabe.html_append("<tr>");
            ausgabe.html_append("<th " + sStyleSpan1 + ">" + "Stunde" + "</th>");
            for (int c = 0; c < header.length; c++) {
                ausgabe.html_append("<th " + sStyleSpanN + ">" + header[c] + "</th>");
            }
            ausgabe.html_append("</tr>");

            // Überschrift 2
            ausgabe.html_append("<tr>");
            ausgabe.html_append("<th " + sStyleSpan1 + ">" + "" + "</th>");
            for (int c = 0; c < header.length; c++) {
                for (D_Jahrgang_SAT ja : mapJ.values()) {
                    ausgabe.html_append("<th " + sStyleSpan1 + ">" + ja.kuerzel + "</th>");
                }
                ausgabe.html_append("<th " + sStyleSpan1 + ">" + "&sum;" + "</th>");
            }
            ausgabe.html_append("</tr>");

            // Zeilen
            ausgabe.html_append("</tr>");
            for (int s = 0; s < stunden; s++) {
                ausgabe.html_append("<tr>");
                // Erste Spalte
                ausgabe.html_append("<th " + sStyleSpan1 + ">" + (s + 1) + "</th>");
                // Mittlere Spalten
                for (int t = 0; t < WOCHENTAGE_VISUELL.length; t++) {
                    int summe = 0;
                    for (D_Jahrgang_SAT ja : mapJ.values()) {
                        int nKlausuren = ja.get_summe_betroffen_tag_t_stunde_s(t, s);
                        summe += nKlausuren;
                        ausgabe.html_append("<td " + sStyleSpan1 + ">" + convertZero(nKlausuren) + "</td>");
                    }
                    ausgabe.html_append("<td " + sStyleSpan1 + ">" + "<b>" + convertZero(summe) + "</b>" + "</td>");
                }
                // Letzte Spalte
                /* int summe = 0;
                for (D_Jahrgang_SAT ja : mapJ.values()) {
                    int nKlausuren = ja.get_summe_betroffen_stunde_s(s);
                    summe += nKlausuren;
                    ausgabe.html_append("<td " + sStyleSpan1 + ">" + convertZero(nKlausuren) + "</td>");
                }
                ausgabe.html_append("<td " + sStyleSpan1 + ">" + "<b>" + convertZero(summe) + "</b>" + "</td>");
                 */
                // ... Zeilenende
                ausgabe.html_append("</tr>");
            }
            // Letzte Zeile
            ausgabe.html_append("<tr>");
            // Erste Spalte
            ausgabe.html_append("<th " + sStyleSpan1 + ">" + "&sum;" + "</th>");
            // Mittlere Spalten
            for (int t = 0; t < WOCHENTAGE_VISUELL.length; t++) {
                int summe = 0;
                for (D_Jahrgang_SAT ja : mapJ.values()) {
                    summe += ja.get_summe_betroffen_tag_t(t);
                }
                if (summe == 0) {
                    ausgabe.html_append("<td " + sStyleSpanN + ">" + "" + "</td>");
                } else {
                    for (D_Jahrgang_SAT ja : mapJ.values()) {
                        int nKlausuren = ja.get_summe_betroffen_tag_t(t);
                        ausgabe.html_append("<td " + sStyleSpan1 + ">" + convertZero(nKlausuren) + "</td>");
                    }
                    ausgabe.html_append("<td " + sStyleSpan1 + ">" + "<b>" + convertZero(summe) + "</b>" + "</td>");
                }
            }
            // Letzte Spalte
            /*int summe = 0;
            for (D_Jahrgang_SAT ja : mapJ.values()) {
                int nKlausuren = ja.get_summe_betroffen();
                summe += nKlausuren;
                ausgabe.html_append("<td " + sStyleSpan1 + ">" + convertZero(nKlausuren) + "</td>");
            }
            ausgabe.html_append("<td " + sStyleSpan1 + ">" + "<b>" + convertZero(summe) + "</b>" + "</td>");
             */
            // ... Zeilenende
            ausgabe.html_append("</tr>");

            ausgabe.html_append("</table>");
        }

        private void erzeuge_ausgabe_plan() {
            ausgabe.html_append("<h1>Klausurplan " + mapJ.keySet() + "</h1>");

            int n = mapJ.size() + 1;
            String sStyleTable = "style=\"width:100%; border: 1px solid black; border-collapse: collapse;\"";
            String sStyleSpan1l = "style=\"text-align:left; font-size: 9px; border: 1px solid black; border-collapse: collapse;\" colspan=\"1\"";
            String sStyleSpan1c = "style=\"text-align:center; font-size: 9px; border: 1px solid black; border-collapse: collapse;\" colspan=\"1\"";
            String sStyleSpanN = "style=\"text-align:center; border: 1px solid black; border-collapse: collapse;\" colspan=\"" + n + "\"";

            String[] header = WOCHENTAGE_VISUELL;

            for (int w = 0; w < wochen; w++) {
                String sKW = Static.datum_kalenderwoche(planS[w][0]);
                ausgabe.html_append("<br>Woche " + (w + 1) + " (KW " + sKW + ")<br>");

                // Table...
                ausgabe.html_append("<table " + sStyleTable + ">");

                // Überschrift 1
                ausgabe.html_append("<tr>");
                ausgabe.html_append("<th " + sStyleSpan1c + ">" + "Stunde" + "</th>");
                for (int t = 0; t < header.length; t++) {
                    if (planB[w][t] == true) {
                        ausgabe.html_append("<th " + sStyleSpanN + ">" + header[t] + ", " + planS[w][t] + "</th>");
                    } else {
                        ausgabe.html_append("<th " + sStyleSpanN + ">" + "---" + "</th>");
                    }
                }
                ausgabe.html_append("</tr>");

                // Überschrift 2
                ausgabe.html_append("<tr>");
                ausgabe.html_append("<th " + sStyleSpan1c + ">" + "" + "</th>");
                for (int c = 0; c < header.length; c++) {
                    for (D_Jahrgang_SAT ja : mapJ.values()) {
                        ausgabe.html_append("<th " + sStyleSpan1c + ">" + ja.kuerzel + "</th>");
                    }
                    ausgabe.html_append("<th " + sStyleSpan1c + ">" + "&sum;" + "</th>");
                }
                ausgabe.html_append("</tr>");

                // Zeilen
                ausgabe.html_append("</tr>");
                for (int s = 0; s < stunden; s++) {
                    ausgabe.html_append("<tr>");
                    // Erste Spalte
                    ausgabe.html_append("<th " + sStyleSpan1c + ">" + (s + 1) + "</th>");
                    // Mittlere Spalten
                    for (int t = 0; t < WOCHENTAGE_VISUELL.length; t++) {
                        int summeSG = 0;
                        int summeSK = 0;
                        for (D_Jahrgang_SAT ja : mapJ.values()) {
                            int summeS = 0;
                            int summeK = 0;
                            for (D_Kurs_SAT ku : ja.get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                                if (c_get(ku.wts[w][t][s])) {
                                    summeS += ku.mapS_schriftlich.size();
                                    summeK++;
                                }
                            }
                            summeSG += summeS;
                            summeSK += summeK;
                            // Inhalt: Kurse mit Klausur
                            String sInhalt = "";
                            if (summeS > 0) {
                                sInhalt += summeS + " S.<br>";
                            }
                            for (D_Kurs_SAT ku : ja.get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                                if (c_get(ku.wts[w][t][s])) {
                                    //sInhalt += ku.name + " (" + ku.lehrkraft + ", " + ku.mapS_schriftlich.size() + "/" + ku.mapS_alle.size() + ")" + "<br>";
                                    sInhalt += ku.name + ", " + ku.lehrkraft + "<br>";
                                }
                            }
                            ausgabe.html_append("<td valign=\"top\" " + sStyleSpan1l + ">" + sInhalt + "</td>");
                        }
                        String sInhaltG = "";
                        if (summeSG > 0) {
                            sInhaltG += summeSG + " S.<br>";
                            sInhaltG += summeSK + " K.<br>";
                        }
                        ausgabe.html_append("<td valign=\"top\" " + sStyleSpan1c + ">" + "<b>" + sInhaltG + "</b>" + "</td>");
                    }
                    // ... Zeilenende
                    ausgabe.html_append("</tr>");
                }
                // Letzte Zeile
                ausgabe.html_append("<tr>");

                // ... Zeilenende
                ausgabe.html_append("</tr>");

                ausgabe.html_append("</table>");
            }

        }

        private void erzeuge_csv_schueler() {
            LinkedList<String> csv = new LinkedList<>();
            csv.addLast("Stufe;Name;Vorname;GebDat;Geschlecht;AnzKlGes;AnzWo2Kl;AnzWo3Kl;AnzWo2FoKl;AnzWo3FoKl;Startdatum;Klausurtermine" + "\n");
            for (D_Jahrgang_SAT ja : mapJ.values()) {
                for (D_Schueler_SAT sc : ja.mapS.values()) {
                    // Daten der Spalten
                    String sStufe = ja.kuerzel;
                    String sName = sc.name;
                    String sVorname = sc.vorname;
                    String sGebDat = sc.geburtsdatum;
                    String sGeschlecht = sc.geschlecht;
                    int iAnzKlGes = sc.get_klausuren_gesamt();
                    int iAnzWo2Kl = sc.get_anzahl_an_wochen_mit_2_klausuren();
                    int iAnzWo3Kl = sc.get_anzahl_an_wochen_mit_3_klausuren();
                    int iAnzWo2FoKl = sc.get_anzahl_an_wochen_mit_2_klausuren_in_folge();
                    int iAnzWo3FoKl = sc.get_anzahl_an_wochen_mit_3_klausuren_in_folge();
                    String sStartdatum = planS[0][0];
                    String sKlausurtermine = sc.get_ausgabe_matrix();
                    // Zeile des Schülers
                    csv.addLast("" + sStufe + ";" + sName + ";" + sVorname + ";" + sGebDat + ";" + sGeschlecht + ";" + iAnzKlGes + ";" + iAnzWo2Kl + ";" + iAnzWo3Kl + ";" + iAnzWo2FoKl + ";" + iAnzWo3FoKl + ";" + sStartdatum + ";" + sKlausurtermine + "" + "\n");
                }
            }
            ausgabe.csv_set_schueler(csv);
        }

        private void erzeuge_csv_kurse_nach_kursnamen() {
            LinkedList<String> csv = new LinkedList<>();
            csv.addLast("Stufe;Kursname;Datum;Stunde;AnzKlSchueler;AnzGesSchueler" + "\n");
            for (D_Jahrgang_SAT ja : mapJ.values()) {
                for (D_Kurs_SAT ku : ja.get_alle_kurse(WILDCARD)) {
                    for (int w = 0; w < wochen; w++) {
                        for (int t = 0; t < tage; t++) {
                            if (ja.ist_klausur_tag(w, t) == true) {
                                for (int s = 0; s < stunden; s++) {
                                    if (ku.ist_planmaessig(w, t, s) == true) {
                                        if (ja.ist_klausur_stunde(w, t, s) == true) {
                                            String sStufe = ja.kuerzel;
                                            String sKursname = ku.name;
                                            String sDatum = planS[w][t];
                                            int iStunde = s + 1;
                                            int iAnzKlSchueler = ku.get_klausurschueler_in(w, t, s);
                                            int iAnzGesSchueler = ku.get_gesamtschueler();
                                            // Zeile des Schülers
                                            if (iAnzKlSchueler > 0) {
                                                csv.addLast("" + sStufe + ";" + sKursname + ";" + sDatum + ";" + iStunde + ";" + iAnzKlSchueler + ";" + iAnzGesSchueler + "" + "\n");
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            ausgabe.csv_set_kurse_nach_kursnamen(csv);
        }

        private void erzeuge_csv_kurse_nach_datum() {
            LinkedList<String> csv = new LinkedList<>();
            csv.addLast("Datum;Stunde;Stufe;Kursname;AnzKlSchueler;AnzGesSchueler" + "\n");
            for (int w = 0; w < wochen; w++) {
                for (int t = 0; t < tage; t++) {
                    for (int s = 0; s < stunden; s++) {
                        for (D_Jahrgang_SAT ja : mapJ.values()) {
                            if (ja.ist_klausur_tag(w, t) == true) {
                                for (D_Kurs_SAT ku : ja.get_alle_kurse(WILDCARD)) {
                                    if (ku.ist_planmaessig(w, t, s) == true) {
                                        if (ja.ist_klausur_stunde(w, t, s) == true) {
                                            String sStufe = ja.kuerzel;
                                            String sKursname = ku.name;
                                            String sDatum = planS[w][t];
                                            int iStunde = s + 1;
                                            int iAnzKlSchueler = ku.get_klausurschueler_in(w, t, s);
                                            int iAnzGesSchueler = ku.get_gesamtschueler();
                                            // Zeile des Schülers
                                            if (iAnzKlSchueler > 0) {
                                                csv.addLast("" + sDatum + ";" + iStunde + ";" + sStufe + ";" + sKursname + ";" + iAnzKlSchueler + ";" + iAnzGesSchueler + "" + "\n");
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            ausgabe.csv_set_kurse_nach_datum(csv);
        }

        private int get_klausurtage() {
            int summe = 0;
            for (int w = 0; w < wochen; w++) {
                for (int t = 0; t < tage; t++) {
                    for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                        if (c_get(ku.wt[w][t])) {
                            summe++;
                            break;
                        }
                    }
                }
            }
            return summe;
        }

        private int get_klausurstunden_laut_plan() {
            int summe = 0;
            for (D_Jahrgang_SAT ja : mapJ.values()) {
                summe += ja.get_klausurstunden_laut_plan();
            }
            return summe;
        }

        private int get_klausurstunden_insgesamt() {
            int summe = 0;
            for (D_Jahrgang_SAT ja : mapJ.values()) {
                summe += ja.get_klausurstunden_insgesamt();
            }
            return summe;
        }

        private int get_unterrichtsstunden_mit_fehlenden_sus() {
            int summe = 0;
            for (D_Jahrgang_SAT ja : mapJ.values()) {
                summe += ja.get_unterrichtsstunden_mit_fehlenden_sus();
            }
            return summe;
        }

        private LinkedList<D_Kurs_SAT> get_schriftliche_kurse_mit_kursnamen(String b) {
            LinkedList<D_Kurs_SAT> list = new LinkedList<>();
            for (D_Jahrgang_SAT ja : mapJ.values()) {
                for (D_Kurs_SAT ku : ja.get_schriftliche_kurse_mit_kursnamen(b)) {
                    list.addLast(ku);
                }
            }
            return list;
        }

        private void regeln_klausurenanzahl(int nr, String text) throws UserException {
            String[] t = regeln_check_and_extract(text);
            switch (nr) {
                case 100: // 100#Klausurphasen: Stufe <A> und Kurs <B> schreibt <C> Klausuren.
                    regel_100_klausurphase_stufe_a_kurs_b_schreibt_c_klausuren(t[0], t[1], t[2]);
                    break;
                case 101: // 101#Klausurphasen: Stufe <A> und Fach/Kursart <B>/<C> schreibt <D> Klausuren.
                    regel_101_klausurphase_stufe_a_fach_kursart_b_c_schreibt_d_klausuren(t[0], t[1], t[2], t[3]);
                    break;
            }
        }

        private void regeln_pre(int nr, String text) throws UserException {
            String[] t = regeln_check_and_extract(text);
            switch (nr) {
                case 102: // 102#Klausurphasen: Stufe <A> und Kurs <B> mit Klausurnummer <C> entfällt.
                    regel_102_klausurphase_stufe_a_kurs_b_klausur_c_entfaellt(t[0], t[1], t[2]);
                    break;
                case 103: // 103#Klausurphasen: Stufe <A> und Fach/Kursart <B>/<C> mit Klausurnummer <D> entfällt.
                    regel_103_klausurphase_stufe_a_fach_kursart_b_c_klausur_d_entfaellt(t[0], t[1], t[2], t[3]);
                    break;
                case 104: // 104#Klausurphasen: Stufe <A> und Kurs <B>: Abstand zweier Klausuren ist mindestens <C> Tage.
                    regel_104_klausurphasen_stufe_a_kurs_b_abstand_zweier_klausuren_min_c(t[0], t[1], t[2]);
                    break;
                case 105: // 105#Klausurphasen: Stufe <A> und Fach/Kursart <B>/<C>: Abstand zweier Klausuren ist mindestens <D> Tage.
                    regel_105_klausurphasen_stufe_a_fach_kursart_b_c_abstand_zweier_klausuren_min_d(t[0], t[1], t[2], t[3]);
                    break;
                case 106: // 106#Klausurphasen: Stufe <A> und Schiene <B>: Abstand zweier Klausuren ist mindestens <C> Tage.
                    regel_106_klausurphasen_stufe_a_schiene_b_abstand_zweier_klausuren_min_c(t[0], t[1], t[2]);
                    break;

                case 150: // 150#Klausurlänge: Stufe <A> und Kurs <B> mit Klausurnummer <C> hat Länge <D>.                    
                    regel_150_klausurlaenge_stufe_a_kurs_b_klausurnummer_c_hat_laenge_d(t[0], t[1], t[2], t[3]);
                    break;
                case 151: // 151#Klausurlänge: Stufe <A> und Fach/Kursart <B>/<C> mit Klausurnummer <D> hat Länge <E>.
                    regel_151_klausurlaenge_stufe_a_fach_kursart_b_c_klausurnummer_d_hat_laenge_e(t[0], t[1], t[2], t[3], t[4]);
                    break;
                case 152: // 152#Klausurlänge: Stufe <A> und Schiene <B> mit Klausurnummer <C> hat Länge <D>.                    
                    regel_152_klausurlaenge_stufe_a_schiene_b_klausurnummer_c_hat_laenge_d(t[0], t[1], t[2], t[3]);
                    break;

                case 170: // 170#Klausurtyp: Stufe <A> und Kurs <B> mit Klausurnummer <C> prüft mündlich über <D> Tage.
                    regel_170_klausurtyp_stufe_a_kurs_b_klausurnummer_c_muendlich_mit_d_tagen(t[0], t[1], t[2], t[3]);
                    break;
                case 171: // 171#Klausurtyp: Stufe <A> und Fach/Kursart <B>/<C> mit Klausurnummer <D> prüft mündlich über <E> Tage.
                    regel_171_klausurtyp_stufe_a_kurs_fachart_b_c_klausurnummer_d_muendlich_mit_e_tagen(t[0], t[1], t[2], t[3], t[4]);
                    break;
            }
        }

        private void regeln_post(int nr, String text) throws ParseException, UserException, IOException {
            String[] t = regeln_check_and_extract(text);

            switch (nr) {
                case 200: // // 200#Plan: Stufe <A> und Datum <B> bis Datum <C> gesperrt.
                    regel_200_plan_stufe_a_datum_b_bis_datum_c_gesperrt(t[0], t[1], t[2]);
                    break;
                case 201: // 201#Plan: Stufe <A> und Stunden <B> bis <C> gesperrt.
                    regel_201_plan_stufe_a_stunden_b_bis_c_gesperrt(t[0], t[1], t[2]);
                    break;
                case 202: // 202#Plan: Stufe <A> und Datum <B> und Stunden <C> bis <D> gesperrt.
                    regel_202_plan_stufe_a_datum_b_stunden_c_bis_d_gesperrt(t[0], t[1], t[2], t[3]);
                    break;

                case 300: // 300#SchülerIn: Stufe <A> und SchülerIn <B> in keiner Woche <C> Klausurtage aufeinanderfolgend.
                    regel_300_stufe_a_schueler_b_nicht_c_klausurtage_aufeinanderfolgend(t[0], t[1], t[2]);
                    break;
                case 301: // 301#SchülerIn: Stufe <A> und SchülerIn <B> Wochen-Summe mit <C> Klausurtagen aufeinanderfolgend maximal <D>.
                    regel_301_stufe_a_schueler_b_wochensumme_mit_c_klausurtage_aufeinanderfolgend_max_d(t[0], t[1], t[2], t[3]);
                    break;
                case 302: // 302#SchülerIn: Stufe <A> Schüler-Wochen-Summe mit <B> Klausurtagen aufeinanderfolgend maximal <C>.
                    regel_302_stufe_a_schueler_wochensumme_mit_b_klausurtagen_aufeinanderfolgend_max_c(t[0], t[1], t[2]);
                    break;
                case 303: // 303#SchülerIn: Stufen-Schüler-Wochen-Summe mit <A> Klausurtagen aufeinanderfolgend maximal <B>.
                    regel_303_stufen_schueler_wochensumme_mit_a_klausurtagen_aufeinanderfolgend_max_b(t[0], t[1]);
                    break;
                case 304: // 304#SchülerIn: Stufe <A> und SchülerIn <B> mit Klausur der Fach/Kursart <C>/<D> impliziert danach keinen Klausurtag.
                    regel_304_stufe_a_schueler_b_mit_fach_kursart_c_d_implizieren_danach_keinen_klausurtag(t[0], t[1], t[2], t[3]);
                    break;
                case 305: // 305#SchülerIn: Stufe <A> und SchülerIn <B> mit Klausur der Fach/Kursart <C>/<D> impliziert davor keinen Klausurtag.
                    regel_305_stufe_a_schueler_b_mit_fach_kursart_c_d_implizieren_davor_keinen_klausurtag(t[0], t[1], t[2], t[3]);
                    break;
                case 306: // 306#SchülerIn: Stufe <A> und SchülerIn <B> mit Fach/Kursart <C>/<D> nicht in gleicher Woche mit Fach/Kursart <E>/<F>.
                    regel_306_stufe_a_schueler_b_mit_fach_kursart_c_d_nicht_in_gleicher_woche_mit_fachart_kursart_e_f(t[0], t[1], t[2], t[3], t[4], t[5]);
                    break;
                case 307: // 307#SchülerIn: Stufe <A> und SchülerIn <B> mit Fach/Kursart <C>/<D> in einer Woche begrenzen die Klausurtage auf <E>.                    
                    regel_307_stufe_a_schueler_b_mit_fach_kursart_c_d_begrenzen_klausurtage_auf_e(t[0], t[1], t[2], t[3], t[4]);
                    break;

                case 400: // 400#Klausurbeginn gemeinsam: Stufe <A> alle Kurse (außer LK und mündl. Prüfungen) mit gleichem Fach.
                    regel_400_klausurbeginn_gemeinsam_stufe_a_alle_kurse_mit_gleichem_fach_ausser_lk_und_muendlich(t[0]);
                    break;
                case 401: // 401#Klausurbeginn gemeinsam: Stufe <A> für Kurs <B> und Kurs <C>.
                    regel_401_klausurbeginn_gemeinsam_stufe_a_fuer_kurs_b_und_kurs_c(t[0], t[1], t[2]);
                    break;
                case 402:  // 402#Klausurbeginn gemeinsam: Stufe <A> für Kurs <B> und allen Kursen mit Fach/Kursart <C>/<D>.
                    regel_402_klausurbeginn_gemeinsam_stufe_a_fuer_kurs_b_und_fach_kursart_c_d(t[0], t[1], t[2], t[3]);
                    break;
                case 403: // 403#Klausurbeginn gemeinsam: Stufe <A> für alle Kurse mit Fach/Kursart <B>/<C>.
                    regel_403_klausurbeginn_gemeinsam_stufe_a_fuer_fach_kursart_b_c(t[0], t[1], t[2]);
                    break;
                case 404: // 404#Klausurbeginn gemeinsam: Stufe <A> für alle Kurse mit Fach/Kursart <B>/<C> und alle mit Fach/Kursart <D>/<E>.
                    regel_404_klausurbeginn_gemeinsam_stufe_a_fuer_fach_kursart_b_c_und_fach_kursart_d_e(t[0], t[1], t[2], t[3], t[4]);
                    break;
                case 405: // 405#Klausurbeginn gemeinsam: Stufe <A> für alle Kurse des gleichen Tages.
                    regel_405_klausurbeginn_gemeinsam_stufe_a_alle_kurses_des_gleichen_tages(t[0]);
                    break;
                case 406: //406#Klausurbeginn gemeinsam: Stufe <A> für alle Kurse der gleichen Lehrkraft pro Fach/Kursart.
                    regel_406_klausurbeginn_gemeinsam_stufe_a_alle_kurse_der_gleichen_lehrkraft_pro_fach_kursart(t[0]);
                    break;
                case 407: //407#Klausurbeginn gemeinsam: Stufe <A> für alle Kurse der Schiene <B>.
                    regel_407_klausurbeginn_gemeinsam_stufe_a_alle_kurse_der_schiene_b(t[0], t[1]);
                    break;
                case 408: // 408#Klausurbeginn gemeinsam: Stufe <A> für Kurs <B> und Kurs <C> für Klausurnummer <D> (auch mündl. Prüfungen).
                    regel_408_klausurbeginn_gemeinsam_stufe_a_kurs_b_kurs_c_klausurnummer_d_auch_muendlich(t[0], t[1], t[2], t[3]);
                    break;
                case 409: // 409#Klausurbeginn gemeinsam: Stufe <A> für alle Kurse mit Fach/Kursart <B>/<C> für Klausurnummer <D> (auch mündl. Prüfungen).
                    regel_409_klausurbeginn_gemeinsam_stufe_a_fach_kursart_b_c_klausurnummer_d_auch_muendlich(t[0], t[1], t[2], t[3]);
                    break;

                case 450: // 450#Klausurbeginn: Ist in Stufe <A> für alle Kurse aus Schiene <B> mit Klausurnummer <C> am Datum <D> in Stunde <E>.
                    regel_450_klausurbeginn_ist_in_stufe_a_fuer_alle_kurse_schiene_b_klausurnummer_c_datum_d_stunde_e(t[0], t[1], t[2], t[3], t[4]);
                    break;
                case 451: // 451#Klausurbeginn: Ist in Stufe <A> für Kurs <B> mit Klausurnummer <C>  am Datum <D> in Stunde <E>.
                    regel_451_klausurbeginn_ist_in_stufe_a_fuer_kurs_b_klausurnummer_c_datum_d_stunde_e(t[0], t[1], t[2], t[3], t[4]);
                    break;
                case 452: // 452#Klausurbeginn: Ist in Stufe <A> für Fach/Kursart <B>/<C> mit Klausurnummer <D> am Datum <E> in Stunde <F>.
                    regel_452_klausurbeginn_ist_in_stufe_a_fuer_kurs_fachart_b_c_klausurnummer_d_datum_e_stunde_f(t[0], t[1], t[2], t[3], t[4], t[5]);
                    break;
                case 453: // 453#Klausurbeginn: Ist in Stufe <A> für Kurs <B> mit Klausurnummer <C> zwischen Datum <D> und <E>.
                    regel_453_klausurbeginn_ist_in_stufe_a_fuer_kurs_b_klausurnummer_c_zwischen_datum_d_und_e(t[0], t[1], t[2], t[3], t[4]);
                    break;
                case 454: // 454#Klausurbeginn: Ist in Stufe <A> für Fach/Kursart <B>/<C> mit Klausurnummer <D> zwischen Datum <E> und <F>.
                    regel_454_klausurbeginn_ist_in_stufe_a_fuer_fach_kursart_b_c_klausurnummer_d_zwischen_datum_e_und_f(t[0], t[1], t[2], t[3], t[4], t[5]);
                    break;
                case 455: // 455#Klausurbeginn: Ist in Stufe <A> generell verboten in Stunde <B>.
                    regel_455_stufe_a_klausurbeginn_verbiete_stufe_a_stunde_b(t[0], t[1]);
                    break;
                case 456: // 456#Klausurbeginn: Ist in Stufe <A> generell verboten für Kursart <B> am Wochentag <C> von Stunde <D> bis <E>.
                    regel_456_stufe_a_klausurbeginn_verbiete_stufe_a_kursart_b_wochentag_c_von_stunde_d_bis_e(t[0], t[1], t[2], t[3], t[4]);
                    break;

                case 900: // 900#Optimiere: Stufe <A> hat maximal <B> Klausurtage.
                    regel_900_optimiere_stufe_a_maximal_b_klausurtage(t[0], t[1]);
                    break;
                case 901: // 901#Optimiere: Alle Stufen haben zusammen maximal <A> Klausurtage.
                    regel_901_optimiere_alle_stufen_zusammen_maximal_a_klausurtage(t[0]);
                    break;
                case 902: // 902#Optimiere: Stufe <A> hat mit Stufe <B> höchstens <C> gemeinsame Klausurtage.
                    regel_902_optimiere_stufe_a_hat_mit_stufe_b_hoechstens_c_gemeinsame_klausurtage(t[0], t[1], t[2]);
                    break;
                case 903: // 903#Optimiere: Alle Stufen haben höchstens <A> gemeinsame Klausurtage.
                    regel_903_optimiere_alle_stufen_haben_maximal_a_gemeinsame_klausurtage(t[0]);
                    break;
                case 904: // 904#Optimiere: Stufe <A> hat mit Stufe <B> höchstens <C> gemeinsame Klausurstunden.
                    regel_904_optimiere_stufe_a_hat_mit_stufe_b_hoechstens_c_gemeinsame_klausurstunden(t[0], t[1], t[2]);
                    break;
                case 905: // 905#Optimiere: Alle Stufen haben höchstens <A> gemeinsame Klausurstunden.
                    regel_905_optimiere_alle_stufen_haben_maximal_a_gemeinsame_klausurstunden(t[0]);
                    break;
                case 906: // 906#Optimiere: Alle Stufen zusammen am Datum <A> in Stunde <B> höchstens <C> KlausurschreiberInnen.
                    regel_906_optimiere_alle_stufen_zusammen_am_datum_a_in_stunde_b_hoechstens_c_klausurschreiber(t[0], t[1], t[2]);
                    break;
                case 907: // 907#Optimiere: Stufe hat pro Stundenplanzelle (Summe über alle Wochen) maximal <B> Klausurtermine.
                    regel_907_optimiere_stufe_a_stundenplanzelle_maximal_b_klausurtermine(t[0], t[1]);
                    break;
                case 908: // 908#Optimiere: Stufe <A> hat pro Stundenplanzelle (Summe über alle Wochen) höchstens <B> mal maximal <C> (oder mehr) Klausurtermine.
                    regel_908_optimiere_stufe_a_stundenplanzelle_hoechstens_b_mal_maximal_c_klausurtermine(t[0], t[1], t[2]);
                    break;
                case 909: // 909#Optimiere: Stufe <A> hat pro Stundenplanzelle (Summe über alle Wochen) am Wochentag <B> in Stunde <C> maximal <D> Klausurtermine.
                    regel_909_optimiere_stufe_a_stundenplanzelle_am_tag_b_in_stunde_c_maximal_d_klausurtermine(t[0], t[1], t[2], t[3]);
                    break;

                case 910: // 910#Optimiere: Stufe <A> mit Kurs <B> ist maximal <C> Unterrichtsstunden von Fremdklausuren betroffen.
                    regel_910_optimiere_stufe_a_kurs_b_maximal_c_fremdklausurstunden(t[0], t[1], t[2]);
                    break;
                case 911: // 911#Optimiere: Stufe <A> alle Kurse sind insgesamt maximal <B> Unterrichtsstunden von Fremdklausuren betroffen.
                    regel_911_optimiere_stufe_a_maximal_b_fremdklausurstunden(t[0], t[1]);
                    break;
                case 912: // 912#Optimiere: Stufenübergreifend sind insgesamt maximal <A> Unterrichtsstunden von Fremdklausuren betroffen.
                    regel_912_optimiere_maximal_a_fremdklausurstunden(t[0]);
                    break;

                case 920: // 920#Optimiere: Stufe <A> und Fach/Kursart <B>/<C> hat mindestens <D> Klausurstunden, die nach Plan stattfinden.
                    regel_920_optimiere_stufe_a_fach_kursart_b_c_mindestens_d_klausurstunden_laut_plan(t[0], t[1], t[2], t[3]);
                    break;
                case 921: // 921#Optimiere: Stufe <A> hat in der Summe mindestens <B> Klausurstunden, die laut Plan stattfinden.
                    regel_921_optimiere_stufe_a_summe_mindestens_b_klausurstunden_laut_plan(t[0], t[1]);
                    break;
                case 922: // 922#Optimiere: Insgesamt gibt es mindestens <A> Klausurstunden, die laut Plan stattfinden.
                    regel_922_optimiere_insgesamt_mindestens_a_klausurstunden_laut_plan(t[0]);
                    break;

            }
        }

        private LinkedList<D_Jahrgang_SAT> get_stufen(String a) {
            LinkedList<D_Jahrgang_SAT> list = new LinkedList<>();
            for (D_Jahrgang_SAT ja : mapJ.values()) {
                if (a.equals(WILDCARD) || ja.kuerzel.equals(a)) {
                    list.addLast(ja);
                }
            }
            return list;
        }

        private LinkedList<D_Kurs_SAT> get_kurse() {
            LinkedList<D_Kurs_SAT> list = new LinkedList<>();
            for (D_Jahrgang_SAT ja : get_stufen(WILDCARD)) {
                for (D_Kurs_SAT ku : ja.get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                    list.addLast(ku);
                }
            }
            return list;
        }

        private D_Kurs_SAT get_schlimmsten_kurs_mit_fehlenden_sus() {
            LinkedList<D_Jahrgang_SAT> list = new LinkedList<>(mapJ.values());
            D_Kurs_SAT max = list.get(0).get_schlimmsten_kurs_mit_fehlenden_sus();
            for (D_Jahrgang_SAT ja : list) {
                D_Kurs_SAT next = ja.get_schlimmsten_kurs_mit_fehlenden_sus();
                if (next.get_unterrichtsstunden_mit_fehlenden_sus() > max.get_unterrichtsstunden_mit_fehlenden_sus()) {
                    max = next;
                }
            }
            return max;
        }

        private String[] regeln_check_and_extract(String text) throws UserException {
            String[] result = new String[6];
            for (int i = 0; i < result.length; i++) {
                result[i] = "";
            }

            int index_insert = 0;
            int indexS = 0;
            int stack_size = 0;
            for (int indexE = 0; indexE < text.length(); indexE++) {
                char c = text.charAt(indexE);
                if (c == '<') {
                    if (stack_size == 0) {
                        indexS = indexE;
                        stack_size++;
                    } else {
                        throw new UserException("Symbol '<' kommt unerwartet in Zeile: " + text);
                    }
                }
                if (c == '>') {
                    if (stack_size == 1) {
                        stack_size--;
                        if (index_insert < result.length) {
                            result[index_insert] = text.substring(indexS + 1, indexE);
                            if (result[index_insert].length() == 0) {
                                throw new UserException("Argument ist leer in Zeile: " + text);
                            }
                            index_insert++;
                        } else {
                            throw new UserException("Zu viele Argumente in Zeile: " + text);
                        }
                    } else {
                        throw new UserException("Symbol '>' kommt unerwartet in Zeile: " + text);
                    }
                }
            }
            if (stack_size == 1) {
                throw new UserException("Symbol '>' fehlt in Zeile: " + text);
            }

            return result;
        }

        private void regel_100_klausurphase_stufe_a_kurs_b_schreibt_c_klausuren(String a, String b, String c) {
            int c2 = Integer.parseInt(c);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_100_klausurphase_stufe_a_kurs_b_schreibt_c_klausuren(b, c2);
            }
        }

        private void regel_101_klausurphase_stufe_a_fach_kursart_b_c_schreibt_d_klausuren(String a, String b, String c, String d) {
            int d2 = Integer.parseInt(d);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_101_klausurphase_stufe_a_fach_kursart_b_c_schreibt_d_klausuren(b, c, d2);
            }
        }

        private void regel_102_klausurphase_stufe_a_kurs_b_klausur_c_entfaellt(String a, String b, String c) {
            int c2 = Integer.parseInt(c);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_102_klausurphase_stufe_a_kurs_b_klausur_c_entfaellt(b, c2);
            }
        }

        private void regel_103_klausurphase_stufe_a_fach_kursart_b_c_klausur_d_entfaellt(String a, String b, String c, String d) {
            int d2 = Integer.parseInt(d);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_103_klausurphase_stufe_a_fach_kursart_b_c_klausur_d_entfaellt(b, c, d2);
            }
        }

        private void regel_104_klausurphasen_stufe_a_kurs_b_abstand_zweier_klausuren_min_c(String a, String b, String c) {
            int c2 = Integer.parseInt(c);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_104_klausurphasen_stufe_a_kurs_b_abstand_zweier_klausuren_min_c(b, c2);
            }
        }

        private void regel_105_klausurphasen_stufe_a_fach_kursart_b_c_abstand_zweier_klausuren_min_d(String a, String b, String c, String d) {
            int d2 = Integer.parseInt(d);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_105_klausurphasen_stufe_a_fach_kursart_b_c_abstand_zweier_klausuren_min_d(b, c, d2);
            }
        }

        private void regel_106_klausurphasen_stufe_a_schiene_b_abstand_zweier_klausuren_min_c(String a, String b, String c) {
            int b2 = Integer.parseInt(b) - 1;
            int c2 = Integer.parseInt(c);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_106_klausurphasen_stufe_a_schiene_b_abstand_zweier_klausuren_min_c(b2, c2);
            }
        }

        private void regel_150_klausurlaenge_stufe_a_kurs_b_klausurnummer_c_hat_laenge_d(String a, String b, String c, String d) {
            int d2 = Integer.parseInt(d);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_150_klausurlaenge_stufe_a_kurs_b_klausurnummer_c_hat_laenge_d(b, c, d2);
            }
        }

        private void regel_151_klausurlaenge_stufe_a_fach_kursart_b_c_klausurnummer_d_hat_laenge_e(String a, String b, String c, String d, String e) {
            int e2 = Integer.parseInt(e);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_151_klausurlaenge_stufe_a_fach_kursart_b_c_klausurnummer_d_hat_laenge_e(b, c, d, e2);
            }
        }

        private void regel_152_klausurlaenge_stufe_a_schiene_b_klausurnummer_c_hat_laenge_d(String a, String b, String c, String d) {
            int b2 = Integer.parseInt(b) - 1;
            int d2 = Integer.parseInt(d);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_152_klausurlaenge_stufe_a_schiene_b_klausurnummer_c_hat_laenge_d(b2, c, d2);
            }
        }

        private void regel_170_klausurtyp_stufe_a_kurs_b_klausurnummer_c_muendlich_mit_d_tagen(String a, String b, String c, String d) {
            int c2 = Integer.parseInt(c);
            int d2 = Integer.parseInt(d);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_170_klausurtyp_stufe_a_kurs_b_klausurnummer_c_muendlich_mit_d_tagen(b, c2, d2);
            }
        }

        private void regel_171_klausurtyp_stufe_a_kurs_fachart_b_c_klausurnummer_d_muendlich_mit_e_tagen(String a, String b, String c, String d, String e) {
            int d2 = Integer.parseInt(d);
            int e2 = Integer.parseInt(e);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_171_klausurtyp_stufe_a_kurs_fachart_b_c_klausurnummer_d_muendlich_mit_e_tagen(b, c, d2, e2);
            }
        }

        private void regel_200_plan_stufe_a_datum_b_bis_datum_c_gesperrt(String a, String b, String c) throws ParseException, UserException, IOException {
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_200_plan_stufe_a_datum_b_bis_datum_c_gesperrt(b, c);
            }
        }

        private void regel_201_plan_stufe_a_stunden_b_bis_c_gesperrt(String a, String b, String c) throws IOException, UserException {
            int b2 = Integer.parseInt(b);
            int c2 = Integer.parseInt(c);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_201_plan_stufe_a_stunden_b_bis_c_gesperrt(b2, c2);
            }
        }

        private void regel_202_plan_stufe_a_datum_b_stunden_c_bis_d_gesperrt(String a, String b, String c, String d) throws ParseException, UserException, IOException {
            int c2 = Integer.parseInt(c);
            int d2 = Integer.parseInt(d);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_202_plan_stufe_a_datum_b_stunden_c_bis_d_gesperrt(b, c2, d2);
            }
        }

        private void regel_300_stufe_a_schueler_b_nicht_c_klausurtage_aufeinanderfolgend(String a, String b, String c) throws IOException, UserException {
            int c2 = Integer.parseInt(c);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_300_stufe_a_schueler_b_nicht_c_klausurtage_aufeinanderfolgend(b, c2);
            }
        }

        private void regel_301_stufe_a_schueler_b_wochensumme_mit_c_klausurtage_aufeinanderfolgend_max_d(String a, String b, String c, String d) throws IOException, UserException {
            int c2 = Integer.parseInt(c);
            int d2 = Integer.parseInt(d);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_301_stufe_a_schueler_b_wochensumme_mit_c_klausurtage_aufeinanderfolgend_max_d(b, c2, d2);
            }
        }

        private void regel_302_stufe_a_schueler_wochensumme_mit_b_klausurtagen_aufeinanderfolgend_max_c(String a, String b, String c) throws IOException, UserException {
            int b2 = Integer.parseInt(b);
            int c2 = Integer.parseInt(c);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_302_stufe_a_schueler_wochensumme_mit_b_klausurtagen_aufeinanderfolgend_max_c(b2, c2);
            }
        }

        private void regel_303_stufen_schueler_wochensumme_mit_a_klausurtagen_aufeinanderfolgend_max_b(String a, String b) throws IOException, UserException {
            int a2 = Integer.parseInt(a);
            int b2 = Integer.parseInt(b);

            LinkedList<Integer> list = new LinkedList<>();
            for (D_Jahrgang_SAT ja : get_stufen(WILDCARD)) {
                list.addAll(ja.get_schuelerwochensumme_klausurtage_aufeinanderfolgend(a2));
            }
            c_at_most_GENERIC(list, b2);
        }

        private void regel_304_stufe_a_schueler_b_mit_fach_kursart_c_d_implizieren_danach_keinen_klausurtag(String a, String b, String c, String d) throws IOException, UserException {
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_304_stufe_a_schueler_b_mit_fach_kursart_c_d_implizieren_danach_keinen_klausurtag(b, c, d);
            }
        }

        private void regel_305_stufe_a_schueler_b_mit_fach_kursart_c_d_implizieren_davor_keinen_klausurtag(String a, String b, String c, String d) throws IOException, UserException {
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_305_stufe_a_schueler_b_mit_fach_kursart_c_d_implizieren_davor_keinen_klausurtag(b, c, d);
            }
        }

        private void regel_306_stufe_a_schueler_b_mit_fach_kursart_c_d_nicht_in_gleicher_woche_mit_fachart_kursart_e_f(String a, String b, String c, String d, String e, String f) throws IOException, UserException {
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_306_stufe_a_schueler_b_mit_fach_kursart_c_d_nicht_in_gleicher_woche_mit_fachart_kursart_e_f(b, c, d, e, f);
            }
        }

        private void regel_307_stufe_a_schueler_b_mit_fach_kursart_c_d_begrenzen_klausurtage_auf_e(String a, String b, String c, String d, String e) throws IOException, UserException {
            int e2 = Integer.parseInt(e);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_307_stufe_a_schueler_b_mit_fach_kursart_c_d_begrenzen_klausurtage_auf_e(b, c, d, e2);
            }
        }

        private void regel_400_klausurbeginn_gemeinsam_stufe_a_alle_kurse_mit_gleichem_fach_ausser_lk_und_muendlich(String a) throws IOException, UserException {
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_400_stufe_a_klausurbeginn_gemeinsam_alle_kurse_mit_gleichem_fach_ausser_lk_und_muendlich();
            }
        }

        private void regel_401_klausurbeginn_gemeinsam_stufe_a_fuer_kurs_b_und_kurs_c(String a, String b, String c) throws IOException, UserException {
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_401_klausurbeginn_gemeinsam_stufe_a_fuer_kurs_b_und_kurs_c(b, c);
            }
        }

        private void regel_402_klausurbeginn_gemeinsam_stufe_a_fuer_kurs_b_und_fach_kursart_c_d(String a, String b, String c, String d) throws IOException, UserException {
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_402_klausurbeginn_gemeinsam_stufe_a_fuer_kurs_b_und_fach_kursart_c_d(b, c, d);
            }
        }

        private void regel_403_klausurbeginn_gemeinsam_stufe_a_fuer_fach_kursart_b_c(String a, String b, String c) throws IOException, UserException {
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_403_stufe_a_fuer_fach_kursart_b_c(b, c);
            }
        }

        private void regel_404_klausurbeginn_gemeinsam_stufe_a_fuer_fach_kursart_b_c_und_fach_kursart_d_e(String a, String b, String c, String d, String e) throws IOException, UserException {
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_404_stufe_a_fuer_fach_kursart_b_c_und_fach_kursart_d_e(b, c, d, e);
            }
        }

        private void regel_405_klausurbeginn_gemeinsam_stufe_a_alle_kurses_des_gleichen_tages(String a) throws IOException, UserException {
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_405_klausurbeginn_gemeinsam_stufe_a_alle_kurses_des_gleichen_tages();
            }
        }

        private void regel_406_klausurbeginn_gemeinsam_stufe_a_alle_kurse_der_gleichen_lehrkraft_pro_fach_kursart(String a) throws IOException, UserException {
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_406_klausurbeginn_gemeinsam_stufe_a_alle_kurse_der_gleichen_lehrkraft_pro_fach_kursart();
            }
        }

        private void regel_407_klausurbeginn_gemeinsam_stufe_a_alle_kurse_der_schiene_b(String a, String b) throws IOException, UserException {
            int b2 = Integer.parseInt(b) - 1;
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_407_klausurbeginn_gemeinsam_stufe_a_alle_kurse_der_schiene_b(b2);
            }
        }

        private void regel_408_klausurbeginn_gemeinsam_stufe_a_kurs_b_kurs_c_klausurnummer_d_auch_muendlich(String a, String b, String c, String d) throws IOException, UserException {
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_408_klausurbeginn_gemeinsam_stufe_a_kurs_b_kurs_c_klausurnummer_d_auch_muendlich(b, c, d);
            }
        }

        private void regel_409_klausurbeginn_gemeinsam_stufe_a_fach_kursart_b_c_klausurnummer_d_auch_muendlich(String a, String b, String c, String d) throws IOException, UserException {
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_409_klausurbeginn_gemeinsam_stufe_a_fach_kursart_b_c_klausurnummer_d_auch_muendlich(b, c, d);
            }
        }

        private void regel_450_klausurbeginn_ist_in_stufe_a_fuer_alle_kurse_schiene_b_klausurnummer_c_datum_d_stunde_e(String a, String b, String c, String d, String e) throws IOException, UserException {
            int b2 = Integer.parseInt(b) - 1;
            int c2 = Integer.parseInt(c);
            int e2 = Integer.parseInt(e) - 1;
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_450_stufe_a_klausurbeginn_fixieren_alle_kurse_schiene_b_klausurnummer_c_datum_d_stunde_e(b2, c2, d, e2);
            }
        }

        private void regel_451_klausurbeginn_ist_in_stufe_a_fuer_kurs_b_klausurnummer_c_datum_d_stunde_e(String a, String b, String c, String d, String e) throws IOException, UserException {
            int c2 = Integer.parseInt(c);
            int e2 = Integer.parseInt(e) - 1;
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_451_klausurbeginn_ist_in_stufe_a_fuer_kurs_b_klausurnummer_c_datum_d_stunde_e(b, c2, d, e2);
            }
        }

        private void regel_452_klausurbeginn_ist_in_stufe_a_fuer_kurs_fachart_b_c_klausurnummer_d_datum_e_stunde_f(String a, String b, String c, String d, String e, String f) throws IOException, UserException {
            int d2 = Integer.parseInt(d);
            int f2 = Integer.parseInt(f) - 1;
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_452_klausurbeginn_ist_in_stufe_a_fuer_kurs_fachart_b_c_klausurnummer_d_datum_e_stunde_f(b, c, d2, e, f2);
            }
        }

        private void regel_453_klausurbeginn_ist_in_stufe_a_fuer_kurs_b_klausurnummer_c_zwischen_datum_d_und_e(String a, String b, String c, String d, String e) throws ParseException, UserException, IOException {
            int c2 = Integer.parseInt(c);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_453_klausurbeginn_ist_in_stufe_a_fuer_kurs_b_klausurnummer_c_zwischen_datum_d_und_e(b, c2, d, e);
            }
        }

        private void regel_454_klausurbeginn_ist_in_stufe_a_fuer_fach_kursart_b_c_klausurnummer_d_zwischen_datum_e_und_f(String a, String b, String c, String d, String e, String f) throws ParseException, UserException, IOException {
            int d2 = Integer.parseInt(d);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_454_klausurbeginn_ist_in_stufe_a_fuer_fach_kursart_b_c_klausurnummer_d_zwischen_datum_e_und_f(b, c, d2, e, f);
            }
        }

        private void regel_455_stufe_a_klausurbeginn_verbiete_stufe_a_stunde_b(String a, String b) throws IOException, UserException {
            int b2 = Integer.parseInt(b) - 1;
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_455_stufe_a_klausurbeginn_verbiete_stufe_a_stunde_b(b2);
            }
        }

        private void regel_456_stufe_a_klausurbeginn_verbiete_stufe_a_kursart_b_wochentag_c_von_stunde_d_bis_e(String a, String b, String c, String d, String e) throws IOException, UserException {
            int d2 = Integer.parseInt(d) - 1;
            int e2 = Integer.parseInt(e) - 1;
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_456_stufe_a_klausurbeginn_verbiete_stufe_a_kursart_b_wochentag_c_von_stunde_d_bis_e(b, c, d2, e2);
            }
        }

        private void regel_900_optimiere_stufe_a_maximal_b_klausurtage(String a, String b) throws IOException, UserException {
            int b2 = Integer.parseInt(b);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_900_optimiere_stufe_a_maximal_b_klausurtage(b2);
            }
        }

        private void regel_901_optimiere_alle_stufen_zusammen_maximal_a_klausurtage(String a) throws IOException, UserException {
            int a2 = Integer.parseInt(a);
            c_at_most_GENERIC(toList(wt), a2);
        }

        private void regel_902_optimiere_stufe_a_hat_mit_stufe_b_hoechstens_c_gemeinsame_klausurtage(String a, String b, String c) throws IOException, UserException {
            int c2 = Integer.parseInt(c);
            for (int w = 0; w < wochen; w++) {
                for (int t = 0; t < tage; t++) {
                    LinkedList<Integer> list = new LinkedList<>();
                    for (D_Jahrgang_SAT ja1 : get_stufen(a)) {
                        for (D_Jahrgang_SAT ja2 : get_stufen(b)) {
                            int var = c_new_var_AND(ja1.wt[w][t], ja2.wt[w][t]);
                            list.addLast(var);
                        }
                    }
                    c_at_most_GENERIC(list, c2);
                }
            }
        }

        private void regel_903_optimiere_alle_stufen_haben_maximal_a_gemeinsame_klausurtage(String a) throws IOException, UserException {
            if (mapJ.size() <= 1) {
                return;
            }
            int a2 = Integer.parseInt(a);
            LinkedList<Integer> list = new LinkedList<>();
            for (int w = 0; w < wochen; w++) {
                for (int t = 0; t < tage; t++) {
                    LinkedList<Integer> listSort = new LinkedList();
                    for (D_Jahrgang_SAT ja : get_stufen(WILDCARD)) {
                        listSort.addLast(ja.wt[w][t]);
                    }
                    c_bitonic_sort(listSort);
                    list.addLast(listSort.get(1));
                }
            }
            c_at_most_GENERIC(list, a2);
        }

        private void regel_904_optimiere_stufe_a_hat_mit_stufe_b_hoechstens_c_gemeinsame_klausurstunden(String a, String b, String c) throws IOException, UserException {
            int c2 = Integer.parseInt(c);
            for (int w = 0; w < wochen; w++) {
                for (int t = 0; t < tage; t++) {
                    for (int s = 0; s < stunden; s++) {
                        LinkedList<Integer> list = new LinkedList<>();
                        for (D_Jahrgang_SAT ja1 : get_stufen(a)) {
                            for (D_Jahrgang_SAT ja2 : get_stufen(b)) {
                                int var = c_new_var_AND(ja1.wts_ohneM[w][t][s], ja2.wts_ohneM[w][t][s]);
                                list.addLast(var);
                            }
                        }
                        c_at_most_GENERIC(list, c2);
                    }
                }
            }
        }

        private void regel_905_optimiere_alle_stufen_haben_maximal_a_gemeinsame_klausurstunden(String a) throws IOException, UserException {
            if (mapJ.size() <= 1) {
                return;
            }
            int a2 = Integer.parseInt(a);
            LinkedList<Integer> list = new LinkedList<>();
            for (int w = 0; w < wochen; w++) {
                for (int t = 0; t < tage; t++) {
                    for (int s = 0; s < stunden; s++) {
                        LinkedList<Integer> listSort = new LinkedList();
                        for (D_Jahrgang_SAT ja : get_stufen(WILDCARD)) {
                            listSort.addLast(ja.wts_ohneM[w][t][s]);
                        }
                        c_bitonic_sort(listSort);
                        list.addLast(listSort.get(1));
                    }
                }
            }
            c_at_most_GENERIC(list, a2);
        }

        private void regel_906_optimiere_alle_stufen_zusammen_am_datum_a_in_stunde_b_hoechstens_c_klausurschreiber(String a, String b, String c) throws IOException, UserException {
            int c2 = Integer.parseInt(c);
            for (int w = 0; w < wochen; w++) {
                for (int t = 0; t < tage; t++) {
                    if (a.equals(WILDCARD) || a.equals(planS[w][t])) {
                        for (int s = 0; s < stunden; s++) {
                            if (b.equals(WILDCARD) || (b.equals("" + (s + 1)))) {
                                LinkedList<Integer> listMAX = new LinkedList<>();
                                for (D_Jahrgang_SAT ja : mapJ.values()) {
                                    ja.regel_add_vars_klausurschreiber(listMAX, w, t, s);
                                }
                                c_at_most_GENERIC(listMAX, c2);
                            }
                        }
                    }
                }
            }
        }

        private void regel_907_optimiere_stufe_a_stundenplanzelle_maximal_b_klausurtermine(String a, String b) throws IOException, UserException {
            int b2 = Integer.parseInt(b);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_907_optimiere_stufe_a_stundenplanzelle_maximal_b_klausurtermine(b2);
            }
        }

        private void regel_908_optimiere_stufe_a_stundenplanzelle_hoechstens_b_mal_maximal_c_klausurtermine(String a, String b, String c) throws IOException, UserException {
            int b2 = Integer.parseInt(b);
            int c2 = Integer.parseInt(c);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_908_optimiere_stufe_a_stundenplanzelle_hoechstens_b_mal_maximal_c_klausurtermine(b2, c2);
            }
        }

        private void regel_909_optimiere_stufe_a_stundenplanzelle_am_tag_b_in_stunde_c_maximal_d_klausurtermine(String a, String b, String c, String d) throws IOException, UserException {
            int d2 = Integer.parseInt(d);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_909_optimiere_stufe_a_stundenplanzelle_am_tag_b_in_stunde_c_maximal_d_klausurtermine(a, b, c, d2);
            }
        }

        private void regel_910_optimiere_stufe_a_kurs_b_maximal_c_fremdklausurstunden(String a, String b, String c) throws IOException, UserException {
            int c2 = Integer.parseInt(c);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_910_optimiere_stufe_a_kurs_b_maximal_c_fremdklausurstunden(b, c2);
            }
        }

        private void regel_911_optimiere_stufe_a_maximal_b_fremdklausurstunden(String a, String b) throws IOException, UserException {
            int b2 = Integer.parseInt(b);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_911_optimiere_stufe_a_maximal_b_fremdklausurstunden(b2);
            }
        }

        private void regel_912_optimiere_maximal_a_fremdklausurstunden(String a) throws IOException, UserException {
            int a2 = Integer.parseInt(a);
            LinkedList<Integer> list = new LinkedList<>();
            for (D_Jahrgang_SAT ja : mapJ.values()) {
                list.addAll(ja.regel_get_fremdklausurstunden());
            }
            c_at_most_GENERIC(list, a2);
        }

        private void regel_920_optimiere_stufe_a_fach_kursart_b_c_mindestens_d_klausurstunden_laut_plan(String a, String b, String c, String d) throws IOException, UserException {
            int d2 = Integer.parseInt(d);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_920_optimiere_stufe_a_fach_kursart_b_c_mindestens_d_klausurstunden_laut_plan(b, c, d2);
            }
        }

        private void regel_921_optimiere_stufe_a_summe_mindestens_b_klausurstunden_laut_plan(String a, String b) throws IOException, UserException {
            int b2 = Integer.parseInt(b);
            for (D_Jahrgang_SAT ja : get_stufen(a)) {
                ja.regel_921_optimiere_stufe_a_summe_mindestens_b_klausurstunden_laut_plan(b2);
            }
        }

        private void regel_922_optimiere_insgesamt_mindestens_a_klausurstunden_laut_plan(String a) throws IOException, UserException {
            int a2 = Integer.parseInt(a);
            LinkedList<Integer> list = new LinkedList<>();
            for (D_Jahrgang_SAT ja : mapJ.values()) {
                list.addAll(ja.regel_gib_klausurstunden_nach_plan());
            }
            c_at_least_GENERIC(list, a2);
        }

    }

    private final class D_Jahrgang_SAT {

        final D_Main_SAT ma;
        final String kuerzel;
        final SortedMap<String, D_Kurs_SAT> mapK;
        final SortedMap<String, D_Schueler_SAT> mapS;
        int[][] wt; // [Woche, Tag]
        int[][][] wts_ohneM; // Woche, Tag, Stunde ohne müdliche

        private D_Jahrgang_SAT(D_Main_SAT ma, String kuerzel) {
            this.ma = ma;
            this.kuerzel = kuerzel;
            this.mapK = new TreeMap<>();
            this.mapS = new TreeMap<>();
        }

        private void create(D_Map mapJ) {
            // Erzeuge Kurse
            for (String kursname : mapJ.get(ID_KURSE).keySetSortedString()) {
                D_Map mapKurs = mapJ.get(ID_KURSE).get(kursname);
                D_Kurs_SAT ku = new D_Kurs_SAT(ma, this, kursname, mapKurs);
                mapK.put(kursname, ku);
            }
            // Erzeuge Schüler
            for (String schueler_key : mapJ.get(ID_SCHUELER).keySetSortedString()) {
                D_Map mapSchueler = mapJ.get(ID_SCHUELER).get(schueler_key);
                D_Schueler_SAT sc = new D_Schueler_SAT(ma, this, schueler_key, mapSchueler);
                mapS.put(schueler_key, sc);
            }

            // Verbinde Kurse mit Schülern
            for (String kursname : mapJ.get(ID_KURSE).keySetSortedString()) {
                D_Map mapKurs = mapJ.get(ID_KURSE).get(kursname);
                mapK.get(kursname).verbindeS(mapKurs);
            }
            // Verbinde Schüler mit Kursen
            for (String schueler_key : mapJ.get(ID_SCHUELER).keySetSortedString()) {
                D_Map mapSchueler = mapJ.get(ID_SCHUELER).get(schueler_key);
                mapS.get(schueler_key).verbindeK(mapSchueler);
            }

        }

        private void create_klausuren() {
            for (D_Kurs_SAT ku : mapK.values()) {
                ku.create_klausuren();
            }
        }

        private void cnf_pre() {
            wt = ma.get_array_wt();
            wts_ohneM = ma.get_array_wts();

            for (D_Kurs_SAT ku : mapK.values()) {
                ku.cnf_pre();
            }
            for (D_Schueler_SAT sc : mapS.values()) {
                sc.cnf_pre();
            }
        }

        private void cnf_post() throws IOException, UserException {
            // wt = OR (Kurse.wt)
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    LinkedList<Integer> listWT = new LinkedList<>();
                    for (D_Kurs_SAT ku : mapK.values()) {
                        ku.add_w_t(listWT, w, t);
                    }
                    int varOR = c_OR_tree(listWT);
                    c_equal(varOR, wt[w][t]);
                }
            }

            // wts_ohneM
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    for (int s = 0; s < ma.stunden; s++) {
                        LinkedList<Integer> listWTS_ohne = new LinkedList<>();
                        for (D_Kurs_SAT ku : mapK.values()) {
                            listWTS_ohne.addLast(ku.wts_ohneM[w][t][s]);
                        }
                        int rootWTS_ohneM = c_OR_tree(listWTS_ohne);
                        c_equal(wts_ohneM[w][t][s], rootWTS_ohneM);
                    }
                }
            }

            // POST rekursiv
            for (D_Kurs_SAT ku : mapK.values()) {
                ku.cnf_post();
            }
            for (D_Schueler_SAT sc : mapS.values()) {
                sc.cnf_post();
            }
        }

        private int get_kursanzahl() {
            return mapK.size();
        }

        private int get_kursanzahl_schriftlich() {
            int summe = 0;
            for (D_Kurs_SAT ku : mapK.values()) {
                if (ku.mapS_schriftlich.size() > 0) {
                    summe++;
                }
            }
            return summe;
        }

        private int get_klausuren_am(int w, int t) {
            int summe = 0;
            for (D_Kurs_SAT ku : mapK.values()) {
                if (c_get(ku.wt[w][t])) {
                    summe++;
                }
            }
            return summe;
        }

        private int get_klausuren_in_woche(int w) {
            int summe = 0;
            for (int t = 0; t < ma.tage; t++) {
                summe += get_klausuren_am(w, t);
            }
            return summe;
        }

        private int get_klausuren_am_tag(int t) {
            int summe = 0;
            for (int w = 0; w < ma.wochen; w++) {
                summe += get_klausuren_am(w, t);
            }
            return summe;
        }

        private int get_klausuren_gesamt() {
            int summe = 0;
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    summe += get_klausuren_am(w, t);
                }
            }
            return summe;
        }

        private void erzeuge_ausgabe_klausur_matrix() {
            ausgabe.html_append("<h1>Klausurmatrix [" + kuerzel + "]</h1>");
            ausgabe.html_append("<tt>");
            int klausurtage = 0;
            for (int w = 0; w < ma.planS.length; w++) {
                String line = "";
                for (int t = 0; t < ma.planS[w].length; t++) {
                    if (c_get(wt[w][t])) {
                        klausurtage++;
                    }
                    if (ma.planB[w][t]) {
                        int summe = 0;
                        for (D_Kurs_SAT ku : mapK.values()) {
                            if (c_get(ku.wt[w][t])) {
                                summe++;
                            }
                        }
                        line += Static.html_add_space(4, summe + "");
                    } else {
                        line += Static.html_add_space(4, "x");
                    }
                }
                line += "&nbsp;&nbsp;&nbsp;&nbsp;[" + ma.planS[w][0] + " bis " + ma.planS[w][6] + "]";
                ausgabe.html_append(line + "<br>");
            }
            ausgabe.html_append("</tt>");
            ausgabe.html_append("Klausurtage = " + klausurtage + "<br>");
        }

        private void erzeuge_ausgabe_zellen() {
            ausgabe.html_append("<h1>Zellenübersicht [" + kuerzel + "]</h1>");
            ausgabe.html_append("(Anzahl der Tage, an denen die Zelle betroffen ist)" + "<br>");
            ausgabe.html_append("(ohne mündliche Prüfungen)" + "<br>");
            String[][][] planBZ = new String[ma.stunden][ma.tage][2];
            for (int s = 0; s < ma.stunden; s++) {
                for (int t = 0; t < ma.tage; t++) {
                    int summeZelle = 0;
                    for (int w = 0; w < ma.wochen; w++) {
                        int summeWoche = 0;
                        for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                            if (c_get(ku.wts_ohneM[w][t][s])) {
                                summeWoche++;
                            }
                        }
                        if (summeWoche > 0) {
                            summeZelle++;
                        }
                    }
                    planBZ[s][t][0] = "" + summeZelle;
                    planBZ[s][t][1] = "" + summeZelle;
                }
            }
            Static.toHtmlTable(planBZ, ausgabe.html_get_list(), new String[]{"Stunde", "Mo", "Di", "Mi", "Do", "Fr", "Sa", "So"});
        }

        private void erzeuge_ausgabe_plan() {
            // Ausgabe der Klausuren pro Woche / Tag /Stunde / 
            ausgabe.html_append("<h1>Klausurplan [" + kuerzel + "]</h1>");
            for (int w = 0; w < ma.wochen; w++) {
                ausgabe.html_append("<br>");
                String sKW = Static.datum_kalenderwoche(ma.planS[w][0]);
                ausgabe.html_append("Woche " + (w + 1) + " (KW " + sKW + ") [" + kuerzel + "]<br>");

                String[][][] planK = new String[ma.stunden][ma.tage][2];
                for (int s = 0; s < ma.stunden; s++) {
                    for (int t = 0; t < ma.tage; t++) {
                        if (ma.planB[w][t] == false) {
                            planK[s][t][0] = "---";
                            planK[s][t][1] = "---";
                            continue;
                        }
                        String sInhalt = "";
                        for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                            if (c_get(ku.wts[w][t][s])) {
                                sInhalt += ku.name + " (" + ku.lehrkraft + ", " + ku.mapS_schriftlich.size() + "/" + ku.mapS_alle.size() + ")" + "<br>";
                            }
                        }
                        planK[s][t][0] = sInhalt;
                        planK[s][t][1] = sInhalt;
                    }
                }

                String[] header = new String[]{"Stunde", "Mo", "Di", "Mi", "Do", "Fr", "Sa", "So"};
                for (int t = 0; t < ma.tage; t++) {
                    header[t + 1] = ma.planS[w][t];
                    if (ma.planB[w][t] == false) {
                        header[t + 1] += "x";
                    }
                }
                Static.toHtmlTable(planK, ausgabe.html_get_list(), header);
            }
        }

        private void erzeuge_ausgabe_schueler_matrix() {
            Object[][] data = new Object[mapS.size()][4];
            TreeSet<Object[]> set = new TreeSet<>(new Comparator<Object[]>() {
                @Override
                public int compare(Object[] a, Object[] b) {
                    int a3 = (int) a[2];
                    int b3 = (int) b[2];
                    if (a3 > b3) {
                        return -1;
                    }
                    if (a3 < b3) {
                        return +1;
                    }
                    int a2 = (int) a[3];
                    int b2 = (int) b[3];
                    if (a2 > b2) {
                        return -1;
                    }
                    if (a2 < b2) {
                        return +1;
                    }
                    String a1 = (String) a[1];
                    String b1 = (String) b[1];
                    return a1.compareTo(b1);
                }
            });

            int index = 0;
            int sum2 = 0;
            int sum3 = 0;
            for (D_Schueler_SAT sc : mapS.values()) {
                data[index][0] = sc.get_ausgabe_matrix();
                data[index][1] = sc.key;
                data[index][2] = sc.get_ausgabe_3er();
                sum3 += sc.get_ausgabe_3er();
                data[index][3] = sc.get_ausgabe_2er();
                sum2 += sc.get_ausgabe_2er();
                set.add(data[index]);
                index++;
            }
            // Unsortierte Schülerliste
            ausgabe.html_append("<h1>Schülermatrix [" + kuerzel + "]</h1>");
            ausgabe.html_append("<tt>");
            for (int i = 0; i < data.length; i++) {
                ausgabe.html_append("" + data[i][0] + Static.html_add_space(4, data[i][2] + "") + html_add_space(4, data[i][3] + "") + "&nbsp;&nbsp;&nbsp;&nbsp;" + data[i][1] + "<br>");
            }
            ausgabe.html_append("</tt>");
            // Sortierte Schülerliste
            ausgabe.html_append("<h1>Schülermatrix [" + kuerzel + "] (SUM-3er=" + sum3 + ", SUM-2er=" + sum2 + ")</h1>");
            ausgabe.html_append("<tt>");
            for (Object[] o : set) {
                ausgabe.html_append("" + o[0] + Static.html_add_space(4, o[2] + "") + html_add_space(4, o[3] + "") + "&nbsp;&nbsp;&nbsp;&nbsp;" + o[1] + "<br>");
            }
            ausgabe.html_append("</tt>");

        }

        private void erzeuge_ausgabe_kurse() {
            ausgabe.html_append("<h1>Kursmatrix [" + kuerzel + "]</h1>");
            ausgabe.html_append("<tt>");
            for (D_Kurs_SAT ku : mapK.values()) {
                ku.erzeuge_ausgabe_matrix();
            }
            ausgabe.html_append("</tt>");
        }

        private void erzeuge_xml() {
            // Fülle mit Daten ...
            D_Map mapX = new D_Map();
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                for (D_Klausur_SAT kl : ku.mapKL.values()) {
                    D_Map quartal = mapX.get(kl.klausurnummer); // Sortiere nach Phasen
                    for (int w = 0; w < ma.wochen; w++) {
                        for (int t = 0; t < ma.tage; t++) {
                            if (c_get(kl.wt[w][t])) {
                                String datum = ma.planS[w][t];
                                D_Map mapSchieneDatum = quartal.get(datum);
                                for (int s = 0; s < ma.stunden; s++) {
                                    if (c_get(kl.startet_in_der_stunde[s])) {
                                        D_Map mapSchieneVon = mapSchieneDatum.get(s + 1); // Index-Verschiebung
                                        D_Map mapSchieneBis = mapSchieneVon.get(s + 1 + kl.klausurlaenge - 1); // Index-Verschiebung
                                        mapSchieneBis.get(ku);
                                    }
                                }
                            }
                        }
                    }
                }
            }

            // Eingabedaten lesen
            String sBlockung = map.get(ID_STUFEN).get(kuerzel).get(ID_BLOCKUNG).first_key_string();
            String sBlockgruppe = map.get(ID_STUFEN).get(kuerzel).get(ID_BLOCKGRUPPE).first_key_string();
            String sJahr = map.get(ID_STUFEN).get(kuerzel).get(ID_JAHR).first_key_string();
            String sAbschnitt = map.get(ID_STUFEN).get(kuerzel).get(ID_ABSCHNITT).first_key_string();
            int iStundenDauer = map.get(ID_STUNDENDAUER).first_key_int();

            // Ausgabe konstruieren
            //System.out.println("XML Ausgabe");
            LinkedList<String> xml = new LinkedList<>();
            xml.addLast("<KLAUSURTERMINE Blockung=\"" + sBlockung + "\" Blockgruppe=\"" + sBlockgruppe + "\" Jahr=\"" + sJahr + "\" Abschnitt=\"" + sAbschnitt + "\">");
            for (int sQuartal : mapX.keySetSortedInteger()) {
                xml.addLast("<KLAUSURTERMIN Bez=\"" + sQuartal + ". Quartal\">");
                int schienenCount = 0;
                //System.out.println("mapX.get(sQuartal) = " + mapX.get(sQuartal).keySet());

                for (String sDatum : mapX.get(sQuartal).keySetSortedDate()) {
                    for (int iVon : mapX.get(sQuartal).get(sDatum).keySetSortedInteger()) {
                        for (int iBis : mapX.get(sQuartal).get(sDatum).get(iVon).keySetSortedInteger()) {
                            int iKlausurlaenge = iBis - iVon + 1;
                            int iDauer = iKlausurlaenge * iStundenDauer;
                            schienenCount++;
                            String sBeginn = map.get(ID_STUNDE + "_" + iVon).first_key_string();
                            String sSchiene = "Schiene " + schienenCount;
                            //if (mapSchieneZuBeschriftung.containsKey(schienenCount)) {
                            //    sSchiene = mapSchieneZuBeschriftung.get(schienenCount);
                            //}
                            xml.addLast("<KLAUSURSCHIENE Bez=\"" + sSchiene + "\" Datum=\"" + sDatum + "\" VonStd=\"" + iVon + "\" BisStd=\"" + iBis + "\" Beginn=\"" + sBeginn + "\" Dauer=\"" + iDauer + "\">");
                            for (Object oKurs : mapX.get(sQuartal).get(sDatum).get(iVon).get(iBis).keySet()) {
                                D_Kurs_SAT ku = (D_Kurs_SAT) oKurs;
                                xml.addLast("<KURS Bez=\"" + ku.name + "\" Fach=\"" + ku.fach + "\" Kursart=\"" + ku.kursart + "\" Lehrer=\"" + ku.lehrkraft + "\" Klausurraum=\"--\" Schuelerzahl=\"" + ku.mapS_alle.size() + "\" Klausurschreiber=\"" + ku.mapS_schriftlich.size() + "\">");
                                for (int std = 0; std < iKlausurlaenge; std++) {
                                    xml.addLast("<STUNDE Stunde=\"" + std + "\" Aufsicht=\"" + "--" + "\"/>");
                                }
                                for (D_Schueler_SAT sc : ku.mapS_alle.values()) {
                                    String sKS = ku.mapS_schriftlich.containsKey(sc.key) ? "j" : "n";
                                    xml.addLast("<SCHUELER DbIdnr=\"" + sc.dbid + "\" Name=\"" + sc.name + "\" Vorname=\"" + sc.vorname + "\" Klausurschreiber=\"" + sKS + "\"/>");
                                }
                                xml.addLast("</KURS>");
                            }
                            xml.addLast("</KLAUSURSCHIENE>");
                        }
                    }
                }
                xml.addLast("</KLAUSURTERMIN>");
            }
            xml.addLast("</KLAUSURTERMINE>");

            ausgabe.xml_set(kuerzel, xml);
        }

        private LinkedList<D_Kurs_SAT> get_schriftliche_kurse_aus_schiene_0_indiziert(int schiene) {
            LinkedList<D_Kurs_SAT> list = new LinkedList<>();
            for (D_Kurs_SAT ku : mapK.values()) {
                if (ku.ist_schriftlich()) {
                    if (ku.schienen.contains(schiene)) {
                        list.addLast(ku);
                    }
                }
            }
            return list;
        }

        private LinkedList<D_Kurs_SAT> get_schriftliche_kurse_mit_fach_kursart(String b, String c) {
            LinkedList<D_Kurs_SAT> list = new LinkedList<>();
            for (D_Kurs_SAT ku : mapK.values()) {
                if (b.equals(WILDCARD) || ku.fach.equals(b)) {
                    if (c.equals(WILDCARD) || ku.kursart.equals(c)) {
                        if (ku.ist_schriftlich()) {
                            list.addLast(ku);
                        }
                    }
                }
            }
            return list;
        }

        private LinkedList<D_Kurs_SAT> get_schriftliche_kurse_mit_kursnamen(String b) {
            LinkedList<D_Kurs_SAT> list = new LinkedList<>();
            for (D_Kurs_SAT ku : mapK.values()) {
                if (ku.ist_schriftlich()) {
                    if (b.equals(WILDCARD) || ku.name.equals(b)) {
                        list.addLast(ku);
                    }
                }
            }
            return list;
        }

        private LinkedList<D_Kurs_SAT> get_alle_kurse(String b) {
            LinkedList<D_Kurs_SAT> list = new LinkedList<>();
            for (D_Kurs_SAT ku : mapK.values()) {
                if (b.equals(WILDCARD) || ku.name.equals(b)) {
                    list.addLast(ku);
                }
            }
            return list;
        }

        private LinkedList<D_Kurs_SAT> get_schriftliche_kurse_ausser_lk() {
            LinkedList<D_Kurs_SAT> list = new LinkedList<>();
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                if (ku.ist_nicht_lk()) {
                    list.addLast(ku);
                }
            }
            return list;
        }

        private LinkedList<D_Schueler_SAT> gib_schueler(String b) {
            LinkedList<D_Schueler_SAT> list = new LinkedList<>();
            for (D_Schueler_SAT sc : mapS.values()) {
                if (b.equals(WILDCARD) || sc.key.equals(b)) {
                    list.addLast(sc);
                }
            }
            return list;
        }

        private int get_summe_betroffen_tag_t_stunde_s(int t, int s) {
            int summeZelle = 0;
            for (int w = 0; w < ma.wochen; w++) {
                int summeWoche = 0;
                for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                    if (c_get(ku.wts_ohneM[w][t][s])) {
                        summeWoche++;
                    }
                }
                if (summeWoche > 0) {
                    summeZelle++;
                }
            }
            return summeZelle;
        }

        private int get_summe_betroffen_stunde_s(int s) {
            int summeZelle = 0;
            for (int t = 0; t < ma.tage; t++) {
                summeZelle += get_summe_betroffen_tag_t_stunde_s(t, s);
            }
            return summeZelle;
        }

        private int get_summe_betroffen_tag_t(int t) {
            int summeZelle = 0;
            for (int s = 0; s < ma.stunden; s++) {
                summeZelle += get_summe_betroffen_tag_t_stunde_s(t, s);
            }
            return summeZelle;
        }

        private int get_summe_betroffen() {
            int summeZelle = 0;
            for (int s = 0; s < ma.stunden; s++) {
                summeZelle += get_summe_betroffen_stunde_s(s);
            }
            return summeZelle;
        }

        private int get_klausurtage() {
            int summe = 0;
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                        if (c_get(ku.wt[w][t])) {
                            summe++;
                            break;
                        }
                    }
                }
            }
            return summe;
        }

        private int get_klausurstunden_laut_plan() {
            int summe = 0;
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                summe += ku.get_klausurstunden_laut_plan();
            }
            return summe;
        }

        private int get_klausurstunden_insgesamt() {
            int summe = 0;
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                summe += ku.get_klausurstunden_insgesamt();
            }
            return summe;
        }

        private int get_unterrichtsstunden_mit_fehlenden_sus() {
            int summe = 0;
            for (D_Kurs_SAT ku : get_alle_kurse(WILDCARD)) {
                summe += ku.get_unterrichtsstunden_mit_fehlenden_sus();
            }
            return summe;
        }

        private D_Kurs_SAT get_schlimmsten_kurs_mit_fehlenden_sus() {
            LinkedList<D_Kurs_SAT> list = get_alle_kurse(WILDCARD);
            D_Kurs_SAT max = list.getFirst();
            for (D_Kurs_SAT ku : list) {
                if (ku.get_unterrichtsstunden_mit_fehlenden_sus() > max.get_unterrichtsstunden_mit_fehlenden_sus()) {
                    max = ku;
                }
            }
            return max;
        }

        private void regel_100_klausurphase_stufe_a_kurs_b_schreibt_c_klausuren(String b, int c) {
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(b)) {
                ku.regel_setze_klausurenanzahl_auf(c);
            }
        }

        private void regel_101_klausurphase_stufe_a_fach_kursart_b_c_schreibt_d_klausuren(String b, String c, int d) {
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_fach_kursart(b, c)) {
                ku.regel_setze_klausurenanzahl_auf(d);
            }
        }

        private void regel_102_klausurphase_stufe_a_kurs_b_klausur_c_entfaellt(String b, int c) {
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(b)) {
                ku.regel_setze_klausur_nr_entfaellt(c);
            }
        }

        private void regel_103_klausurphase_stufe_a_fach_kursart_b_c_klausur_d_entfaellt(String b, String c, int d) {
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_fach_kursart(b, c)) {
                ku.regel_setze_klausur_nr_entfaellt(d);
            }
        }

        private void regel_104_klausurphasen_stufe_a_kurs_b_abstand_zweier_klausuren_min_c(String b, int c) {
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(b)) {
                ku.regel_setze_klausurenabstand_auf(c);
            }
        }

        private void regel_105_klausurphasen_stufe_a_fach_kursart_b_c_abstand_zweier_klausuren_min_d(String b, String c, int d) {
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_fach_kursart(b, c)) {
                ku.regel_setze_klausurenabstand_auf(d);
            }
        }

        private void regel_106_klausurphasen_stufe_a_schiene_b_abstand_zweier_klausuren_min_c(int b, int c) {
            for (D_Kurs_SAT ku : get_schriftliche_kurse_aus_schiene_0_indiziert(b)) {
                ku.regel_setze_klausurenabstand_auf(c);
            }
        }

        private void regel_150_klausurlaenge_stufe_a_kurs_b_klausurnummer_c_hat_laenge_d(String b, String c, int d) {
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(b)) {
                ku.regel_setze_klausurnummer_c_auf_laenge_d(c, d);
            }
        }

        private void regel_151_klausurlaenge_stufe_a_fach_kursart_b_c_klausurnummer_d_hat_laenge_e(String b, String c, String d, int e) {
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_fach_kursart(b, c)) {
                ku.regel_setze_klausurnummer_c_auf_laenge_d(d, e);
            }
        }

        private void regel_152_klausurlaenge_stufe_a_schiene_b_klausurnummer_c_hat_laenge_d(int b, String c, int d) {
            for (D_Kurs_SAT ku : get_schriftliche_kurse_aus_schiene_0_indiziert(b)) {
                ku.regel_setze_klausurnummer_c_auf_laenge_d(c, d);
            }
        }

        private void regel_170_klausurtyp_stufe_a_kurs_b_klausurnummer_c_muendlich_mit_d_tagen(String b, int c, int d) {
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(b)) {
                ku.regel_setze_klausurnummer_c_muendlich_mit_d_tagen(c, d);
            }
        }

        private void regel_171_klausurtyp_stufe_a_kurs_fachart_b_c_klausurnummer_d_muendlich_mit_e_tagen(String b, String c, int d, int e) {
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_fach_kursart(b, c)) {
                ku.regel_setze_klausurnummer_c_muendlich_mit_d_tagen(d, e);
            }
        }

        private void regel_200_plan_stufe_a_datum_b_bis_datum_c_gesperrt(String b, String c) throws ParseException, UserException, IOException {
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    String datum = ma.planS[w][t];
                    if (b.equals(WILDCARD) || (datum_compare(b, datum) <= 0)) {
                        if (c.equals(WILDCARD) || (datum_compare(datum, c) <= 0)) {
                            c_1(-wt[w][t]);
                        }
                    }
                }
            }
        }

        private void regel_201_plan_stufe_a_stunden_b_bis_c_gesperrt(int b, int c) throws IOException, UserException {
            int b2 = Math.min(b, c) - 1;
            int c2 = Math.max(b, c) - 1;
            for (D_Kurs_SAT ku : mapK.values()) {
                ku.regel_201_plan_stufe_a_stunden_b_bis_c_gesperrt(b2, c2);
            }
        }

        private void regel_202_plan_stufe_a_datum_b_stunden_c_bis_d_gesperrt(String b, int c, int d) throws ParseException, UserException, IOException {
            int c2 = Math.min(c, d) - 1;
            int d2 = Math.max(c, d) - 1;
            for (D_Kurs_SAT ku : mapK.values()) {
                ku.regel_202_plan_stufe_a_datum_b_stunden_c_bis_d_gesperrt(b, c2, d2);
            }
        }

        private void regel_300_stufe_a_schueler_b_nicht_c_klausurtage_aufeinanderfolgend(String b, int c) throws IOException, UserException {
            for (D_Schueler_SAT sc : gib_schueler(b)) {
                sc.regel_300_stufe_a_schueler_b_nicht_c_klausurtage_aufeinanderfolgend(c);
            }
        }

        private void regel_301_stufe_a_schueler_b_wochensumme_mit_c_klausurtage_aufeinanderfolgend_max_d(String b, int c, int d) throws IOException, UserException {
            for (D_Schueler_SAT sc : gib_schueler(b)) {
                sc.regel_301_stufe_a_schueler_b_wochensumme_mit_c_klausurtage_aufeinanderfolgend_max_d(c, d);
            }
        }

        private void regel_302_stufe_a_schueler_wochensumme_mit_b_klausurtagen_aufeinanderfolgend_max_c(int b, int c) throws IOException, UserException {
            c_at_most_GENERIC(get_schuelerwochensumme_klausurtage_aufeinanderfolgend(b), c);
        }

        private LinkedList<Integer> get_schuelerwochensumme_klausurtage_aufeinanderfolgend(int b) throws IOException, UserException {
            LinkedList<Integer> list = new LinkedList<>();
            for (D_Schueler_SAT sc : gib_schueler(WILDCARD)) {
                list.addAll(sc.regel_get_wochensumme_klausurtage_aufeinanderfolgend(b));
            }
            return list;
        }

        private void regel_304_stufe_a_schueler_b_mit_fach_kursart_c_d_implizieren_danach_keinen_klausurtag(String b, String c, String d) throws IOException, UserException {
            for (D_Schueler_SAT sc : gib_schueler(b)) {
                sc.regel_304_stufe_a_schueler_b_mit_fach_kursart_c_d_implizieren_danach_keinen_klausurtag(c, d);
            }
        }

        private void regel_305_stufe_a_schueler_b_mit_fach_kursart_c_d_implizieren_davor_keinen_klausurtag(String b, String c, String d) throws IOException, UserException {
            for (D_Schueler_SAT sc : gib_schueler(b)) {
                sc.regel_305_stufe_a_schueler_b_mit_fach_kursart_c_d_implizieren_davor_keinen_klausurtag(c, d);
            }
        }

        private void regel_306_stufe_a_schueler_b_mit_fach_kursart_c_d_nicht_in_gleicher_woche_mit_fachart_kursart_e_f(String b, String c, String d, String e, String f) throws IOException, UserException {
            for (D_Schueler_SAT sc : gib_schueler(b)) {
                sc.regel_306_stufe_a_schueler_b_mit_fach_kursart_c_d_nicht_in_gleicher_woche_mit_fachart_kursart_e_f(c, d, e, f);
            }
        }

        private void regel_307_stufe_a_schueler_b_mit_fach_kursart_c_d_begrenzen_klausurtage_auf_e(String b, String c, String d, int e) throws IOException, UserException {
            for (D_Schueler_SAT sc : gib_schueler(b)) {
                sc.regel_307_stufe_a_schueler_b_mit_fach_kursart_c_d_begrenzen_klausurtage_auf_e(c, d, e);
            }
        }

        private void regel_400_stufe_a_klausurbeginn_gemeinsam_alle_kurse_mit_gleichem_fach_ausser_lk_und_muendlich() throws IOException, UserException {
            for (D_Kurs_SAT ku1 : get_schriftliche_kurse_ausser_lk()) {
                for (D_Kurs_SAT ku2 : get_schriftliche_kurse_ausser_lk()) {
                    if (ku1.fach.equals(ku2.fach)) {
                        if (ku1.kursart.equals(ku2.kursart)) {
                            ku1.regel_klausurbeginn_gemeinsam(ku2, WILDCARD, false);
                        }
                    }
                }
            }
        }

        private void regel_401_klausurbeginn_gemeinsam_stufe_a_fuer_kurs_b_und_kurs_c(String b, String c) throws IOException, UserException {
            for (D_Kurs_SAT ku1 : get_schriftliche_kurse_mit_kursnamen(b)) {
                for (D_Kurs_SAT ku2 : get_schriftliche_kurse_mit_kursnamen(c)) {
                    ku1.regel_klausurbeginn_gemeinsam(ku2, WILDCARD, false);
                }
            }
        }

        private void regel_402_klausurbeginn_gemeinsam_stufe_a_fuer_kurs_b_und_fach_kursart_c_d(String b, String c, String d) throws IOException, UserException {
            for (D_Kurs_SAT ku1 : get_schriftliche_kurse_mit_kursnamen(b)) {
                for (D_Kurs_SAT ku2 : get_schriftliche_kurse_mit_fach_kursart(c, d)) {
                    ku1.regel_klausurbeginn_gemeinsam(ku2, WILDCARD, false);
                }
            }
        }

        private void regel_403_stufe_a_fuer_fach_kursart_b_c(String b, String c) throws IOException, UserException {
            for (D_Kurs_SAT ku1 : get_schriftliche_kurse_mit_fach_kursart(b, c)) {
                for (D_Kurs_SAT ku2 : get_schriftliche_kurse_mit_fach_kursart(b, c)) {
                    ku1.regel_klausurbeginn_gemeinsam(ku2, WILDCARD, false);
                }
            }
        }

        private void regel_404_stufe_a_fuer_fach_kursart_b_c_und_fach_kursart_d_e(String b, String c, String d, String e) throws IOException, UserException {
            for (D_Kurs_SAT ku1 : get_schriftliche_kurse_mit_fach_kursart(b, c)) {
                for (D_Kurs_SAT ku2 : get_schriftliche_kurse_mit_fach_kursart(d, e)) {
                    ku1.regel_klausurbeginn_gemeinsam(ku2, WILDCARD, false);
                }
            }
        }

        private void regel_405_klausurbeginn_gemeinsam_stufe_a_alle_kurses_des_gleichen_tages() throws IOException, UserException {
            for (D_Kurs_SAT ku1 : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                for (D_Kurs_SAT ku2 : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                    ku1.regel_klausurbeginn_gemeinsam_wenn_gleicher_tag(ku2);
                }
            }
        }

        private void regel_406_klausurbeginn_gemeinsam_stufe_a_alle_kurse_der_gleichen_lehrkraft_pro_fach_kursart() throws IOException, UserException {
            for (D_Kurs_SAT ku1 : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                for (D_Kurs_SAT ku2 : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                    if (ku1.ist_gleiche_lehrkraft_wie(ku2) == true) {
                        if (ku1.ist_gleiche_fach_kursart(ku2) == true) {
                            ku1.regel_klausurbeginn_gemeinsam(ku2, WILDCARD, false);
                        }
                    }
                }
            }
        }

        private void regel_407_klausurbeginn_gemeinsam_stufe_a_alle_kurse_der_schiene_b(int b) throws IOException, UserException {
            for (D_Kurs_SAT ku1 : get_schriftliche_kurse_aus_schiene_0_indiziert(b)) {
                for (D_Kurs_SAT ku2 : get_schriftliche_kurse_aus_schiene_0_indiziert(b)) {
                    ku1.regel_klausurbeginn_gemeinsam(ku2, WILDCARD, false);
                }
            }
        }

        private void regel_408_klausurbeginn_gemeinsam_stufe_a_kurs_b_kurs_c_klausurnummer_d_auch_muendlich(String b, String c, String d) throws IOException, UserException {
            for (D_Kurs_SAT ku1 : get_schriftliche_kurse_mit_kursnamen(b)) {
                for (D_Kurs_SAT ku2 : get_schriftliche_kurse_mit_kursnamen(c)) {
                    ku1.regel_klausurbeginn_gemeinsam(ku2, d, true);
                }
            }
        }

        private void regel_409_klausurbeginn_gemeinsam_stufe_a_fach_kursart_b_c_klausurnummer_d_auch_muendlich(String b, String c, String d) throws IOException, UserException {
            for (D_Kurs_SAT ku1 : get_schriftliche_kurse_mit_fach_kursart(b, c)) {
                for (D_Kurs_SAT ku2 : get_schriftliche_kurse_mit_fach_kursart(b, c)) {
                    ku1.regel_klausurbeginn_gemeinsam(ku2, d, true);
                }
            }
        }

        private void regel_450_stufe_a_klausurbeginn_fixieren_alle_kurse_schiene_b_klausurnummer_c_datum_d_stunde_e(int b, int c, String d, int e) throws IOException, UserException {
            for (D_Kurs_SAT ku1 : get_schriftliche_kurse_aus_schiene_0_indiziert(b)) {
                ku1.regel_setze_klausurnummer_c_auf_datum_d_in_stunde_e(c, d, e);
            }
        }

        private void regel_451_klausurbeginn_ist_in_stufe_a_fuer_kurs_b_klausurnummer_c_datum_d_stunde_e(String b, int c, String d, int e) throws IOException, UserException {
            for (D_Kurs_SAT ku1 : get_schriftliche_kurse_mit_kursnamen(b)) {
                ku1.regel_setze_klausurnummer_c_auf_datum_d_in_stunde_e(c, d, e);
            }
        }

        private void regel_452_klausurbeginn_ist_in_stufe_a_fuer_kurs_fachart_b_c_klausurnummer_d_datum_e_stunde_f(String b, String c, int d, String e, int f) throws IOException, UserException {
            for (D_Kurs_SAT ku1 : get_schriftliche_kurse_mit_fach_kursart(b, c)) {
                ku1.regel_setze_klausurnummer_c_auf_datum_d_in_stunde_e(d, e, f);
            }
        }

        private void regel_453_klausurbeginn_ist_in_stufe_a_fuer_kurs_b_klausurnummer_c_zwischen_datum_d_und_e(String b, int c, String d, String e) throws ParseException, UserException, IOException {
            for (D_Kurs_SAT ku1 : get_schriftliche_kurse_mit_kursnamen(b)) {
                ku1.regel_setze_klausurnummer_c_zwischen_datum_d_und_e(c, d, e);
            }
        }

        private void regel_454_klausurbeginn_ist_in_stufe_a_fuer_fach_kursart_b_c_klausurnummer_d_zwischen_datum_e_und_f(String b, String c, int d, String e, String f) throws ParseException, UserException, IOException {
            for (D_Kurs_SAT ku1 : get_schriftliche_kurse_mit_fach_kursart(b, c)) {
                ku1.regel_setze_klausurnummer_c_zwischen_datum_d_und_e(d, e, f);
            }
        }

        private void regel_455_stufe_a_klausurbeginn_verbiete_stufe_a_stunde_b(int b) throws IOException, UserException {
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                ku.regel_klausurbeginn_verbieten_in_stunde(b);
            }
        }

        private void regel_456_stufe_a_klausurbeginn_verbiete_stufe_a_kursart_b_wochentag_c_von_stunde_d_bis_e(String b, String c, int d, int e) throws IOException, UserException {
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_fach_kursart(WILDCARD, b)) {
                ku.regel_klausurbeginn_verbieten_am_wochentag_c_von_stunde_d_bis_e(c, d, e);
            }
        }

        private void regel_add_vars_klausurschreiber(LinkedList<Integer> list, int w, int t, int s) {
            for (D_Schueler_SAT sc : mapS.values()) {
                sc.regel_add_vars_klausurschreiber(list, w, t, s);
            }
        }

        private void regel_900_optimiere_stufe_a_maximal_b_klausurtage(int b) throws IOException, UserException {
            c_at_most_GENERIC(toList(wt), b);
        }

        private void regel_907_optimiere_stufe_a_stundenplanzelle_maximal_b_klausurtermine(int b) throws IOException, UserException {
            for (int t = 0; t < ma.tage; t++) {
                for (int s = 0; s < ma.stunden; s++) {
                    LinkedList<Integer> listAlleWochen = new LinkedList<>();
                    for (int w = 0; w < ma.wochen; w++) {
                        LinkedList<Integer> listDieseWoche = new LinkedList<>();
                        for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                            if (ku.ist_schriftlich()) {
                                listDieseWoche.addLast(ku.wts_ohneM[w][t][s]);
                            }
                        }
                        int varOR = c_OR_tree(listDieseWoche);
                        listAlleWochen.addLast(varOR);
                    }
                    c_at_most_GENERIC(listAlleWochen, b);
                }
            }
        }

        private void regel_908_optimiere_stufe_a_stundenplanzelle_hoechstens_b_mal_maximal_c_klausurtermine(int b, int c) throws IOException, UserException {
            LinkedList<Integer> listHoechstens = new LinkedList<>();
            for (int t = 0; t < ma.tage; t++) {
                for (int s = 0; s < ma.stunden; s++) {

                    LinkedList<Integer> listAlleWochen = new LinkedList<>();
                    for (int w = 0; w < ma.wochen; w++) {
                        LinkedList<Integer> listDieseWoche = new LinkedList<>();
                        for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                            if (ku.ist_schriftlich()) {
                                listDieseWoche.addLast(ku.wts_ohneM[w][t][s]);
                            }
                        }
                        int varOR = c_OR_tree(listDieseWoche);
                        listAlleWochen.addLast(varOR);
                    }
                    c_bitonic_sort(listAlleWochen);

                    if ((c - 1 >= 0) && (c - 1 < listAlleWochen.size())) {
                        int var = listAlleWochen.get(c - 1);
                        listHoechstens.addLast(var);
                    }

                }
            }
            c_at_most_GENERIC(listHoechstens, b);
        }

        private void regel_909_optimiere_stufe_a_stundenplanzelle_am_tag_b_in_stunde_c_maximal_d_klausurtermine(String a, String b, String c, int d) throws IOException, UserException {
            for (int t = 0; t < WOCHENTAGE_VISUELL.length; t++) {
                if (b.equals(WILDCARD) || b.equals(WOCHENTAGE_VISUELL[t])) {
                    for (int s = 0; s < ma.stunden; s++) {
                        if (c.equals(WILDCARD) || c.equals((s + 1) + "")) {
                            LinkedList<Integer> listAlleWochen = new LinkedList<>();
                            for (int w = 0; w < ma.wochen; w++) {
                                LinkedList<Integer> listDieseWoche = new LinkedList<>();
                                for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                                    if (ku.ist_schriftlich()) {
                                        listDieseWoche.addLast(ku.wts_ohneM[w][t][s]);
                                    }
                                }
                                int varOR = c_OR_tree(listDieseWoche);
                                listAlleWochen.addLast(varOR);
                            }
                            c_at_most_GENERIC(listAlleWochen, d);
                        }
                    }
                }
            }
        }

        private void regel_910_optimiere_stufe_a_kurs_b_maximal_c_fremdklausurstunden(String b, int c) throws IOException, UserException {
            for (D_Kurs_SAT ku : get_alle_kurse(b)) { // WICHTIG ALLE KURSE!!!
                ku.regel_910_optimiere_stufe_a_kurs_b_maximal_c_fremdklausurstunden(c);
            }
        }

        private LinkedList<Integer> regel_get_fremdklausurstunden() throws IOException, UserException {
            LinkedList<Integer> list = new LinkedList<>();
            for (D_Kurs_SAT ku : get_alle_kurse(WILDCARD)) { // WICHTIG ALLE KURSE!!!
                list.addAll(ku.regel_get_fremdklausurstunden());
            }
            return list;
        }

        private void regel_911_optimiere_stufe_a_maximal_b_fremdklausurstunden(int b) throws IOException, UserException {
            c_at_most_GENERIC(regel_get_fremdklausurstunden(), b);
        }

        private void regel_920_optimiere_stufe_a_fach_kursart_b_c_mindestens_d_klausurstunden_laut_plan(String b, String c, int d) throws IOException, UserException {
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_fach_kursart(b, c)) {
                c_at_least_GENERIC(ku.regel_gib_klausurstunden_nach_plan(), d);
            }
        }

        private void regel_921_optimiere_stufe_a_summe_mindestens_b_klausurstunden_laut_plan(int b) throws IOException, UserException {
            LinkedList<Integer> list = new LinkedList<>();
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                list.addAll(ku.regel_gib_klausurstunden_nach_plan());
            }
            c_at_least_GENERIC(list, b);
        }

        private LinkedList<Integer> regel_gib_klausurstunden_nach_plan() {
            LinkedList<Integer> list = new LinkedList<>();
            for (D_Kurs_SAT ku : get_schriftliche_kurse_mit_kursnamen(WILDCARD)) {
                list.addAll(ku.regel_gib_klausurstunden_nach_plan());
            }
            return list;
        }

        private boolean ist_klausur_stunde(int w, int t, int s) {
            return c_get(wts_ohneM[w][t][s]);
        }

        private boolean ist_klausur_tag(int w, int t) {
            return c_get(wt[w][t]);
        }

    }

    private final class D_Schueler_SAT {

        final D_Main_SAT ma;
        final D_Jahrgang_SAT ja;
        final String key;
        final String name;
        final String vorname;
        final String geburtsdatum;
        final String geschlecht;
        final int nr;
        final String dbid;
        final TreeMap<String, D_Kurs_SAT> mapK_alle;
        final TreeMap<String, D_Kurs_SAT> mapK_schriftlich;

        int[][] wt; // [Woche, Tag]
        int[][][] wts; // [Woche, Tag, Stunde]
        int[][][] wts_ohneM; // Woche, Tag, Stunde ohne müdliche

        private D_Schueler_SAT(D_Main_SAT ma, D_Jahrgang_SAT ja, String key, D_Map mapSchueler) {
            this.ma = ma;
            this.ja = ja;
            this.key = key;
            this.nr = mapSchueler.get(ID_SCHUELER_NR).first_key_int();
            this.name = mapSchueler.get(ID_SCHUELER_NAME).first_key_string();
            this.vorname = mapSchueler.get(ID_SCHUELER_VORNAME).first_key_string();
            this.geburtsdatum = mapSchueler.get(ID_SCHUELER_GEBDATUM).first_key_string();
            this.geschlecht = mapSchueler.get(ID_SCHUELER_GESCHLECHT).first_key_string();
            this.dbid = mapSchueler.get(ID_SCHUELER_DBID).first_key_string();
            this.mapK_alle = new TreeMap<>();
            this.mapK_schriftlich = new TreeMap<>();
        }

        private void verbindeK(D_Map mapSchueler) {
            for (String kursname : mapSchueler.get(ID_SCHUELER_KURSE_ALLE).keySetSortedString()) {
                D_Kurs_SAT ku = ja.mapK.get(kursname);
                mapK_alle.put(kursname, ku);
            }
            for (String kursname : mapSchueler.get(ID_SCHUELER_KURSE_SCHRIFTLICH).keySetSortedString()) {
                D_Kurs_SAT ku = ja.mapK.get(kursname);
                mapK_schriftlich.put(kursname, ku);
            }
        }

        private void cnf_pre() {
            wt = ma.get_array_wt();
            wts = ma.get_array_wts();
            wts_ohneM = ma.get_array_wts();
        }

        private void cnf_post() throws IOException, UserException {
            // wt, wts, wts_ohneM 
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    LinkedList<Integer> listWT = new LinkedList<>();
                    for (D_Kurs_SAT ku : mapK_schriftlich.values()) {
                        ku.add_w_t(listWT, w, t);
                    }
                    int rootWT = c_at_most_one_tree(listWT); // pro tag max. eine Klausur
                    c_equal(wt[w][t], rootWT);

                    for (int s = 0; s < ma.stunden; s++) {
                        LinkedList<Integer> listWTS = new LinkedList<>();
                        LinkedList<Integer> listWTS_ohne = new LinkedList<>();
                        for (D_Kurs_SAT ku : mapK_schriftlich.values()) {
                            ku.add_w_t_s(listWTS, w, t, s); // pro stunde max. eine Klausur
                            ku.add_w_t_s_ohne(listWTS_ohne, w, t, s); // pro stunde max. eine Klausur
                        }
                        int rootWTS = c_at_most_one_tree(listWTS);
                        int rootWTS_ohne = c_at_most_one_tree(listWTS_ohne);
                        c_equal(wts[w][t][s], rootWTS);
                        c_equal(wts_ohneM[w][t][s], rootWTS_ohne);
                    }
                }
            }

            // wts, wts_ohneM 
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    for (int s = 0; s < ma.stunden; s++) {
                        LinkedList<Integer> listWTS = new LinkedList<>();
                        LinkedList<Integer> listWTS_ohne = new LinkedList<>();
                        for (D_Kurs_SAT ku : mapK_schriftlich.values()) {
                            ku.add_w_t_s(listWTS, w, t, s); // pro stunde max. eine Klausur
                            ku.add_w_t_s_ohne(listWTS_ohne, w, t, s); // pro stunde max. eine Klausur
                        }
                        int rootWTS = c_at_most_one_tree(listWTS);
                        int rootWTS_ohne = c_at_most_one_tree(listWTS_ohne);
                        c_equal(wts[w][t][s], rootWTS);
                        c_equal(wts_ohneM[w][t][s], rootWTS_ohne);
                    }
                }
            }

            // max_klausuren_pro_schueler_pro_woche
            if (ma.max_klausuren_pro_schueler_pro_woche > 0) {
                for (int w = 0; w < ma.wochen; w++) {
                    c_at_most_GENERIC(toList(wt[w]), ma.max_klausuren_pro_schueler_pro_woche);
                }
            }
        }

        private String get_ausgabe_matrix() {
            String s = "";
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    if (ma.planB[w][t] == true) {
                        String sKurs = "0";
                        for (D_Kurs_SAT ku : mapK_schriftlich.values()) {
                            if (c_get(ku.wt[w][t])) {
                                if (ku.ist_muendlich(w, t)) {
                                    sKurs = "M";
                                } else {
                                    if (ku.ist_nicht_lk()) {
                                        sKurs = "G";
                                    } else {
                                        sKurs = "L";
                                    }
                                }
                            }
                        }
                        s = s + sKurs;
                    } else {
                        s = s + "x";
                    }

                }
                s = s + " ";
            }
            return s;
        }

        private int get_ausgabe_2er() {
            int summe = 0;
            boolean back1 = false;
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    boolean current = c_get(wt[w][t]);
                    if (back1 && current) {
                        summe++;
                    }
                    back1 = current;
                }
            }
            return summe;
        }

        private int get_ausgabe_3er() {
            int summe = 0;
            boolean back1 = false;
            boolean back2 = false;
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    boolean current = c_get(wt[w][t]);
                    if (back1 && back2 && current) {
                        summe++;
                    }
                    back2 = back1;
                    back1 = current;
                }
            }
            return summe;
        }

        private int get_klausuren_gesamt() {
            int summe = 0;
            for (D_Kurs_SAT ku : mapK_schriftlich.values()) {
                summe += ku.get_klausuren_insgesamt();
            }
            return summe;
        }

        private int get_anzahl_an_wochen_mit_2_klausuren() {
            int summe = 0;
            for (int w = 0; w < ma.wochen; w++) {
                int wochen_counter = 0;
                for (int t = 0; t < ma.tage; t++) {
                    boolean current = c_get(wt[w][t]);
                    if (current) {
                        wochen_counter++;
                    }
                }
                if (wochen_counter == 2) {
                    summe++;
                }
            }
            return summe;
        }

        private int get_anzahl_an_wochen_mit_3_klausuren() {
            int summe = 0;
            for (int w = 0; w < ma.wochen; w++) {
                int wochen_counter = 0;
                for (int t = 0; t < ma.tage; t++) {
                    boolean current = c_get(wt[w][t]);
                    if (current) {
                        wochen_counter++;
                    }
                }
                if (wochen_counter == 3) {
                    summe++;
                }
            }
            return summe;
        }

        private int get_anzahl_an_wochen_mit_2_klausuren_in_folge() {
            int summe = 0;
            for (int w = 0; w < ma.wochen; w++) {
                int in_folge_counter = 0;
                for (int t = 0; t < ma.tage; t++) {
                    boolean current = c_get(wt[w][t]);
                    if (current) {
                        in_folge_counter++;
                    } else {
                        if (in_folge_counter == 2) {
                            summe++;
                        }
                        in_folge_counter = 0;
                    }
                }
                if (in_folge_counter == 2) {
                    summe++;
                }
            }
            return summe;
        }

        private int get_anzahl_an_wochen_mit_3_klausuren_in_folge() {
            int summe = 0;
            for (int w = 0; w < ma.wochen; w++) {
                int in_folge_counter = 0;
                for (int t = 0; t < ma.tage; t++) {
                    boolean current = c_get(wt[w][t]);
                    if (current) {
                        in_folge_counter++;
                    } else {
                        if (in_folge_counter == 3) {
                            summe++;
                        }
                        in_folge_counter = 0;
                    }
                }
                if (in_folge_counter == 3) {
                    summe++;
                }
            }
            return summe;
        }

        private void regel_300_stufe_a_schueler_b_nicht_c_klausurtage_aufeinanderfolgend(int c) throws IOException, UserException {
            c_at_most_GENERIC(regel_get_wochensumme_klausurtage_aufeinanderfolgend(c), 0);
        }

        private void regel_301_stufe_a_schueler_b_wochensumme_mit_c_klausurtage_aufeinanderfolgend_max_d(int c, int d) throws IOException, UserException {
            c_at_most_GENERIC(regel_get_wochensumme_klausurtage_aufeinanderfolgend(c), d);
        }

        private LinkedList<Integer> regel_get_wochensumme_klausurtage_aufeinanderfolgend(int c) throws IOException, UserException {
            LinkedList<Integer> list = new LinkedList<>();
            if (c == 2) {
                for (int w = 0; w < ma.wochen; w++) {
                    for (int t = 1; t < ma.tage; t++) {
                        int var = c_new_var_AND(wt[w][t - 1], wt[w][t]);
                        list.addLast(var);
                    }
                }
            }
            if (c == 3) {
                for (int w = 0; w < ma.wochen; w++) {
                    for (int t = 2; t < ma.tage; t++) {
                        int var = c_new_var_AND(c_new_var_AND(wt[w][t - 2], wt[w][t - 1]), wt[w][t]);
                        list.addLast(var);
                    }
                }
            }
            return list;
        }

        private void regel_304_stufe_a_schueler_b_mit_fach_kursart_c_d_implizieren_danach_keinen_klausurtag(String c, String d) throws IOException, UserException {
            LinkedList<D_Kurs_SAT> list = get_schuelerkurse_schriftlich_mit_fach_kursart(c, d);
            int[][] tempWT = get_schueler_kurs_AMO_ueberlagerung(list);
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 1; t < ma.tage; t++) {
                    c_2(-tempWT[w][t - 1], -wt[w][t]);
                }
            }
        }

        private void regel_305_stufe_a_schueler_b_mit_fach_kursart_c_d_implizieren_davor_keinen_klausurtag(String c, String d) throws IOException, UserException {
            LinkedList<D_Kurs_SAT> list = get_schuelerkurse_schriftlich_mit_fach_kursart(c, d);
            int[][] tempWT = get_schueler_kurs_AMO_ueberlagerung(list);
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 1; t < ma.tage; t++) {
                    c_2(-wt[w][t - 1], -tempWT[w][t]);
                }
            }
        }

        private void regel_306_stufe_a_schueler_b_mit_fach_kursart_c_d_nicht_in_gleicher_woche_mit_fachart_kursart_e_f(String c, String d, String e, String f) throws IOException, UserException {
            LinkedList<D_Kurs_SAT> list1 = get_schuelerkurse_schriftlich_mit_fach_kursart(c, d);
            LinkedList<D_Kurs_SAT> list2 = get_schuelerkurse_schriftlich_mit_fach_kursart(e, f);
            int[][] tempWT1 = get_schueler_kurs_AMO_ueberlagerung(list1);
            int[][] tempWT2 = get_schueler_kurs_AMO_ueberlagerung(list2);
            // nicht beides
            for (int w = 0; w < ma.wochen; w++) {
                LinkedList<Integer> listWoche1 = new LinkedList<>();
                LinkedList<Integer> listWoche2 = new LinkedList<>();
                for (int t = 0; t < ma.tage; t++) {
                    listWoche1.addLast(tempWT1[w][t]);
                    listWoche2.addLast(tempWT2[w][t]);
                }
                int istInWocheVar1 = c_OR_tree(listWoche1);
                int istInWocheVar2 = c_OR_tree(listWoche2);
                c_2(-istInWocheVar1, -istInWocheVar2);
            }
        }

        private void regel_307_stufe_a_schueler_b_mit_fach_kursart_c_d_begrenzen_klausurtage_auf_e(String c, String d, int e) throws IOException, UserException {
            LinkedList<D_Kurs_SAT> list = get_schuelerkurse_schriftlich_mit_fach_kursart(c, d);
            int[][] temp = get_schueler_kurs_AMO_ueberlagerung(list);
            // begrenzen
            for (int w = 0; w < ma.wochen; w++) {
                LinkedList<Integer> listWoche = new LinkedList<>();
                LinkedList<Integer> listWT = new LinkedList<>();
                for (int t = 0; t < ma.tage; t++) {
                    listWoche.addLast(temp[w][t]);
                    listWT.addLast(wt[w][t]);
                }
                // WENN var DANN max e Klausurtage
                c_bitonic_sort(listWT);
                int var1 = c_OR_tree(listWoche);
                int var2 = listWT.get(e);
                c_2(-var1, -var2);
            }
        }

        private LinkedList<D_Kurs_SAT> get_schuelerkurse_schriftlich_mit_fach_kursart(String b, String c) {
            LinkedList<D_Kurs_SAT> list = new LinkedList<>();
            for (D_Kurs_SAT ku : mapK_schriftlich.values()) {
                if (b.equals(WILDCARD) || ku.fach.equals(b)) {
                    if (c.equals(WILDCARD) || ku.kursart.equals(c)) {
                        list.addLast(ku);
                    }
                }
            }
            return list;
        }

        private int[][] get_schueler_kurs_AMO_ueberlagerung(LinkedList<D_Kurs_SAT> listK) throws IOException, UserException {
            int[][] temp = new int[ma.wochen][ma.tage];
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    LinkedList<Integer> listAMO = new LinkedList<>();
                    for (D_Kurs_SAT ku : listK) {
                        ku.add_w_t(listAMO, w, t);
                    }
                    temp[w][t] = c_at_most_one_tree(listAMO);
                }
            }
            return temp;
        }

        private void regel_add_vars_klausurschreiber(LinkedList<Integer> list, int w, int t, int s) {
            list.addLast(wts_ohneM[w][t][s]);
        }

    }

    private final class D_Kurs_SAT {

        final D_Main_SAT ma;
        final D_Jahrgang_SAT ja;
        final String name;
        final int kurs_nr;
        final String fach;
        final String lehrkraft;
        final int std;
        final String raum;
        final String kursart;
        final TreeMap<String, D_Schueler_SAT> mapS_alle;
        final TreeMap<String, D_Schueler_SAT> mapS_schriftlich;
        final TreeMap<Integer, D_Klausur_SAT> mapKL;
        final TreeSet<Integer> schienen;
        final boolean[][][] laut_plan;
        int klausurenanzahl;

        int[][] wt; // [Woche, Tag]
        int[][][] wts; // Woche, Tag, Stunde
        int[][][] wts_ohneM; // Woche, Tag, Stunde ohne müdliche

        private D_Kurs_SAT(D_Main_SAT ma, D_Jahrgang_SAT ja, String name, D_Map mapKurs) {
            this.ma = ma;
            this.ja = ja;
            this.name = name;
            this.kurs_nr = mapKurs.get(ID_KURS_NR).first_key_int();
            this.fach = mapKurs.get(ID_KURS_FACH).first_key_string();
            this.lehrkraft = mapKurs.get(ID_KURS_LEHRER).first_key_string();
            this.std = mapKurs.get(ID_KURS_STD).first_key_int();
            this.raum = mapKurs.get(ID_KURS_RAUM).first_key_string();
            this.kursart = mapKurs.get(ID_KURS_KURSART).first_key_string();
            this.mapS_alle = new TreeMap<>();
            this.mapS_schriftlich = new TreeMap<>();
            this.mapKL = new TreeMap<>();
            this.schienen = mapKurs.get(ID_KURS_SCHIENEN).keySetSortedInteger();
            this.klausurenanzahl = 0;

            this.laut_plan = new boolean[2][ma.tage][ma.stunden];
            for (String sWoche : mapKurs.get(ID_KURS_PLAN).keySetSortedString()) {
                for (String sTag : mapKurs.get(ID_KURS_PLAN).get(sWoche).keySetSortedString()) {
                    for (String sStunde : mapKurs.get(ID_KURS_PLAN).get(sWoche).get(sTag).keySetSortedString()) {
                        for (String sRaum : mapKurs.get(ID_KURS_PLAN).get(sWoche).get(sTag).get(sStunde).keySetSortedString()) {
                            for (String sLehrer : mapKurs.get(ID_KURS_PLAN).get(sWoche).get(sTag).get(sStunde).get(sRaum).keySetSortedString()) {
                                int iWoche = Integer.parseInt(sWoche); // TODO ggf. iWoche drehen?
                                int iTag = Integer.parseInt(sTag);
                                int iStunde = Integer.parseInt(sStunde);
                                this.laut_plan[iWoche][iTag - 1][iStunde - 1] = true;
                                //("  "+sWoche+", "+sTag+", "+sStunde+", "+sRaum+", "+sLehrer);
                            }
                        }
                    }
                }
            }

        }

        private void verbindeS(D_Map mapKurs) {
            for (String schueler_key : mapKurs.get(ID_KURS_SCHUELER_ALLE).keySetSortedString()) {
                D_Schueler_SAT sc = ja.mapS.get(schueler_key);
                mapS_alle.put(schueler_key, sc);
            }
            for (String schueler_key : mapKurs.get(ID_KURS_SCHUELER_SCHRIFTLICH).keySetSortedString()) {
                D_Schueler_SAT sc = ja.mapS.get(schueler_key);
                mapS_schriftlich.put(schueler_key, sc);
            }
        }

        private boolean ist_schriftlich() {
            return mapS_schriftlich.size() > 0;
        }

        private void add_w_t(LinkedList<Integer> list, int w, int t) {
            if (ist_schriftlich() == false) {
                return;
            }
            if (ma.planB[w][t] == false) {
                return;
            }
            list.addLast(wt[w][t]);
        }

        private void add_w_t_s(LinkedList<Integer> listWTS, int w, int t, int s) {
            if (ist_schriftlich() == false) {
                return;
            }
            if (ma.planB[w][t] == false) {
                return;
            }
            listWTS.addLast(wts[w][t][s]);
        }

        private void add_w_t_s_ohne(LinkedList<Integer> listWTS_ohne, int w, int t, int s) {
            if (ist_schriftlich() == false) {
                return;
            }
            if (ma.planB[w][t] == false) {
                return;
            }
            listWTS_ohne.addLast(wts_ohneM[w][t][s]);
        }

        private int get_max_klausur_nr() {
            return mapKL.size();
        }

        private int get_klausurstunden_laut_plan() {
            int summe = 0;
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    if (ma.planB[w][t]) {
                        int wocheMOD2 = Integer.parseInt(Static.datum_kalenderwoche(ma.planS[w][t])) % 2;
                        for (int s = 0; s < ma.stunden; s++) {
                            if (laut_plan[wocheMOD2][t][s]) {
                                if (c_get(wts_ohneM[w][t][s])) {
                                    summe++;
                                }
                            }
                        }
                    }
                }
            }
            return summe;
        }

        private int get_klausurstunden_insgesamt() {
            int summe = 0;
            for (D_Klausur_SAT kl : mapKL.values()) {
                summe += kl.get_klausurstunden_insgesamt();
            }
            return summe;
        }

        private boolean ist_nicht_lk() {
            return !kursart.equals("LK");
        }

        private boolean ist_gleiche_lehrkraft_wie(D_Kurs_SAT that) {
            if (this.lehrkraft.equals(that.lehrkraft) == false) {
                return false;
            }
            if (this.ist_dummy_lehrkraft() == true) {
                return false;
            }
            return true;
        }

        private boolean ist_dummy_lehrkraft() {
            if (this.lehrkraft.equals("-")) {
                return false;
            }
            if (this.lehrkraft.equals("-")) {
                return false;
            }
            if (this.lehrkraft.equals("--")) {
                return false;
            }
            if (this.lehrkraft.equals("---")) {
                return false;
            }
            return this.lehrkraft.equals("");
        }

        private boolean ist_muendlich(int w, int t) {
            for (D_Klausur_SAT kl : mapKL.values()) {
                if (kl.ist_muendlich()) {
                    if (c_get(kl.wt[w][t])) {
                        return true;
                    }
                }
            }
            return false;
        }

        private boolean ist_gleiche_fach_kursart(D_Kurs_SAT that) {
            return this.fach.equals(that.fach) && this.kursart.equals(that.kursart);
        }

        private void create_klausuren() {
            if (ist_schriftlich() == false) {
                return;
            }
            for (int kl_nr = 1; kl_nr <= klausurenanzahl; kl_nr++) {
                D_Klausur_SAT kl = new D_Klausur_SAT(ma, this, kl_nr);
                mapKL.put(kl_nr, kl);
            }
        }

        private void cnf_pre() {
            wt = ma.get_array_wt();
            wts = ma.get_array_wts();
            wts_ohneM = ma.get_array_wts();

            // PRE rekursiv
            for (D_Klausur_SAT kl : mapKL.values()) {
                kl.cnf_pre();
            }
        }

        private void cnf_post() throws IOException, UserException {
            // wt, wts, wts_ohneM
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    // wt = c_at_most_one_tree 
                    LinkedList<Integer> listWT = new LinkedList<>();
                    for (D_Klausur_SAT kl : mapKL.values()) {
                        listWT.addLast(kl.wt[w][t]);
                    }
                    int varOR = c_at_most_one_tree(listWT);
                    c_equal(varOR, wt[w][t]);
                    // wts, wts_ohneM = c_at_most_one_tree
                    for (int s = 0; s < ma.stunden; s++) {
                        LinkedList<Integer> listWTS = new LinkedList<>();
                        LinkedList<Integer> listWTS_ohne = new LinkedList<>();
                        for (D_Klausur_SAT kl : mapKL.values()) {
                            listWTS.addLast(kl.wts[w][t][s]);
                            if (kl.muendliche_tage <= 0) {
                                listWTS_ohne.addLast(kl.wts[w][t][s]);
                            }
                        }
                        int rootWTS = c_at_most_one_tree(listWTS);
                        int rootWTS_ohneM = c_at_most_one_tree(listWTS_ohne);
                        c_equal(wts[w][t][s], rootWTS);
                        c_equal(wts_ohneM[w][t][s], rootWTS_ohneM);
                    }
                }
            }

            // POST rekursiv
            for (D_Klausur_SAT kl : mapKL.values()) {
                kl.cnf_post();
            }
        }

        private void erzeuge_ausgabe_matrix() {
            if (ist_schriftlich() == false) {
                return;
            }
            String s = "";
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    if (ma.planB[w][t] == true) {
                        s = s + (c_get(wt[w][t]) ? "1" : "0");
                    } else {
                        s = s + "x";
                    }

                }
                s = s + " ";
            }
            ausgabe.html_append(s + "   " + name + "[" + mapKL.size() + "]<br>");

            /*
            for (D_Klausur_SAT kl : mapKL.values()) {
                kl.erzeuge_ausgabe_matrix();
            }
             */
        }

        private void regel_setze_klausurenanzahl_auf(int anzahl) {
            if (ist_schriftlich() == false) {
                return;
            }
            klausurenanzahl = anzahl;
        }

        private void regel_setze_klausur_nr_entfaellt(int nummer) {
            if (ist_schriftlich() == false) {
                return;
            }
            mapKL.get(nummer).entfaellt = true;
        }

        private void regel_setze_klausurenabstand_auf(int abstand) {
            if (ist_schriftlich() == false) {
                return;
            }
            for (D_Klausur_SAT kl : mapKL.values()) {
                kl.regel_setze_abstand_zur_naechsten_klausur_auf(abstand);
            }
        }

        private void regel_setze_klausurnummer_c_auf_laenge_d(String nummer, int laenge) {
            if (ist_schriftlich() == false) {
                return;
            }
            for (D_Klausur_SAT kl : mapKL.values()) {
                if (nummer.equals(WILDCARD) || (nummer.equals("" + kl.klausurnummer))) {
                    kl.regel_setze_klausurlaenge_auf(laenge);
                }
            }
        }

        private void regel_201_plan_stufe_a_stunden_b_bis_c_gesperrt(int b, int c) throws IOException, UserException {
            if (ist_schriftlich() == false) {
                return;
            }
            for (D_Klausur_SAT kl : mapKL.values()) {
                kl.regel_stunden_b_bis_c_gesperrt(b, c);
            }
        }

        private void regel_202_plan_stufe_a_datum_b_stunden_c_bis_d_gesperrt(String b, int c, int d) throws ParseException, UserException, IOException {
            if (ist_schriftlich() == false) {
                return;
            }
            for (D_Klausur_SAT kl : mapKL.values()) {
                kl.regel_datum_b_stunden_c_bis_d_gesperrt(b, c, d);
            }
        }

        private void regel_klausurbeginn_gemeinsam(D_Kurs_SAT that, String sKlausurnummer, boolean auch_muendlich) throws IOException, UserException {
            if (!this.ist_schriftlich() || !that.ist_schriftlich()) {
                return;
            }
            if (sKlausurnummer.equals(WILDCARD)) {
                for (int klausurnummer : mapKL.keySet()) {
                    if (that.mapKL.containsKey(klausurnummer)) {
                        D_Klausur_SAT kl1 = this.mapKL.get(klausurnummer);
                        D_Klausur_SAT kl2 = that.mapKL.get(klausurnummer);
                        kl1.regel_klausurbeginn_gemeinsam(kl2, auch_muendlich);
                    }
                }
            } else {
                int klausurnummer = Integer.parseInt(sKlausurnummer);
                if (this.mapKL.containsKey(klausurnummer)) {
                    if (that.mapKL.containsKey(klausurnummer)) {
                        D_Klausur_SAT kl1 = this.mapKL.get(klausurnummer);
                        D_Klausur_SAT kl2 = that.mapKL.get(klausurnummer);
                        kl1.regel_klausurbeginn_gemeinsam(kl2, auch_muendlich);
                    }
                }
            }
        }

        private void regel_setze_klausurnummer_c_auf_datum_d_in_stunde_e(int c, String d, int e) throws IOException, UserException {
            if (mapKL.containsKey(c)) {
                mapKL.get(c).regel_setze_auf_datum_d_in_stunde_e(d, e);
            }
        }

        private void regel_setze_klausurnummer_c_zwischen_datum_d_und_e(int c, String d, String e) throws ParseException, UserException, IOException {
            if (mapKL.containsKey(c)) {
                mapKL.get(c).regel_setze_zwischen_datum_d_und_e(d, e);
            }
        }

        private void regel_klausurbeginn_gemeinsam_wenn_gleicher_tag(D_Kurs_SAT that) throws IOException, UserException {
            for (D_Klausur_SAT kl1 : this.mapKL.values()) {
                for (D_Klausur_SAT kl2 : that.mapKL.values()) {
                    kl1.regel_klausurbeginn_gemeinsam_wenn_gleicher_tag(kl2);
                }
            }
        }

        private void regel_klausurbeginn_verbieten_in_stunde(int b) throws IOException, UserException {
            for (D_Klausur_SAT kl : mapKL.values()) {
                kl.regel_klausurbeginn_verbieten_in_stunde(b);
            }
        }

        private void regel_klausurbeginn_verbieten_am_wochentag_c_von_stunde_d_bis_e(String c, int d, int e) throws IOException, UserException {
            for (D_Klausur_SAT kl : mapKL.values()) {
                kl.regel_klausurbeginn_verbieten_am_wochentag_c_von_stunde_d_bis_e(c, d, e);
            }
        }

        private void regel_setze_klausurnummer_c_muendlich_mit_d_tagen(int c, int d) {
            if (mapKL.containsKey(c)) {
                mapKL.get(c).regel_setze_muendlich_mit_d_tagen(d);
                mapKL.get(c).regel_setze_klausurlaenge_auf(ma.stunden);
            }
        }

        private LinkedList<Integer> regel_gib_klausurstunden_nach_plan() {
            LinkedList<Integer> list = new LinkedList<>();
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    for (int s = 0; s < ma.stunden; s++) {
                        if (ist_planmaessig(w, t, s) == true) {
                            list.addLast(wts_ohneM[w][t][s]);
                        }
                    }
                }
            }
            return list;
        }

        private void regel_910_optimiere_stufe_a_kurs_b_maximal_c_fremdklausurstunden(int c) throws IOException, UserException {
            c_at_most_GENERIC(regel_get_fremdklausurstunden(), c);
        }

        private LinkedList<Integer> regel_get_fremdklausurstunden() throws IOException, UserException {
            LinkedList<Integer> list = new LinkedList<>();
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    if (ma.planB[w][t]) {
                        int wocheMOD2 = Integer.parseInt(Static.datum_kalenderwoche(ma.planS[w][t])) % 2;
                        for (int s = 0; s < ma.stunden; s++) {
                            if (laut_plan[wocheMOD2][t][s]) {
                                // 1) Der Kurs hat Unterricht in Zelle [w][t][s]
                                // 2) Der Kurs schreibt selbst nicht
                                int varSchreibtNicht = -wts[w][t][s];
                                // 3) Mindestens ein Schüler des Kurses schreibt
                                LinkedList<Integer> listS = new LinkedList<>();
                                for (D_Schueler_SAT sc : mapS_schriftlich.values()) {
                                    listS.addLast(sc.wts[w][t][s]);
                                }
                                int varSchuelerSchreibt = c_OR_tree(listS);
                                // 4) Resultat
                                int varGestoert = c_new_var_AND(varSchreibtNicht, varSchuelerSchreibt);
                                list.addLast(varGestoert);
                            }
                        }
                    }
                }
            }
            return list;
        }

        private int get_unterrichtsstunden_mit_fehlenden_sus() {
            int summe = 0;
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    for (int s = 0; s < ma.stunden; s++) {
                        if (ist_planmaessig(w, t, s)) { // 1) Der Kurs hat Unterricht in Zelle [w][t][s]
                            if (c_get(-wts[w][t][s])) { // 2) Der Kurs schreibt nicht selbst
                                int fehlend = 0;
                                for (D_Schueler_SAT sc : mapS_schriftlich.values()) {
                                    if (c_get(sc.wts_ohneM[w][t][s])) { // ignoriere mündliche
                                        fehlend++;
                                    }
                                }
                                if (fehlend > 0) { // 3) Es fehlt mindestens ein S.
                                    summe++;
                                }
                            }
                        }
                    }
                }
            }
            return summe;
        }

        private int get_unterrichtsstunden_mit_fehlenden_sus_im_eigenen_kurs() {
            int summe = 0;
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    for (int s = 0; s < ma.stunden; s++) {
                        if (ist_planmaessig(w, t, s)) { // 1) Der Kurs hat Unterricht in Zelle [w][t][s]
                            if (c_get(wts[w][t][s])) { // 2) Der Kurs schreibt selbst
                                int fehlend = 0;
                                for (D_Schueler_SAT sc : mapS_schriftlich.values()) {
                                    if (c_get(sc.wts_ohneM[w][t][s])) { // ignoriere mündliche
                                        fehlend++;
                                    }
                                }
                                if (fehlend > 0) { // 3) Es fehlt mindestens ein S.
                                    summe++;
                                }
                            }
                        }
                    }
                }
            }
            return summe;
        }

        private int get_klausuren_insgesamt() {
            int summe = 0;
            for (D_Klausur_SAT kl : mapKL.values()) {
                if (kl.entfaellt == false) {
                    summe++;
                }
            }
            return summe;
        }

        private int get_klausurschueler_in(int w, int t, int s) {
            int summe = 0;
            for (D_Schueler_SAT sc : mapS_schriftlich.values()) {
                if (c_get(sc.wts_ohneM[w][t][s])) {
                    summe++;
                }
            }
            return summe;
        }

        private int get_gesamtschueler() {
            return mapS_alle.size();
        }

        private boolean ist_planmaessig(int w, int t, int s) {
            int wocheMOD2 = Integer.parseInt(Static.datum_kalenderwoche(ma.planS[w][t])) % 2;
            return laut_plan[wocheMOD2][t][s];
        }

    }

    private final class D_Klausur_SAT {

        final D_Main_SAT ma;
        final D_Kurs_SAT parent;
        final int klausurnummer;

        int klausurlaenge;
        int muendliche_tage;
        boolean entfaellt;
        int klausurabstand;

        int[][] wt; // woche, tag
        int[][] wt_nicht_davor; // [Woche, Tag]
        int[][][] wts; // Woche, Tag, Stunde
        int[] startet_in_der_stunde;
        int[] ist_in_der_stunde;

        public D_Klausur_SAT(D_Main_SAT ma, D_Kurs_SAT parent, int nr) {
            this.ma = ma;
            this.parent = parent;
            this.klausurnummer = nr;

            this.klausurlaenge = 1;
            this.muendliche_tage = -1;
            this.entfaellt = false;
            this.klausurabstand = 1;
        }

        private void regel_setze_abstand_zur_naechsten_klausur_auf(int abstand) {
            this.klausurabstand = abstand;
        }

        private void regel_setze_klausurlaenge_auf(int laenge) {
            this.klausurlaenge = laenge;
        }

        private boolean ist_muendlich() {
            return muendliche_tage >= 1;
        }

        private void cnf_pre() {
            this.wt = ma.get_array_wt();
            this.wt_nicht_davor = c_new_var_matrix(ma.wochen, ma.tage);
            this.startet_in_der_stunde = c_new_var_array(ma.stunden);
            this.ist_in_der_stunde = c_new_var_array(ma.stunden);
            this.wts = ma.get_array_wts();
        }

        private void cnf_post() throws IOException, UserException {
            if (entfaellt) {
                c_exactly_GENERIC(toList(wt), 0);
                c_exactly_GENERIC(toList(wts), 0);
                c_exactly_GENERIC(toList(startet_in_der_stunde), 0);
                c_exactly_GENERIC(toList(ist_in_der_stunde), 0);
                return;
            }

            // wt
            c_exactly_GENERIC(toList(wt), 1);

            // wt_nicht_davor
            int prev = 0;
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    if (prev != 0) {
                        c_2(-wt_nicht_davor[w][t], prev); // Wenn TRUE dann davor auch
                    }
                    c_2(-wt_nicht_davor[w][t], -wt[w][t]); // Wenn TRUE dann wt FALSE
                    prev = wt_nicht_davor[w][t];
                }
            }

            // Klausurabstände zu allen nächsten Klausuren
            for (D_Klausur_SAT that : parent.mapKL.values()) {
                if (klausurnummer < that.klausurnummer) {
                    LinkedList<Integer> list1 = toList(this.wt);
                    LinkedList<Integer> list2 = toList(that.wt_nicht_davor);
                    // Wenn this.wt --> that.wt_nicht_davor
                    for (int i = 0; i < list1.size(); i++) {
                        if (i + klausurabstand - 1 < list2.size()) {
                            int var1 = list1.get(i);
                            int var2 = list2.get(i + klausurabstand - 1);
                            c_2(-var1, var2);
                        }
                    }
                }
            }

            // startet_in_der_stunde
            c_exactly_GENERIC(toList(startet_in_der_stunde), 1);
            for (int s = ma.stunden - klausurlaenge + 1; s < ma.stunden; s++) {
                c_1(-startet_in_der_stunde[s]);
            }

            // ist_in_der_stunde
            for (int s = 0; s < ma.stunden; s++) {
                LinkedList<Integer> listOR = new LinkedList<>();
                for (int i = 0; i < klausurlaenge; i++) {
                    int var = (s - i >= 0) ? startet_in_der_stunde[s - i] : c_FALSE;
                    listOR.addLast(var);
                }
                int root = c_OR_tree(listOR);
                c_equal(root, ist_in_der_stunde[s]);
            }

            // wts = wt & ist_in_der_stunde
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    for (int s = 0; s < ma.stunden; s++) {
                        int x = wt[w][t];
                        int y = ist_in_der_stunde[s];
                        int root = c_new_var_AND(x, y);
                        c_equal(wts[w][t][s], root);
                    }
                }
            }

            // Mündliche Prüfungen
            if ((muendliche_tage >= 1) && (muendliche_tage <= WOCHENTAGE_VISUELL.length)) {
                // Verbiete bestimmte Tage der Woche
                for (int w = 0; w < ma.wochen; w++) {
                    for (int t = 0; t < ma.tage; t++) {
                        // Zähle mögliche Tage in Folge
                        int count = 0;
                        for (int tDif = 0; tDif < muendliche_tage; tDif++) {
                            if (t + tDif < ma.tage) {
                                if (ma.planB[w][t + tDif] == true) {
                                    count++;
                                }
                            }
                        }
                        // verbiete 
                        if (count < muendliche_tage) {
                            c_1(-wt[w][t]);
                        }
                    }
                }
                // Schüler-Implikationen
                for (D_Schueler_SAT sc : parent.mapS_schriftlich.values()) {
                    for (int w = 0; w < ma.wochen; w++) {
                        for (int t = 0; t < ma.tage; t++) {
                            // Vorwärts verbieten
                            for (int t2 = t + 1; t2 < ma.tage; t2++) {
                                if (t2 - t < muendliche_tage) {
                                    c_2(-wt[w][t], -sc.wt[w][t2]);
                                }
                            }
                        }
                    }
                }
            }

        }

        private void regel_stunden_b_bis_c_gesperrt(int b, int c) throws IOException, UserException {
            if (ist_muendlich()) {
                return;
            }
            for (int s = b; s <= c; s++) {
                c_1(-ist_in_der_stunde[s]);
            }
        }

        private void regel_datum_b_stunden_c_bis_d_gesperrt(String b, int c, int d) throws ParseException, UserException, IOException {
            if (ist_muendlich()) {
                return;
            }
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    if (ma.planS[w][t].equals(b)) {
                        for (int s = c; s <= d; s++) {
                            c_1(-wts[w][t][s]);
                        }
                    }
                }
            }
        }

        private void regel_klausurbeginn_gemeinsam(D_Klausur_SAT that, boolean auch_muendlich) throws IOException, UserException {
            if (this.ist_muendlich() || that.ist_muendlich()) {
                if (!auch_muendlich) {
                    return;
                }
            }
            // Gemeinsamer Tag
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    c_equal(this.wt[w][t], that.wt[w][t]);
                }
            }
            // Gemeinsamer Stundenstart
            for (int s = 0; s < ma.stunden; s++) {
                c_equal(this.startet_in_der_stunde[s], that.startet_in_der_stunde[s]);
            }
        }

        private void regel_klausurbeginn_gemeinsam_wenn_gleicher_tag(D_Klausur_SAT that) throws IOException, UserException {
            if (this.ist_muendlich() || that.ist_muendlich()) {
                return;
            }
            // Gleichen Tag bestimmen
            LinkedList<Integer> listOR1 = new LinkedList<>();
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    int var1 = this.wt[w][t];
                    int var2 = that.wt[w][t];
                    listOR1.addLast(c_new_var_AND(var1, var2));
                }
            }
            int var_gleicher_tag = c_new_var();
            c_equal(var_gleicher_tag, c_OR_tree(listOR1));

            // var_gleicher_beginn
            LinkedList<Integer> listOR2 = new LinkedList<>();
            for (int s = 0; s < ma.stunden; s++) {
                int var1 = this.startet_in_der_stunde[s];
                int var2 = that.startet_in_der_stunde[s];
                listOR2.addLast(c_new_var_AND(var1, var2));
            }
            int var_gleicher_beginn = c_new_var();
            c_equal(var_gleicher_beginn, c_OR_tree(listOR2));

            // var_gleicher_tag --> var_gleicher_beginn
            c_2(-var_gleicher_tag, var_gleicher_beginn);
        }

        private void regel_setze_auf_datum_d_in_stunde_e(String d, int e) throws IOException, UserException {
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    if (ma.planS[w][t].equals(d)) {
                        c_1(wt[w][t]);
                        c_1(startet_in_der_stunde[e]);
                    }
                }
            }
        }

        private void regel_setze_zwischen_datum_d_und_e(String d, String e) throws ParseException, UserException, IOException {
            for (int w = 0; w < ma.wochen; w++) {
                for (int t = 0; t < ma.tage; t++) {
                    if (d.equals(WILDCARD) == false) {
                        if (datum_compare(ma.planS[w][t], d) < 0) {
                            c_1(-wt[w][t]);
                        }
                    }
                    if (e.equals(WILDCARD) == false) {
                        if (datum_compare(ma.planS[w][t], e) > 0) {
                            c_1(-wt[w][t]);
                        }
                    }
                }
            }
        }

        private void regel_klausurbeginn_verbieten_in_stunde(int b) throws IOException, UserException {
            if (ist_muendlich()) {
                return;
            }
            c_1(-startet_in_der_stunde[b]);
        }

        private void regel_klausurbeginn_verbieten_am_wochentag_c_von_stunde_d_bis_e(String c, int d, int e) throws IOException, UserException {
            if (ist_muendlich()) {
                return;
            }
            for (int t = 0; t < WOCHENTAGE_VISUELL.length; t++) {
                if (WOCHENTAGE_VISUELL[t].equals(c)) {
                    for (int w = 0; w < ma.wochen; w++) {
                        for (int s = d; s <= e; s++) {
                            c_1(-wts[w][t][s]);
                        }
                    }
                }
            }
        }

        private void regel_setze_muendlich_mit_d_tagen(int d) {
            this.muendliche_tage = d;
        }

        private int get_klausurstunden_insgesamt() {
            if (ist_muendlich()) {
                return 0;
            } else {
                return klausurlaenge;
            }
        }

    }

// ########## HELPER ##########
    private LinkedList<Integer> toList(int[] a) {
        LinkedList<Integer> list = new LinkedList<>();
        for (int w = 0; w < a.length; w++) {
            list.addLast(a[w]);
        }
        return list;
    }

    private LinkedList<Integer> toList(int[][] wt) {
        LinkedList<Integer> list = new LinkedList<>();
        for (int w = 0; w < wt.length; w++) {
            for (int t = 0; t < wt[w].length; t++) {
                list.addLast(wt[w][t]);
            }
        }
        return list;
    }

    private LinkedList<Integer> toList(int[][][] wts) {
        LinkedList<Integer> list = new LinkedList<>();
        for (int w = 0; w < wts.length; w++) {
            for (int t = 0; t < wts[w].length; t++) {
                for (int s = 0; s < wts[w][t].length; s++) {
                    list.addLast(wts[w][t][s]);
                }
            }
        }
        return list;
    }

    // ########## CNF ##########
    void c_reset() throws IOException, UserException {
        listRND.clear();
        c_vars = 0;
        c_clauses = 0;
        c_TRUE = c_new_var();
        c_1(c_TRUE);
        c_FALSE = -c_TRUE;
    }

    int c_new_var() {
        if (listRND.isEmpty()) {
            for (int i = 0; i < 10000; i++) {
                c_vars++;
                if (rnd.nextBoolean()) {
                    listRND.addLast(+c_vars);
                } else {
                    listRND.addLast(-c_vars);
                }
            }
            Collections.shuffle(listRND, rnd);
        }
        return listRND.pollFirst();
    }

    int[] c_new_var_array(int size) {
        int[] a = new int[size];
        for (int i = 0; i < a.length; i++) {
            a[i] = c_new_var();
        }
        return a;
    }

    int[][] c_new_var_matrix(int rows, int cols) {
        int[][] m = new int[rows][cols];
        for (int r = 0; r < rows; r++) {
            for (int c = 0; c < cols; c++) {
                m[r][c] = c_new_var();
            }
        }
        return m;
    }

    int[][][] c_new_var_matrix_3d2(int d1, int d2, int d3) {
        int[][][] m = new int[d1][d2][d3];
        for (int i1 = 0; i1 < d1; i1++) {
            for (int i2 = 0; i2 < d2; i2++) {
                for (int i3 = 0; i3 < d3; i3++) {
                    m[i1][i2][i3] = c_new_var();
                }
            }
        }
        return m;
    }

    void write_bytes(String line, boolean send) throws IOException {
        if (line != null) {
            baos.write(line.getBytes());
        }
        if ((baos.size() > 1024 * 1024 * 32) || send) {
            process.getOutputStream().write(baos.toByteArray());
            baos.reset();
        }
    }

    void c_1(int x) throws IOException, UserException {
        if (x == 0) {
            throw new UserException("Klausel c_1 enthält den ungültigen Wert x=" + x + "");
        }
        c_clauses++;
        if (!c_simulate) {
            write_bytes(x + " 0" + "\n", false);
        }
    }

    void c_2(int x, int y) throws IOException, UserException {
        if (x == 0) {
            throw new UserException("Klausel c_2 enthält den ungültigen Wert x=" + x + "");
        }
        if (y == 0) {
            throw new UserException("Klausel c_2 enthält den ungültigen Wert y=" + y + "");
        }
        c_clauses++;
        if (!c_simulate) {
            write_bytes(x + " " + y + " 0" + "\n", false);
        }
    }

    void c_3(int x, int y, int z) throws IOException, UserException {
        if (x == 0) {
            throw new UserException("Klausel c_3 enthält den ungültigen Wert x=" + x + "");
        }
        if (y == 0) {
            throw new UserException("Klausel c_3 enthält den ungültigen Wert y=" + y + "");
        }
        if (z == 0) {
            throw new UserException("Klausel c_3 enthält den ungültigen Wert z=" + z + "");
        }
        c_clauses++;
        if (!c_simulate) {
            write_bytes(x + " " + y + " " + z + " 0" + "\n", false);
        }
    }

    int c_at_most_one_tree(LinkedList<Integer> list) throws IOException, UserException {
        while (list.size() >= 2) {
            int a = list.pollFirst();
            int b = list.pollFirst();
            c_2(-a, -b); // not both
            int c = c_new_var_OR(a, b);
            list.addLast(c);
        }
        if (list.isEmpty()) {
            return c_FALSE;
        }
        return list.pollFirst();
    }

    int c_new_var_OR(int x, int y) throws IOException, UserException {
        int z = c_new_var();
        c_2(-x, z);
        c_2(-y, z);
        c_3(x, y, -z);
        return z;
    }

    void c_equal(int x, int y) throws IOException, UserException {
        c_2(-x, +y);
        c_2(+x, -y);
    }

    int c_exactly_one_tree(LinkedList<Integer> list) throws IOException, UserException {
        int var = c_at_most_one_tree(list);
        c_1(var);
        return var;
    }

    int c_OR_tree(LinkedList<Integer> list) throws IOException, UserException {
        while (list.size() >= 2) {
            int a = list.pollFirst();
            int b = list.pollFirst();
            int c = c_new_var_OR(a, b);
            list.addLast(c);
        }
        if (list.isEmpty()) {
            return c_FALSE;
        }
        return list.pollFirst();
    }

    int c_new_var_AND(int x, int y) throws IOException, UserException {
        int c = c_new_var();
        c_2(x, -c);
        c_2(y, -c);
        c_3(-x, -y, c);
        return c;
    }

    boolean c_get(int var
    ) {
        if (var > 0) {
            return state[var];
        }
        return !state[-var];
    }

    void c_exactly_GENERIC(LinkedList<Integer> list, int amount) throws IOException, UserException {
        list = new LinkedList<>(list);
        // special case: 0
        if (amount == 0) {
            for (int x : list) {
                c_1(-x);
            }
            return;
        }
        // special case: N
        if (amount == list.size()) {
            for (int x : list) {
                c_1(+x);
            }
            return;
        }
        // special case: 1
        if (amount == 1) {
            int var = c_exactly_one_tree(list);
            c_1(+var);
            return;
        }
        // else choose...
        c_exactly_NETWORK(list, amount);
    }

    void c_at_least_GENERIC(LinkedList<Integer> list, int minimum) throws IOException, UserException {
        list = new LinkedList<>(list);
        // special case: trivial
        if (minimum <= 0) {
            return;
        }
        // special case: all true
        if (minimum == list.size()) {
            for (int x : list) {
                c_1(+x);
            }
            return;
        }
        // special case: at_least_one
        if (minimum == 1) {
            int var = c_OR_tree(list);
            c_1(+var);
            return;
        }
        // else choose...
        c_at_least_NETWORK(list, minimum);

    }

    void c_at_most_GENERIC(LinkedList<Integer> list, int maximum) throws IOException, UserException {
        list = new LinkedList<>(list);
        // special case: trivial
        if (maximum >= list.size()) {
            return;
        }
        // special case: all false
        if (maximum == 0) {
            for (int x : list) {
                c_1(-x);
            }
            return;
        }
        // special case: at_most_one
        if (maximum == 1) {
            c_at_most_one_tree(list);
            return;
        }
        // else choose...
        c_at_most_NETWORK(list, maximum);

    }

    private void c_at_most_NETWORK(LinkedList<Integer> list, int maximum) throws IOException, UserException {
        // sort
        c_bitonic_sort(list);
        // force FALSE
        for (int i = maximum; i < list.size(); i++) {
            c_1(-list.get(i));
        }
    }

    private void c_at_least_NETWORK(LinkedList<Integer> list, int minimum) throws IOException, UserException {
        // sort
        c_bitonic_sort(list);
        // force FALSE
        for (int i = 0; i < minimum; i++) {
            c_1(+list.get(i));
        }
    }

    private void c_exactly_NETWORK(LinkedList<Integer> list, int amount) throws IOException, UserException {
        // sort
        c_bitonic_sort(list);
        // force TRUE / FALSE
        for (int i = 0; i < list.size(); i++) {
            if (i < amount) {
                c_1(+list.get(i));
            } else {
                c_1(-list.get(i));
            }
        }
    }

    void c_bitonic_sort(LinkedList<Integer> list) throws IOException, UserException {
        c_fill_False_until_power_two(list);
        c_bitonic_sort_power_two(list);
    }

    private void c_fill_False_until_power_two(LinkedList<Integer> list) throws IOException, UserException {
        int size = 1;
        while (size < list.size()) {
            size *= 2;
        }
        while (list.size() < size) {
            list.addLast(c_FALSE);
        }
    }

    private void c_bitonic_sort_power_two(LinkedList<Integer> list) throws IOException, UserException {
        for (int window = 2; window <= list.size(); window *= 2) {
            c_bitonic_sort_spiral(list, window);
            for (int difference = window / 2; difference >= 2; difference /= 2) {
                c_bitonic_sort_difference(list, difference);
            }
        }
    }

    private void c_bitonic_sort_spiral(LinkedList<Integer> list, int size) throws IOException, UserException {
        for (int i = 0; i < list.size(); i += size) {
            for (int i1 = i, i2 = i + size - 1; i1 < i2; i1++, i2--) {
                c_bitonic_comparator(list, i1, i2);
            }
        }
    }

    private void c_bitonic_sort_difference(LinkedList<Integer> list, int size) throws IOException, UserException {
        int half = size / 2;
        for (int i = 0; i < list.size(); i += size) {
            for (int j = 0; j < half; j++) {
                c_bitonic_comparator(list, i + j, i + j + half);
            }
        }
    }

    private void c_bitonic_comparator(LinkedList<Integer> result, int i1, int i2) throws IOException, UserException {
        if (i1 >= i2) {
            throw new UserException("D_Ausgabe: c_bitonic_comparator: " + i1 + "," + i2 + " <-- ERROR!!!");
        }
        int a = result.get(i1);
        int b = result.get(i2);
        result.set(i1, c_new_var_OR(a, b));
        result.set(i2, c_new_var_AND(a, b));
    }

}
