package bar_multi_klausurplaner_cnf;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.TreeSet;
import javax.swing.DefaultListModel;
import static bar_multi_klausurplaner_cnf.Static.*;
import javax.swing.SwingUtilities;

/**
 * @author Benjamin A. Bartsch
 */
public class Data {

    private D_Map mapE = new D_Map(); // Eingabedaten
    private D_Map mapA = new D_Map(); // Ausgabedaten
    private final GUI gui;
    private final SimpleDateFormat sdf = new SimpleDateFormat("dd.MM.yyyy");
    private final Calendar cal1 = new GregorianCalendar();
    private final Calendar cal2 = new GregorianCalendar();
    private final DefaultListModel<D_Berechnung> model_berechnung = new DefaultListModel<>();
    private final DefaultListModel<D_Ausgabe> ausgabe_model = new DefaultListModel<>();
    private int berechnung_laufend = 0;

    public Data(GUI gui) {
        this.gui = gui;
    }

    // PROTECTED
    void misc_import_kurs42_ordner(File[] files) throws IOException, UserException {
        misc_reset();

        for (File file : files) {
            misc_import_kurs42_ordner(file);
        }

    }

    void misc_open(ObjectInputStream ois) throws IOException, ClassNotFoundException, ClassCastException {
        co_berechnung_clear();
        co_ausgabe_clear();
        mapE = (D_Map) ois.readObject();
        mapA = (D_Map) ois.readObject();
        for (int id : mapA.keySetInteger()) {
            D_Ausgabe a = new D_Ausgabe(id);
            a.load_from_map(mapA);
            co_ausgabe_add(a);
        }
        misc_revalidate_data();
    }

    void misc_save(ObjectOutputStream oos) throws IOException {
        oos.writeObject(mapE);
        oos.writeObject(mapA);
    }

    void co_set_max_klausuren_pro_schueler_pro_woche(int value) {
        mapE.get(ID_MAX_KLAUSUREN_PRO_WOCHE).clear();
        mapE.get(ID_MAX_KLAUSUREN_PRO_WOCHE).get(value);
    }

    int co_get_max_klausuren_pro_schueler_pro_woche() {
        if (mapE.has(ID_MAX_KLAUSUREN_PRO_WOCHE) == false) {
            mapE.get(ID_MAX_KLAUSUREN_PRO_WOCHE).get(3); // Default
        }
        return mapE.get(ID_MAX_KLAUSUREN_PRO_WOCHE).first_key_int();
    }

    boolean co_set_datum(String sDatum_ab, String sDatum_bi) {
        try {
            Date date_ab = sdf.parse(sDatum_ab);
            Date date_bi = sdf.parse(sDatum_bi);
            cal1.setTime(date_ab);
            cal2.setTime(date_bi);
            int jahr_ab = cal1.get(Calendar.YEAR);
            int jahr_bi = cal2.get(Calendar.YEAR);
            if ((jahr_ab + 2 <= jahr_bi) || (jahr_ab > jahr_bi)) {
                return false;
            }
            int monat_ab = cal1.get(Calendar.MONTH) + 1; // 0...11
            int monat_bi = cal2.get(Calendar.MONTH) + 1; // 0...11
            if ((jahr_ab == jahr_bi) && (monat_ab > monat_bi)) {
                return false;
            }
            int tag_ab = cal1.get(Calendar.DAY_OF_MONTH);
            int tag_bi = cal2.get(Calendar.DAY_OF_MONTH);
            if ((jahr_ab == jahr_bi) && (monat_ab == monat_bi) && (tag_ab >= tag_bi)) {
                return false;
            }

            String sParsed_ab = sdf.format(date_ab);
            String sParsed_bi = sdf.format(date_bi);
            co_set_wochenplan_ab(sParsed_ab);
            co_set_wochenplan_bi(sParsed_bi);

            co_datum_update_plan(date_ab, date_bi);
            return true;
        } catch (ParseException | UserException ex) {
            ex.printStackTrace();
            return false;
        }
    }

    String co_get_wochenplan_ab() {
        if (mapE.has(ID_WOCHENPLAN_DATUM_AB) == false) {
            Calendar cal = Calendar.getInstance();
            // cal.add(Calendar.MONTH, 0);
            String sDatum = sdf.format(cal.getTime());
            mapE.get(ID_WOCHENPLAN_DATUM_AB).get(sDatum);
        }
        return mapE.get(ID_WOCHENPLAN_DATUM_AB).first_key_string();
    }

    String co_get_wochenplan_bi() {
        if (mapE.has(ID_WOCHENPLAN_DATUM_BI) == false) {
            Calendar cal = Calendar.getInstance();
            cal.add(Calendar.MONTH, 1);
            String sDatum = sdf.format(cal.getTime());
            mapE.get(ID_WOCHENPLAN_DATUM_BI).get(sDatum);
        }
        return mapE.get(ID_WOCHENPLAN_DATUM_BI).first_key_string();
    }

    int co_get_wochenplan_nr(String sDatum) {

        return 0;
    }

    LinkedList<String> co_get_stufen(boolean mitStern) {
        LinkedList<String> list = co_get_liste(mitStern);
        list.addAll(mapE.get(ID_STUFEN).keySetSortedString());
        return list;
    }

    LinkedList<String> co_get_kurse(boolean mitStern) {
        LinkedList<String> list = co_get_liste(mitStern);
        for (String sStufe : co_get_stufen(false)) {
            list.addAll(mapE.get(ID_STUFEN).get(sStufe).get(ID_KURSE).keySetSortedString());
        }
        return list;
    }

    LinkedList<String> co_get_klausurlaengen() {
        int max = co_get_letzte_stunde();
        return co_get_zahlenliste(1, max, false);
    }

    LinkedList<String> co_get_klausurnummern(boolean mitStern) {
        LinkedList<String> list = co_get_liste(mitStern);
        list.addAll(co_get_zahlenliste(1, 3, mitStern));
        return list;
    }

    LinkedList<String> co_get_zahlenliste(int min, int max, boolean mitStern) {
        LinkedList<String> list = co_get_liste(mitStern);
        for (int laenge = min; laenge <= max; laenge++) {
            list.addLast(laenge + "");
        }
        return list;
    }

    LinkedList<String> co_get_faecher(boolean mitStern) {
        LinkedList<String> list = co_get_liste(mitStern);
        TreeSet<String> set = new TreeSet<>();
        for (String sStufe : co_get_stufen(false)) {
            for (String sKursname : mapE.get(ID_STUFEN).get(sStufe).get(ID_KURSE).keySetSortedString()) {
                set.add(mapE.get(ID_STUFEN).get(sStufe).get(ID_KURSE).get(sKursname).get(ID_KURS_FACH).first_key_string());
            }
        }
        list.addAll(set);
        return list;
    }

    LinkedList<String> co_get_kursarten(boolean mitStern) {
        LinkedList<String> list = co_get_liste(mitStern);
        TreeSet<String> set = new TreeSet<>();
        for (String sStufe : co_get_stufen(false)) {
            for (String sKursname : mapE.get(ID_STUFEN).get(sStufe).get(ID_KURSE).keySetSortedString()) {
                set.add(mapE.get(ID_STUFEN).get(sStufe).get(ID_KURSE).get(sKursname).get(ID_KURS_KURSART).first_key_string());
            }
        }
        list.addAll(set);
        return list;
    }

    LinkedList<String> co_get_schienen(boolean mitStern) {
        TreeSet<Integer> set = new TreeSet<>();
        for (String sStufe : co_get_stufen(false)) {
            for (String sKursname : mapE.get(ID_STUFEN).get(sStufe).get(ID_KURSE).keySetSortedString()) {
                set.addAll(mapE.get(ID_STUFEN).get(sStufe).get(ID_KURSE).get(sKursname).get(ID_KURS_SCHIENEN).keySetSortedInteger());
            }
        }
        LinkedList<String> list = co_get_liste(mitStern);
        for (int schiene : set) {
            list.addLast("" + (schiene + 1));
        }
        return list;
    }

    LinkedList<String> co_get_datums(boolean mitStern) {
        LinkedList<String> list = co_get_liste(mitStern);
        String[][] plan_s = co_get_plan_s();
        for (int w = 0; w < plan_s.length; w++) {
            for (int t = 0; t < plan_s[w].length; t++) {
                list.addLast(plan_s[w][t]);
            }
        }
        return list;
    }

    LinkedList<String> co_get_stunden(boolean mitStern) {
        LinkedList<String> list = co_get_liste(mitStern);
        int max = co_get_letzte_stunde();
        for (int i = 1; i <= max; i++) {
            list.addLast("" + i);
        }
        return list;
    }

    LinkedList<String> co_get_schueler(boolean mitStern) {
        LinkedList<String> list = co_get_liste(mitStern);
        for (String sStufen : co_get_stufen(false)) {
            list.addAll(mapE.get(ID_STUFEN).get(sStufen).get(ID_SCHUELER).keySetSortedString());
        }
        return list;
    }

    LinkedList<String> co_get_wochentage(boolean mitStern) {
        LinkedList<String> list = co_get_liste(mitStern);
        for (String sWochentage : WOCHENTAGE_VISUELL) {
            list.addLast(sWochentage);
        }
        return list;
    }

    LinkedList<String> co_get_liste(boolean mitStern) {
        LinkedList<String> list = new LinkedList<>();
        if (mitStern) {
            list.addLast("*");
        }
        return list;
    }

    void co_set_regeln(String text) {
        mapE.get(ID_REGELN).clear();
        mapE.get(ID_REGELN).get(text);
    }

    String co_get_regeln() {
        if (mapE.get(ID_REGELN).size() == 0) {
            mapE.get(ID_REGELN).get("");
        }
        return mapE.get(ID_REGELN).first_key_string();
    }

    D_Berechnung co_berechnung_create() {
        int id = co_get_and_inc_berechnungs_nr();
        D_Berechnung be = new D_Berechnung(id, this);
        model_berechnung.addElement(be);
        return be;
    }

    // Aufgerufen von Worker-Thread aus der run()-Methode, 
    // nicht vom AWT-Thread!
    synchronized void co_berechnung_start() {
        // Warteschleife ++
        try {
            while (berechnung_laufend >= co_berechnung_get_max_parallel()) {
                wait();
            }
            berechnung_laufend++;
        } catch (InterruptedException ex) {
            gui.handle_exception(ex);
        }
        
        // Anzeige aktualisieren
        SwingUtilities.invokeLater(new Runnable() {
            @Override
            public void run() {
                gui.set_aktiv(berechnung_laufend);
            }
        });        
    }

    synchronized void co_berechnung_done(int id) {
        // Warteschleife --
        notify();
        berechnung_laufend--;
        
        // Anzeige aktualisieren
        gui.set_aktiv(berechnung_laufend);
        
        // Ausgabe erzeugen
        for (int index = 0; index < model_berechnung.size(); index++) {
            if (model_berechnung.get(index).get_id() == id) {
                D_Berechnung be = model_berechnung.remove(index);
                if (be.get_user_aborted() == false) {
                    co_ausgabe_add(be.get_ergebnis());
                }
                return;
            }
        }
        
        // Aufräumen
        System.gc();
    }

    synchronized void co_berechnung_set_max_parallel(int value) {
        mapE.get(ID_MAX_PARALLEL).clear();
        mapE.get(ID_MAX_PARALLEL).get(value);
        notifyAll();
    }

    synchronized int co_berechnung_get_max_parallel() {
        if (mapE.has(ID_MAX_PARALLEL) == false) {
            mapE.get(ID_MAX_PARALLEL).get(3); // Default
        }
        return mapE.get(ID_MAX_PARALLEL).first_key_int();
    }

    void co_ausgabe_add(D_Ausgabe a) {
        a.save_to_map(mapA);
        ausgabe_model.addElement(a);
    }

    void co_ausgabe_del(D_Ausgabe a) {
        mapA.del(a.get_id());
        ausgabe_model.removeElement(a);
    }

    DefaultListModel<D_Berechnung> co_berechnung_model() {
        return model_berechnung;
    }

    DefaultListModel<D_Ausgabe> co_ergebnis_model() {
        return ausgabe_model;
    }

    D_Map co_map_get_copy() {
        return mapE.copy();
    }

    void co_berechnung_kill(D_Berechnung be) {
        be.set_user_abort();
    }

    int co_get_echte_klausurtage() {
        boolean[][] planB = co_plan_b_get();
        int s = 0;
        for (int r = 0; r < planB.length; r++) {
            for (int c = 0; c < planB[r].length; c++) {
                if (planB[r][c]) {
                    s++;
                }
            }
        }
        return s;
    }

    void co_check() throws UserException {
        if (mapE.get(ID_STUFEN).size() == 0) {
            throw new UserException("Es wurde noch keine Jahrgangsstufe geladen!");
        }
    }

    void handle_exception(Exception ex) {
        gui.handle_exception(ex);
    }

    void co_berechnung_clear() {
        for (int index = 0; index < model_berechnung.size(); index++) {
            model_berechnung.get(index).set_user_abort();
        }
        model_berechnung.removeAllElements();
    }

    int co_get_wochenanzahl() {
        return co_get_plan_s().length;
    }

    int co_get_schueleranzahl() {
        int summe = 0;
        for (String sStufen : co_get_stufen(false)) {
            summe += mapE.get(ID_STUFEN).get(sStufen).get(ID_SCHUELER).size();
        }
        return summe;
    }

    // PRIVATE
    private void misc_reset() {

        mapE.clear();
        co_set_regeln("100#Klausurphasen: Stufe <*> und Kurs <*> schreibt <1> Klausuren."
                + "\n" + "150#Klausurlänge: Stufe <*> und Kurs <*> mit Klausurnummer <*> hat Länge <2>."
                + "\n");

        co_berechnung_clear();
        co_ausgabe_clear();
        misc_revalidate_data();
    }

    private void misc_revalidate_data() {
        /// nothing yet
    }

    private void misc_import_kurs42_ordner(File kurs42_ordner) throws IOException, UserException {
        gui.debug("\nÖffne Ordner: " + kurs42_ordner);

        // ﻿"IdNr";"BlockungsId";             "Bezeichnung";"Jahrgang"; "Semester";"Schienenzahl";"Erstexport";"KoppIdMaske"
        //      1;            1;"Jahrgang: EF (2019/20.1)";      "EF";      20191;            12;           0;   "BL%s%.2x"        
        File file_blockgruppen = new File(kurs42_ordner, "Blockgruppen.txt");
        String[][] table_blockgruppen = read_kurs42_file_as_table(file_blockgruppen);
        if (table_blockgruppen.length < 1) {
            throw new UserException("Die Datei '" + file_blockgruppen.getAbsolutePath() + "' ist leer!");
        }
        for (String[] line : table_blockgruppen) {
            gui.debug(Arrays.toString(line));
        }

        // Daten sind Zeilentupel
        File file_blockung = new File(kurs42_ordner, "Blockung.txt");
        String[][] table_blockung = read_kurs42_file_as_table(file_blockung);
        if (table_blockung.length < 1) {
            throw new UserException("Die Datei 'Blockung.txt' ist leer!");
        }

        // ﻿"Schiene";"Kursbezeichnung";"Fixiert"
        //         0;          "D-GK1";        1
        File file_blockplan = new File(kurs42_ordner, "Blockplan.txt");
        String[][] table_blockplan = read_kurs42_file_as_table(file_blockplan);
        if (table_blockplan.length < 1) {
            throw new UserException("Die Datei 'Blockplan.txt' ist leer!");
        }

        // ﻿"KursId";"Fach";    "Name";"Lehrer";"Std";"Raum";"Kursart";"Schulnummer"
        //       -1; "VKM";"VKM-VTF1";    "LL";    2;  "--";    "VTF";       123456
        File file_kurse = new File(kurs42_ordner, "Kurse.txt");
        String[][] table_kurse = read_kurs42_file_as_table(file_kurse);
        if (table_kurse.length < 1) {
            throw new UserException("Die Datei 'Kurse.txt' ist leer!");
        }

        // "IdNr";"Status";"Geschlecht";"Jahrgang";"Klasse";  "Name";"Vorname";"CollCount";    "GebDat";"SchulNr";"DB_IdNr";"Tutor";        "PruefOrd"; "EMail";"KoopDBIdNr"
        //      1;       2;           4;      "EF";     "B";"AAAAAA"; "BBBBBB";         -1;"01.01.2001";   123456;     3760;   "AA";"APO-GOSt(B)10/G8"; "a@b.c";           0
        File file_schueler = new File(kurs42_ordner, "Schueler.txt");
        String[][] table_schueler = read_kurs42_file_as_table(file_schueler);
        if (table_schueler.length < 1) {
            throw new UserException("Die Datei 'Schueler.txt' ist leer!");
        }

        //   ﻿"Name";"Vorname";    "GebDat";"Geschlecht";"Fachkrz";"Kursart";"Note";"Kl1";"Kl2";"KLG";"SoMi1";"SoMi2";"SomiG";"Fehl1";"Fehl2";"uFehl1";"uFehl2";"FehlStd";"UFehlStd"
        // "AAAAAA"; "BBBBBB";"01.01.2001";           4;      "D";    "GKS";    "";   17;   17;   17;     17;     17;     17;       0;     0;       0;       0;        0;         0
        File file_fachwahlen = new File(kurs42_ordner, "Fachwahlen.txt");
        String[][] table_fachwahlen = read_kurs42_file_as_table(file_fachwahlen);
        if (table_fachwahlen.length < 1) {
            throw new UserException("Die Datei 'Fachwahlen.txt' ist leer!");
        }

        //   "Name";"Vorname";    "GebDat";"Geschlecht";"Kursbezeichnung"
        // "AAAAAA"; "BBBBBB";"01.01.2001";           4;          "D-GK2"
        File file_kursbelegung = new File(kurs42_ordner, "Kursbelegung.txt");
        String[][] table_kursbelegung = read_kurs42_file_as_table(file_kursbelegung);
        if (table_kursbelegung.length < 1) {
            throw new UserException("Die Datei 'Kursbelegung.txt' ist leer!");
        }

        //     ﻿"Kurs";"Tag";"Std";"Woche";"Raum";"Lehrer"
        // "VKM-VTF1";    3;    1;      0;  "--";    "LL"
        File file_kursplaene = new File(kurs42_ordner, "Kursplaene.txt");
        String[][] table_kursplaene = read_kurs42_file_as_table(file_kursplaene);
        if (table_kursplaene.length < 1) {
            throw new UserException("Die Datei 'Kursplaene.txt' ist leer!");
        }

        // ##### Aufgaben von map #####
        // Blockgruppen.txt
        String sJahrgang = table_blockgruppen[0][3];
        if (mapE.get(ID_STUFEN).has(sJahrgang) == true) {
            throw new UserException("Eine Blockungsgruppe mit dem Jahrgang '" + sJahrgang + "' existiert doppelt!");
        }
        mapE.get(ID_STUFEN).get(sJahrgang);
        mapE.get(ID_STUFEN).get(sJahrgang).get(ID_BLOCKGRUPPE).get(table_blockgruppen[0][2]);

        // Blockung.txt
        HashMap<String, String> mapBlockung = new HashMap<>();
        for (int r = 0; r < table_blockung.length; r++) {
            String[] split = table_blockung[r][0].split("=");
            if (split.length == 2) {
                mapBlockung.put(split[0], split[1]);
            }
        }

        String sJahrgang2 = mapBlockung.get("Jahrgang");
        if (sJahrgang2.equals(sJahrgang) == false) {
            throw new UserException("Jahrgang aus Blockplan.txt (" + sJahrgang + ") ungleich Blockung.txt (" + sJahrgang2 + ")");
        }
        mapE.get(ID_STUFEN).get(sJahrgang).get(ID_JAHR).get(mapBlockung.get("Jahr"));
        gui.debug("  Jahr = " + mapBlockung.get("Jahr"));
        mapE.get(ID_STUFEN).get(sJahrgang).get(ID_ABSCHNITT).get(mapBlockung.get("Abschnitt"));
        gui.debug("  Abschnitt = " + mapBlockung.get("Abschnitt"));
        mapE.get(ID_STUFEN).get(sJahrgang).get(ID_BLOCKUNG).get(mapBlockung.get("Bezeichnung"));
        gui.debug("  Bezeichnung = " + mapBlockung.get("Bezeichnung"));

        // Globale werte
        for (int std = 0; std <= 20; std++) {
            String stunde_neu = mapBlockung.get("Stunde_" + std);
            if (stunde_neu == null) {
                stunde_neu = "";
            }
            if (mapE.has(ID_STUNDE + "_" + std)) {
                String stunde_alt = mapE.get(ID_STUNDE + "_" + std).first_key_string();
                if (stunde_neu.equals(stunde_alt) == false) {
                    throw new UserException("Die Jahrgänge haben verschiedene Stundendefinitionen: Stunde_" + std + ": " + stunde_neu + " vs. " + stunde_alt);
                }
            } else {
                mapE.get(ID_STUNDE + "_" + std).get(stunde_neu);
                gui.debug("  Stunde_" + std + " = " + stunde_neu);
            }
        }

        String sStundendauer_neu = mapBlockung.get("StundenDauer");
        int iStundendauer_neu = Integer.parseInt(sStundendauer_neu);
        if (mapE.has(ID_STUNDENDAUER)) {
            int iStundendauer_alt = mapE.get(ID_STUNDENDAUER).first_key_int();
            if (iStundendauer_neu != iStundendauer_alt) {
                throw new UserException("Die Stundendauern haben verschiedene Stundendefinitionen: " + iStundendauer_neu + " vs. " + iStundendauer_alt);
            }
        } else {
            mapE.get(ID_STUNDENDAUER).get(iStundendauer_neu);
            gui.debug("  Stundendauer = " + iStundendauer_neu);
        }

        gui.debug("\n  Erzeuge Kurse & Fächer");
        for (int r = 0; r < table_kurse.length; r++) {
            int kuID = Integer.parseInt(table_kurse[r][0]);
            String sFach = table_kurse[r][1];
            String sName = table_kurse[r][2];
            String sLehrer = table_kurse[r][3];
            int iStd = Integer.parseInt(table_kurse[r][4]);
            String sRaum = table_kurse[r][5];
            String sKursart = table_kurse[r][6];

            if (mapE.get(ID_STUFEN).get(sJahrgang).get(ID_FAECHER).has(sFach) == false) {
                mapE.get(ID_STUFEN).get(sJahrgang).get(ID_FAECHER).get(sFach);
                gui.debug("  Fach: " + sFach);
            }

            if (mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).has(sName) == true) {
                throw new UserException("Der Kurs '" + sName + "' im Jahrgang '" + sJahrgang + "' existiert doppelt!");
            }

            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).get(sName);
            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).get(sName).get(ID_KURS_NR).get(kuID);
            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).get(sName).get(ID_KURS_FACH).get(sFach);
            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).get(sName).get(ID_KURS_LEHRER).get(sLehrer);
            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).get(sName).get(ID_KURS_STD).get(iStd);
            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).get(sName).get(ID_KURS_RAUM).get(sRaum);
            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).get(sName).get(ID_KURS_KURSART).get(sKursart);
            gui.debug("  Kurs: " + kuID + ", " + sName + ", " + sFach + ", " + sLehrer + ", " + sKursart);
        }

        gui.debug("\n  Lage der Kurse (Schienen)");
        for (int r = 0; r < table_blockplan.length; r++) {
            int schiene = Integer.parseInt(table_blockplan[r][0]);
            String sKursname = table_blockplan[r][1];

            if (mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).has(sKursname) == false) {
                throw new UserException("Im Jahrgang '" + sJahrgang + "' wir im Blockplan der Kurs '" + sKursname + "' referenziert, der nicht existiert!");
            }
            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).get(sKursname).get(ID_KURS_SCHIENEN).get(schiene);
            gui.debug("  Kurs: " + sKursname + " in Schiene " + (schiene + 1));
        }
        for (String sKursname : mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).keySetSortedString()) {
            if (mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).get(sKursname).get(ID_KURS_SCHIENEN).size() == 0) {
                throw new UserException("Im Jahrgang '" + sJahrgang + "' hat der Kurs '" + sKursname + "' keine zugeordneten Schienen!");
            }
        }

        gui.debug("\n  Erzeuge Schüler");
        TreeSet<String> set_schueler = new TreeSet<>();
        for (int r = 0; r < table_schueler.length; r++) {
            int iID = Integer.parseInt(table_schueler[r][0]);
            String sName = table_schueler[r][5];
            String sVorname = table_schueler[r][6];
            String sGebDat = table_schueler[r][8];
            String sGeschlecht = table_schueler[r][2];
            String sDBid = table_schueler[r][10];
            String sKey = sName + ";" + sVorname + ";" + sGebDat + ";" + sGeschlecht;

            if (mapE.get(ID_STUFEN).get(sJahrgang).get(ID_SCHUELER).has(sKey) == true) {
                throw new UserException("Im Jahrgang '" + sJahrgang + "' existiert der Schüler '" + sKey + "' doppelt!");
            }
            if (set_schueler.add(sKey) == false) {
                throw new UserException("Der Schüler '" + sKey + "' existiert in mehreren Jahrgängen!");
            }
            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_SCHUELER).get(sKey).get(ID_SCHUELER_NR).get(iID);
            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_SCHUELER).get(sKey).get(ID_SCHUELER_DBID).get(sDBid);
            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_SCHUELER).get(sKey).get(ID_SCHUELER_NAME).get(sName);
            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_SCHUELER).get(sKey).get(ID_SCHUELER_VORNAME).get(sVorname);
            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_SCHUELER).get(sKey).get(ID_SCHUELER_GEBDATUM).get(sGebDat);
            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_SCHUELER).get(sKey).get(ID_SCHUELER_GESCHLECHT).get(sGeschlecht);
            gui.debug("  Schüler: " + sKey);
        }

        gui.debug("\n  Erzeuge Fachwahlen");
        for (int r = 0; r < table_fachwahlen.length; r++) {
            String sName = table_fachwahlen[r][0];
            String sVorname = table_fachwahlen[r][1];
            String sGebDat = table_fachwahlen[r][2];
            String sGeschlecht = table_fachwahlen[r][3];
            String sKey = sName + ";" + sVorname + ";" + sGebDat + ";" + sGeschlecht;

            String sFachkrz = table_fachwahlen[r][4];
            String sKursart = table_fachwahlen[r][5];

            if (mapE.get(ID_STUFEN).get(sJahrgang).get(ID_SCHUELER).has(sKey) == false) {
                throw new UserException("Im Jahrgang '" + sJahrgang + "' existiert zur Fachwahl '" + sFachkrz + ";" + sKursart + "' nicht der Schüler '" + sKey + "'!");
            }
            if (mapE.get(ID_STUFEN).get(sJahrgang).get(ID_SCHUELER).get(sKey).get(ID_SCHUELER_FACHWAHLEN).get(sFachkrz).has(sKursart) == true) {
                throw new UserException("Im Jahrgang '" + sJahrgang + " hat der Schüler '" + sKey + "' die Fachwahl '" + sFachkrz + ";" + sKursart + "' doppelt!");
            }

            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_SCHUELER).get(sKey).get(ID_SCHUELER_FACHWAHLEN).get(sFachkrz).get(sKursart);
            gui.debug("  Schüler: " + sKey + " mit " + sFachkrz + ";" + sKursart);
        }

        gui.debug("\n  Erzeuge Kurs <---> Schüler Verbindungen");
        for (int r = 0; r < table_kursbelegung.length; r++) {
            String sName = table_kursbelegung[r][0];
            String sVorname = table_kursbelegung[r][1];
            String sGebDat = table_kursbelegung[r][2];
            String sGeschlecht = table_kursbelegung[r][3];
            String sKursbezeichnung = table_kursbelegung[r][4];
            String sKey = sName + ";" + sVorname + ";" + sGebDat + ";" + sGeschlecht;

            if (mapE.get(ID_STUFEN).get(sJahrgang).get(ID_SCHUELER).has(sKey) == false) {
                throw new UserException("Im Jahrgang '" + sJahrgang + "' existiert zum Kurs '" + sKursbezeichnung + "' Schüler/in '" + sKey + " nicht'!");
            }
            if (mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).has(sKursbezeichnung) == false) {
                throw new UserException("Im Jahrgang '" + sJahrgang + "' existiert zu Schüler/in '" + sKey + "' der Kurs '" + sKursbezeichnung + " nicht'!");
            }
            String sFach = mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).get(sKursbezeichnung).get(ID_KURS_FACH).first_key_string();
            if (mapE.get(ID_STUFEN).get(sJahrgang).get(ID_SCHUELER).get(sKey).get(ID_SCHUELER_FACHWAHLEN).has(sFach) == false) {
                throw new UserException("Im Jahrgang '" + sJahrgang + "' ist '" + sKey + "' im Kurs '" + sKursbezeichnung + "' ohne in '" + sFach + "' eine Fachwahl zu haben!");
            }
            String sKursart = mapE.get(ID_STUFEN).get(sJahrgang).get(ID_SCHUELER).get(sKey).get(ID_SCHUELER_FACHWAHLEN).get(sFach).first_key_string();

            // Verbinde...
            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_SCHUELER).get(sKey).get(ID_SCHUELER_KURSE_ALLE).get(sKursbezeichnung);
            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).get(sKursbezeichnung).get(ID_KURS_SCHUELER_ALLE).get(sKey);
            if (misc_ist_klausurschreiber(sKursart)) {
                mapE.get(ID_STUFEN).get(sJahrgang).get(ID_SCHUELER).get(sKey).get(ID_SCHUELER_KURSE_SCHRIFTLICH).get(sKursbezeichnung);
                mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).get(sKursbezeichnung).get(ID_KURS_SCHUELER_SCHRIFTLICH).get(sKey);
                gui.debug("  Kurs: " + sKursbezeichnung + " <---> " + sKey + " (schriftlich)");
            } else {
                gui.debug("  Kurs: " + sKursbezeichnung + " <---> " + sKey);
            }
        }

        gui.debug("\n  Erzeuge Kurslagen im Stundenplan");
        for (int r = 0; r < table_kursplaene.length; r++) {
            /*
            "M-LK1";1;8;0;"--";"SR"
            "M-LK1";1;8;1;"--";"SR"
            "M-LK1";2;1;0;"--";"SR"
            "M-LK1";2;1;1;"--";"SR"
            "M-LK1";2;2;0;"--";"SR"
            "M-LK1";2;2;1;"--";"SR"
            "M-LK1";4;1;0;"--";"SR"
            "M-LK1";4;1;1;"--";"SR"
            "M-LK1";4;2;0;"--";"SR"
            "M-LK1";4;2;1;"--";"SR"
             */
            String sKursname = table_kursplaene[r][0];
            String sTag = table_kursplaene[r][1];
            String sStunde = table_kursplaene[r][2];
            String sWoche = table_kursplaene[r][3];
            String sRaum = table_kursplaene[r][4];
            String sLehrer = table_kursplaene[r][5];

            int kuTag = Integer.parseInt(sTag);
            int kuStunde = Integer.parseInt(sStunde);
            mapE.get(ID_MAX_WOCHEN_TAGE).get(kuTag);
            mapE.get(ID_MAX_WOCHEN_STUNDEN).get(kuStunde);

            if (mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).has(sKursname) == false) {
                throw new UserException("Im Jahrgang '" + sJahrgang + "' existiert nicht der Kurs '" + sKursname + "', obwohl er in de Datei Kursplaene.txt referenziert wird!");
            }

            mapE.get(ID_STUFEN).get(sJahrgang).get(ID_KURSE).get(sKursname).get(ID_KURS_PLAN).get(sWoche).get(sTag).get(sStunde).get(sRaum).get(sLehrer);
        }

        mapE.get(ID_MAX_WOCHEN_TAGE).retain_key_max();
        mapE.get(ID_MAX_WOCHEN_STUNDEN).retain_key_max();
        gui.debug("  MAX_WOCHEN_TAG = " + mapE.get(ID_MAX_WOCHEN_TAGE).first_key_int());
        gui.debug("  MAX_WOCHEN_STUNDEN = " + mapE.get(ID_MAX_WOCHEN_STUNDEN).first_key_int());

        misc_revalidate_data();
    }

    private String[][] read_kurs42_file_as_table(File file) throws FileNotFoundException, IOException {
        return read_kurs42_file_as_table(file, StandardCharsets.UTF_8);
    }

    private String[][] read_kurs42_file_as_table(File file, Charset kurs42_encoding) throws FileNotFoundException, IOException {
        String[][] tabelle;

        try (BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(file), kurs42_encoding))) {
            // erste zeile analysieren
            String first_line = br.readLine();

            int columns1 = 1;
            for (int i = 0; i < first_line.length(); i++) {
                if (first_line.charAt(i) == ';') {
                    columns1++;
                }
            }
            int columns2 = 1;
            for (int i = 0; i < first_line.length(); i++) {
                if (first_line.charAt(i) == ',') {
                    columns2++;
                }
            }
            int columns = columns1;
            String kurs42_header_seperator = ";";
            if (columns2 > columns1) {
                columns = columns2;
                kurs42_header_seperator = ",";
            }

            // alle zeilen in liste
            LinkedList<String> liste = new LinkedList<>();
            while (br.ready()) {
                String line = br.readLine();
                liste.addLast(line);
            }

            // liste --> tabelle
            tabelle = new String[liste.size()][columns];
            for (int i = 0; i < tabelle.length; i++) {
                String[] zellen = liste.pollFirst().split(kurs42_header_seperator);
                for (int j = 0; j < columns; j++) {
                    tabelle[i][j] = j < zellen.length ? zellen[j] : "";
                    tabelle[i][j] = tabelle[i][j].replaceAll("\"", "").trim();
                }
            }
        }

        return tabelle;
    }

    private boolean misc_ist_klausurschreiber(String sKursart) {
        return sKursart.equals("LK1") || sKursart.equals("LK2") || sKursart.equals("GKS") || sKursart.equals("AB3") || sKursart.equals("AB4");
    }

    private void co_set_wochenplan_ab(String sDatum) {
        mapE.get(ID_WOCHENPLAN_DATUM_AB).clear();
        mapE.get(ID_WOCHENPLAN_DATUM_AB).get(sDatum);
    }

    private void co_set_wochenplan_bi(String sDatum) {
        mapE.get(ID_WOCHENPLAN_DATUM_BI).clear();
        mapE.get(ID_WOCHENPLAN_DATUM_BI).get(sDatum);
    }

    private void co_ausgabe_clear() {
        mapA.clear();
        ausgabe_model.removeAllElements();
    }

    private String co_datum_get_montag_in_woche(String datum) throws ParseException {
        Calendar c = Calendar.getInstance();
        Date date = sdf.parse(datum);
        c.setTime(date);
        while (c.get(Calendar.DAY_OF_WEEK) != Calendar.MONDAY) {
            c.add(Calendar.DATE, -1);
        }
        return sdf.format(c.getTime());
    }

    private String datum_get_sonntag_in_woche(String datum) throws ParseException {
        Calendar c = Calendar.getInstance();
        Date date = sdf.parse(datum);
        c.setTime(date);
        while (c.get(Calendar.DAY_OF_WEEK) != Calendar.SUNDAY) {
            c.add(Calendar.DATE, +1);
        }
        return sdf.format(c.getTime());
    }

    private void co_set_plan(String[][] planS, boolean[][] planB) {
        mapE.get(ID_PLAN_S).clear();
        mapE.get(ID_PLAN_B).clear();
        mapE.get(ID_PLAN_S).get(planS);
        mapE.get(ID_PLAN_B).get(planB);
    }

    private String[][] co_get_plan_s() {
        if (mapE.get(ID_PLAN_S).size() == 0) {
            co_set_plan(new String[0][0], new boolean[0][0]);
        }
        return mapE.get(ID_PLAN_S).first_key_string_2d();
    }

    private void co_datum_update_plan(Date date_ab, Date date_bi) throws ParseException, UserException {
        String datumVon = sdf.format(date_ab);
        String datumBis = sdf.format(date_bi);

        String datumJetzt = co_datum_get_montag_in_woche(datumVon);
        String datumEnde = datum_naechster_tag(datum_get_sonntag_in_woche(datumBis));

        LinkedList<String[]> tempS = new LinkedList<>();
        LinkedList<boolean[]> tempB = new LinkedList<>();
        while (datumJetzt.equals(datumEnde) == false) {
            String[] wocheS = new String[7];
            boolean[] wocheB = new boolean[7];
            for (int i = 0; i < 7; i++) {
                wocheS[i] = datumJetzt;
                wocheB[i] = true;
                if (i >= 5) { // nicht am Wochenende
                    wocheB[i] = false;
                }
                if (datum_compare(datumJetzt, datumVon) < 0) // vorher nicht
                {
                    wocheB[i] = false;
                }
                if (datum_compare(datumJetzt, datumBis) > 0) // nachher nicht
                {
                    wocheB[i] = false;
                }
                datumJetzt = datum_naechster_tag(datumJetzt);
            }
            tempS.addLast(wocheS);
            tempB.addLast(wocheB);
        }

        String[][] planS = new String[tempS.size()][7];
        boolean[][] planB = new boolean[tempB.size()][7];
        for (int i = 0; i < planS.length; i++) {
            planS[i] = tempS.pollFirst();
            planB[i] = tempB.pollFirst();
        }

        co_set_plan(planS, planB);
    }

    private boolean[][] co_plan_b_get() {
        if (mapE.get(ID_PLAN_B).size() == 0) {
            mapE.get(ID_PLAN_B).get(new boolean[0][0]);
        }
        return mapE.get(ID_PLAN_B).first_key_boolean_2d();
    }

    private int co_get_letzte_stunde() {
        if (mapE.get(ID_MAX_WOCHEN_STUNDEN).size() == 0) {
            mapE.get(ID_MAX_WOCHEN_STUNDEN).get(1);
        }
        return mapE.get(ID_MAX_WOCHEN_STUNDEN).first_key_int();
    }

    private int co_get_and_inc_berechnungs_nr() {
        if (mapE.get(ID_BERECHNUNG_NR).size() == 0) {
            mapE.get(ID_BERECHNUNG_NR).get(0);
        }

        int id = mapE.get(ID_BERECHNUNG_NR).first_key_int() + 1;
        mapE.get(ID_BERECHNUNG_NR).clear();
        mapE.get(ID_BERECHNUNG_NR).get(id);

        return id;
    }

    int co_get_max_klausurstunden() {
        return co_get_schueleranzahl() * 3 * 3; // TODO better
    }

}
