package bar_multi_klausurplaner_cnf;

import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.prefs.Preferences;

/**
 * @author Benjamin A. Bartsch
 */
public class Static {

    static final String TITLE = "bar_Multi_Klausurplaner_CNF_2020_05_05b (by Benjamin A. Bartsch) (bar@gymnasium-eickel.herne.de) (Licence: CC0 / MIT)";
    static final String[] WOCHENTAGE_VISUELL = {"Mo", "Di", "Mi", "Do", "Fr"};
    static final String DATE_FORMAT = "dd.MM.yyyy";

    static Font FONT_DEFAULT = new Font("", Font.PLAIN, 12);
    static Font FONT_TEXT_AREA = new Font("", Font.PLAIN, 12);
    static Font FONT_CHECK_BOX = new Font("", Font.PLAIN, 12);
    static Font FONT_DEFAULT_SMALLER = new Font("", Font.PLAIN, 9);
    static Font FONT_TITLED_BORDER = new Font("", Font.PLAIN, 9);

    static final HashMap<Integer, String> mapHilfe = new HashMap<>();

    static final String ID_STUFEN = "STUFE";
    static final String ID_JAHR = "JAHR";
    static final String ID_ABSCHNITT = "ABSCHNITT";
    static final String ID_BLOCKUNG = "BEZEICHNUNG";
    static final String ID_BLOCKGRUPPE = "BLOCKGRUPPE";
    static final String ID_STUNDE = "STUNDE";
    static final String ID_STUNDENDAUER = "STUNDENDAUER";
    static final String ID_FAECHER = "FAECHER";
    static final String ID_KURSE = "KURSE";
    static final String ID_KURS_NR = "KURS_NR";
    static final String ID_KURS_FACH = "KURS_FACH";
    static final String ID_KURS_LEHRER = "KURS_LEHRER";
    static final String ID_KURS_STD = "KURS_STUNDEN";
    static final String ID_KURS_RAUM = "KURS_RAUM";
    static final String ID_KURS_KURSART = "KURS_KURSART";
    static final String ID_KURS_SCHIENEN = "KURS_SCHIENEN";
    static final String ID_KURS_SCHUELER_ALLE = "KURS_SCHUELER_ALLE";
    static final String ID_KURS_SCHUELER_SCHRIFTLICH = "KURS_SCHUELER_SCHRIFTLICH";
    static final String ID_KURS_PLAN = "KURS_PLAN";
    static final String ID_SCHUELER = "SCHUELER";
    static final String ID_SCHUELER_NR = "SCHUELER_NR";
    static final String ID_SCHUELER_NAME = "SCHUELER_NAME";
    static final String ID_SCHUELER_VORNAME = "SCHUELER_VORNAME";
    static final String ID_SCHUELER_GEBDATUM = "SCHUELER_GEBDATUM";
    static final String ID_SCHUELER_GESCHLECHT = "ID_SCHUELER_GESCHLECHT";
    static final String ID_SCHUELER_DBID = "SCHUELER_DBID";
    static final String ID_SCHUELER_FACHWAHLEN = "SCHUELER_FACHWAHLEN";
    static final String ID_SCHUELER_KURSE_ALLE = "SCHUELER_KURSE_ALLE";
    static final String ID_SCHUELER_KURSE_SCHRIFTLICH = "SCHUELER_KURSE_SCHRIFTLICH";
    static final String ID_MAX_WOCHEN_TAGE = "MAX_WOCHEN_TAGE";
    static final String ID_MAX_WOCHEN_STUNDEN = "MAX_WOCHEN_STUNDEN";
    static final String ID_MAX_KLAUSUREN_PRO_WOCHE = "MAX_KLAUSUREN_PRO_WOCHE";
    static final String ID_WOCHENPLAN_DATUM_AB = "WOCHENPLAN_DATUM_AB";
    static final String ID_WOCHENPLAN_DATUM_BI = "WOCHENPLAN_DATUM_BIS";
    static final String ID_REGELN = "REGELN";
    static final String ID_CONTENT = "CONTENT";
    static final String ID_CVS_SCHUELER = "CVS_SCHUELER";
    static final String ID_CVS_KURSE_1 = "CVS_KURSE_1";
    static final String ID_CVS_KURSE_2 = "CVS_KURSE_2";
    static final String ID_XML = "XML";
    static final String ID_PLAN_S = "PLAN_S";
    static final String ID_PLAN_B = "PLAN_B";
    static final String ID_BERECHNUNG_NR = "BERECHNUNG_NR";
    static final String ID_MAX_PARALLEL = "MAX_PARALLEL";
    
    static final String WILDCARD = "*";

    static {
        mapHilfe.put(0, "Bitte wählen Sie eine konkrete Regel aus, um Hilfe zu erhalten!");
        
        mapHilfe.put(100, "Sie legen fest, welcher Kurs einer Stufe welche Anzahl von Klausuren schreibt. Wird die Regel mehrfach verwendet, gilt für einen Kurs die letzte Anweisung.");
        mapHilfe.put(101, "Sie legen fest, welches Fach bzw. welche Kursart einer Stufe welche Anzahl von Klausuren schreibt. Wird die Regel mehrfach verwendet, gilt für einen Kurs die letzte Anweisung.");
        mapHilfe.put(102, "Sie legen fest, dass eine bestimmte Klausur eines Kurses nicht stattfindet. Dies ist z.B. sinnvoll, wenn eine übergeordnete Regel wie #100 in einem Spezialfall außer Kraft gesetzt werden soll.");
        mapHilfe.put(103, "Sie legen fest, dass eine bestimmte Klausur eines Faches bzw. einer Kursart nicht stattfindet. Dies ist z.B. sinnvoll, wenn eine übergeordnete Regel wie #100 in einem Spezialfall außer Kraft gesetzt werden soll.");
        mapHilfe.put(104, "Sie legen einen Mindestabstand (in Tagen) zwischen zwei Klausuren eines bestimmten Kurses einer bestimmten Stufe fest. Sinnvoll ist es, diesen Wert für alle Kurse zu maximieren.");
        mapHilfe.put(105, "Sie legen einen Mindestabstand (in Tagen) zwischen zwei Klausuren eines bestimmten Faches oder einer bestimmten Kursart einer bestimmten Stufe fest. Dies ist sinnvoll, um individuelle Abstandsregelungen für LKs oder einzelne Fächer zu realisieren.");
        mapHilfe.put(106, "Sie legen einen Mindestabstand (in Tagen) zwischen zwei Klausuren einer bestimmten Blockungsschiene einer bestimmten Stufe fest. Dies ist sinnvoll, wenn eine Schule Klausuren schienenweise plant und hier (schienen-)individuelle Abstände realisieren will.");
        
        mapHilfe.put(150, "Sie legen fest, welche Klausurlänge (in Schulstunden) eine bestimmte Klausur eines bestimmten Kurses einer bestimmten Stufe hat.");
        mapHilfe.put(151, "Sie legen fest, welche Klausurlänge (in Schulstunden) eine bestimmte Klausur eines bestimmten Faches bzw. einer bestimmten Kursart einer bestimmten Stufe hat.");
        mapHilfe.put(152, "Sie legen fest, welche Klausurlänge (in Schulstunden) eine bestimmte Klausur einer bestimmten Schiene einer bestimmten Stufe hat. Dies ist sinnvoll, um die Klausurlängen von LK-Schienen und GK-Schienen voneinander getrennt zu definieren.");

        mapHilfe.put(170, "Plant eine mündliche Prüfung für einen bestimmten Kurs einer bestimmten Stufe ein. Dazu muss auch die Klausurnummer und die Anzahl der benötigten Tage angegeben werden. Es wird dann keiner der betroffenen SchülerInnen eine Klausur in diesem Zeitraum haben. Im Klausurplan erscheint nur der erste betroffene Tag.");
        mapHilfe.put(171, "Plant eine mündliche Prüfung für ein bestimmtes Fach/eine bestimmte Kursart einer bestimmten Stufe ein. Dazu muss auch die Klausurnummer und die Anzahl der benötigten Tage angegeben werden. Es wird dann keiner der betroffenen SchülerInnen eine Klausur in diesem Zeitraum haben. Im Klausurplan erscheint nur der erste betroffene Tag.");
        
        mapHilfe.put(200, "Sie sperren einen Zeitraum von mehreren Tagen für Klausuren einer bestimmten Stufe. Dies ist beispielsweise sinnvoll, um lange Wochenenden und Ferien vom Programm berücksichtigen zu lassen.");
        mapHilfe.put(201, "Sie sperren einige Stunden grundsätzlich für Klausuren einer bestimmten Stufe. Dies ist beispielsweise sinnvoll, um bspw. keine Klausuren in den Nachmittag rechnen zu lassen. Diese Regel ignoriert mündliche Prüfungen!");
        mapHilfe.put(202, "Sie sperren einige Stunden eines bestimmten Tages grundsätzlich für Klausuren einer bestimmten Stufe. Dies ist beispielsweise sinnvoll, um Termine der Schulgemeinde zu berücksichtigen, zu denen keine Klausuren stattfinden dürfen. Diese Regel ignoriert mündliche Prüfungen!");
        
        mapHilfe.put(300, "Ein bestimmter Schüler bzw. eine bestimmte Schülerin hat in keiner Woche die angegebene Zahl aufeinander folgender Klausurtage (2 oder 3). Diese Regel ist insbesondere sinnvoll, um mit dem Sternchen alle SchülerInnen einer Stufe zu verarbeiten und eine gewisse Anzahl von aufeinander folgenden Klausurtagen zu verhindern.");
        mapHilfe.put(301, "Sinnvoll als Ergänzung zu Regel 300. Man kann hier die Häufigkeit regulieren, mit der ein bestimmter Schüler bzw. eine bestimmte Schülerin einer bestimmten Jahrgangsstufe mehrere Klausurtage (2 oder 3) am Stück hat.");
        mapHilfe.put(302, "Diese Regel nimmt pro Jahrgangsstufe die Summe über alle SchülerInnen und Wochen und zählt aufeinanderfolgende Klausurtage (2 oder 3) und begrenzt diese.");
        mapHilfe.put(303, "Diese Regel nimmt die Summe über alle Jahrgangsstufen und SchülerInnen und Wochen und zählt aufeinanderfolgende Klausurtage (2 oder 3) und begrenzt diese.");
        mapHilfe.put(304, "Sie sperren den Tag nach einer Klausur eines bestimmten Faches oder einer bestimmten Kursart für einen bestimmten Schüler bzw. eine bestimmte Schülerin einer bestimmten Stufe. Wendet man die Regel für alle SchülerInnen einer bestimmten Stufe an, so kann man beispielsweise den Tag nach einer LK-Klausur frei halten.");
        mapHilfe.put(305, "Sie sperren den Tag vor einer Klausur eines bestimmten Faches oder einer bestimmten Kursart für einen bestimmten Schüler bzw. eine bestimmte Schülerin einer bestimmten Stufe. Wendet man die Regel für alle SchülerInnen einer bestimmten Stufe an, so kann man beispielsweise den Tag vor einer LK-Klausur frei halten.");
        mapHilfe.put(306, "Sie verhindern, dass für einen bestimmten Schüler bzw. eine bestimmte Schülerin einer bestimmten Stufe die angegebenen Klausuren eines Faches oder einer Kursart in der gleichen Woche liegen. Wendet man die Regel für alle SchülerInnen einer bestimmten Stufe an, so kann man grundsätzlich ausschließen, dass Klausuren von zwei bestimmten Fächern bzw. Kursarten in einer Woche liegen.");
        mapHilfe.put(307, "Sie setzen eine Obergrenze für Klausurtage in einer Woche, bezogen auf einen bestimmten Schüler bzw. eine bestimmte Schülerin einer bestimmten Stufe, wenn ein bestimmtes Fach bzw. eine bestimmte Kursart in der Woche eine Klausur hat. Wendet man die Regel für alle SchülerInnen einer bestimmten Stufe an, so kann man beispielsweise ausschließen, dass in der Woche einer LK-Klausur noch zwei weitere Klausuren liegen.");

        mapHilfe.put(400, "Sie definieren einen gemeinsamen Klausurbeginn für alle Klausuren des gleichen Fachs. Ausgenommen sind hier mündliche Prüfungen (die aber separat definiert werden müssen) und LK-Klausuren. Diese Regel ist nötig, wenn eine Schule Klausuren nicht in Stundenplanschienen schreibt, sondern nach Fächern. Wendet man diese Regel für alle Stufen an (Sternchen), so wird jede Stufe für sich gerechnet.");
        mapHilfe.put(401, "Sie definieren einen gemeinsamen Klausurbeginn für alle Klausuren von zwei bestimmten Kursen. Mündliche Klausuren werden ignoriert.");
        mapHilfe.put(402, "Sie definieren einen gemeinsamen Klausurbeginn für alle Klausuren eines bestimmten Kurses mit allen Kursen eines Faches bzw. einer Kursart. Sie könnten damit beispielsweise dafür sorgen, dass die Klausuren eines L8-Kurses gemeinsam geschrieben wird mit alles L6-Kursen. Mündliche Klausuren werden ignoriert.");
        mapHilfe.put(403, "Sie definieren einen gemeinsamen Klausurbeginn für alle Klausuren eines bestimmten Faches bzw. einer bestimmten Kursart. Sie könnten damit beispielsweise dafür sorgen, dass alle Mathematik-Leistungskurse parallel schreiben. Mündliche Klausuren werden ignoriert.");
        mapHilfe.put(404, "Sie definieren einen gemeinsamen Klausurbeginn für alle Klausuren eines bestimmten Faches bzw. einer bestimmten Kursart mit allen Klausuren eines anderen Faches bzw. einer anderen Kursart. Sie könnten damit beispielsweise dafür sorgen, dass alle Sozialwissenschaften-Kurse ihre Klausuren parallel zu den Sozialwissenschaften (Englisch)-Kursen schreiben. Mündliche Klausuren werden ignoriert.");
        mapHilfe.put(405, "Sie legen fest, dass alle Klausuren einer Stufe, die auf einen Tag gerechnet werden, in der gleichen Stunde beginnen. Dies ist beispielsweise von Vorteil, wenn man möglichst wenige Räume und möglichst wenige Aufsichten einsetzen möchte. Mündliche Klausuren werden ignoriert.");
        mapHilfe.put(406, "Sie legen fest, dass alle Kurse einer Stufe und eines Faches bzw. einer Kursart, die durch die gleiche Lehrkraft unterrichtet werden, zur gleichen Zeit beginnen sollen. Dies minimiert die Anzahl von Klausuren, die eine Lehrkraft konzipieren muss und ist eine weniger große Einschränkung im Vergleich zu Regel 403. Mündliche Klausuren werden ignoriert.");
        mapHilfe.put(407, "Sie legen fest, dass alle Kurse einer Stufe und einer (Blockungs-)Schiene des Stundenplans zur gleichen Zeit beginnen sollen. Mündliche Klausuren werden ignoriert.");
        mapHilfe.put(408, "Im Unterschied zur Regel 400-407 sind hier mündliche Prüfungen berücksichtigt. Damit können unterschiedliche Kurse gleichzeitig für mündliche Prüfungen verplant werden.");
        mapHilfe.put(409, "Im Unterschied zur Regel 400-407 sind hier mündliche Prüfungen berücksichtigt. Damit können unterschiedliche Fächer/Kursarten gleichzeitig für mündliche Prüfungen verplant werden.");

        mapHilfe.put(450, "Sie legen den Beginn einer bestimmten Klausur einer Stufe und eine (Blockungs-)Schiene des Stundenplans manuell fest.");
        mapHilfe.put(451, "Sie legen den Beginn einer bestimmten Klausur eines bestimmten Kurses einer Stufe manuell fest.");
        mapHilfe.put(452, "Sie legen den Beginn einer bestimmten Klausur einer Stufe und eines Faches bzw. einer Kursart manuell fest.");
        mapHilfe.put(453, "Sie legen einen Datumsbereich für den Beginn einer bestimmten Klausur eines bestimmten Kurses einer Stufe fest. Dies ist sinnvoll, um Klausuren zweier Quartale voneinander zu trennen.");
        mapHilfe.put(454, "Sie legen einen Datumsbereich für den Beginn einer bestimmten Klausur eines bestimmten Faches bzw. einer Kursart einer Stufe fest. Dies ist sinnvoll, um Klausuren zweier Quartale voneinander zu trennen.");
        mapHilfe.put(455, "Sie sperren eine bestimmte Stunde im Stundenplan einer Stufe für den Klausurbeginn. Mündliche Klausuren werden ignoriert.");
        mapHilfe.put(456, "Sie sperren eine Stundenbereich eines bestimmten Wochentages (alle Wochen) im Stundenplan einer Stufe für den Klausurbeginn einer Kursart. Mündliche Klausuren werden ignoriert.");
        
        mapHilfe.put(900, "Sie legen die maximale Anzahl von Klausurtagen einer Stufe fest. Dies ein Optimierungsparameter, der beim Erstellen des Klausurplans heruntergerechnet werden muss, um unterschiedliche Klausuren zusammen zu legen.");
        mapHilfe.put(901, "Sie legen die maximale Anzahl von Klausurtagen aller Stufen zusammen fest. Im Vergleich zu Regel 900 führt er dazu, dass Klausuren von mehreren Stufen mit höherer Wahrscheinlichkeit am gleichen Tag landen und dadurch sehr volle Klausurtage (hohe Anzahl an Klausurschreibern führen evtl. zu einem Raum-/Aufsichtsproblem) entstehen können, wenn man nicht mit einer Zellenregel (vgl. Regel 903-905) ein Maximum festlegt.");
        mapHilfe.put(902, "Zur besseren Verteilung können Sie einen Höchstwert für gemeinsame Klausurtage zweier Stufen festlegen. Die Stunden des jeweiligen Tages sind dabei egal.");
        mapHilfe.put(903, "Zur besseren Verteilung können Sie einen Höchstwert für gemeinsame Klausurtage aller Stufen festlegen. Die Stunden des jeweiligen Tages sind dabei egal und es zählt auch eine Klausur als \"gemeinsam\", wenn sie nur zwei (von drei oder mehr) Stufen betrifft.");
        mapHilfe.put(904, "Mit dieser Regel kann die höchste Anzahl an gemeinsamen Klausurstunden zweier Stufen festgelegt werden. Dies ist sinnvoll, um große Anzahlen von Klausurschreibern zu verhindern, die ggf. zu einem Raum- bzw. Aufsichtsproblem führen. Eine gemeinsame Klausurstunde ist eine Stunde an einem bestimmten Tag, in der in mehr als einer Jahrgangsstufe eine Klausur stattfindet.");
        mapHilfe.put(905, "Mit dieser Regel kann die höchste Anzahl an gemeinsamen Klausurstunden aller Stufen festgelegt werden. Dies ist sinnvoll, um große Anzahlen von Klausurschreibern zu verhindern, die ggf. zu einem Raum- bzw. Aufsichtsproblem führen. Eine gemeinsame Klausurstunde ist eine Stunde an einem bestimmten Tag, in der in mehr als einer Jahrgangsstufe eine Klausur stattfindet.");
        mapHilfe.put(906, "Sie legen eine Höchstgrenze an Klausurschreibern über alle Stufen für eine bestimmte Stunde eines bestimmten Datums fest. Achtung, rechenintensive Regel, falls für alle Daten und alle Stunden gerechnet wird!");
        mapHilfe.put(907, "Sie legen eine Höchstgrenze für die Anzahl an Klausuren fest, die eine bestimmte Stunde eines bestimmten Wochentages (Summe über alle Wochen des Klausurplans) betreffen.");
        mapHilfe.put(908, "Ergänzung zu Regel 903: Sie legen eine Höchstgrenze für die Anzahl an Klausuren fest, die eine bestimmte Schulstunde eines bestimmten Wochentages (Summe über alle Wochen des Klausurplans) betreffen.");
        mapHilfe.put(909, "Sie legen eine Höchstgrenze für die Anzahl an Klausuren fest, die eine bestimmte Schulstunde eines bestimmten Tages (Summe über alle Wochen des Klausurplans) betreffen.");
        mapHilfe.put(910, "Sie legen eine Höchstgrenze für Stunden eines bestimmten Kurses fest, in denen SchülerInnen aufgrund von Klausuren fehlen. Verwendet man *, so gilt die Höchstgrenze \"pro Kurs\". ");
        mapHilfe.put(911, "Kursübergreifende Variante von Regel 910. Sie legen die Summe von Unterrichtsstunden einer bestimmten Stufe fest, in denen SchülerInnen aufgrund von Klausuren fehlen. ");
        mapHilfe.put(912, "Stufenübergreifende Variante von Regel 910. Sie legen die Summe von Unterrichtsstunden aller Stufen fest, in denen SchülerInnen aufgrund von Klausuren fehlen. ");
        
        mapHilfe.put(920, "Die Klausuren werden so geplant, dass eine bestimmte Mindestanzahl von Klausurstunden eines Faches bzw. einer Kursart innerhalb der regulären Unterrichtszeit der einzelnen Kurse stattfindet. Achtung: Wenn alle Kurse eines Faches parallel schreiben sollen, darf die Mindestzahl nicht zu hoch gewählt werden.");
        mapHilfe.put(921, "Fächerübergreifende Variante von Regel 920. Wenn bei Stufe * gewählt wird, gilt die gewählte Mindestzahl von Klausurstunden \"pro Stufe\".");
        mapHilfe.put(922, "Fächer- und Stufenübergreifende Variante von Regel 920.");

        mapHilfe.put(-1, "");
        mapHilfe.put(-1, "");
        mapHilfe.put(-1, "");
        mapHilfe.put(-1, "");
        mapHilfe.put(-1, "");
        mapHilfe.put(-1, "");
    }

    static Rectangle get_centered_rectangle(double percent) {
        Dimension dim1 = Toolkit.getDefaultToolkit().getScreenSize();
        Dimension dim2 = new Dimension((int) (dim1.width * percent), (int) (dim1.height * percent));
        int x = (int) ((dim1.getWidth() - dim2.width) / 2);
        int y = (int) ((dim1.getHeight() - dim2.height) / 2);
        return new Rectangle(x, y, dim2.width, dim2.height);
    }

    static String getStandardPfadKurs42() {
        Preferences pref = Preferences.userNodeForPackage(Static.class);
        return pref.get("StandardSpeicherpfadKurs42", "");
    }

    static void setStandardPfadKurs42(String value) {
        Preferences pref = Preferences.userNodeForPackage(Static.class);
        pref.put("StandardSpeicherpfadKurs42", value);
    }

    static String getStandardPfadSpeichern() {
        Preferences pref = Preferences.userNodeForPackage(Static.class);
        return pref.get("StandardSpeicherpfad", "");
    }

    static void setStandardPfadSpeichern(String value) {
        Preferences pref = Preferences.userNodeForPackage(Static.class);
        pref.put("StandardSpeicherpfad", value);
    }

    static String get_path_crypto() {
        Preferences pref = Preferences.userNodeForPackage(Static.class);
        return pref.get("StandardPfadCrypto", "");
    }

    static void set_path_crypto(String pfad) {
        Preferences pref = Preferences.userNodeForPackage(Static.class);
        pref.put("StandardPfadCrypto", pfad);
    }

    static void open_webpage(String urlString) {
        try {
            Desktop.getDesktop().browse(new URL(urlString).toURI());
        } catch (IOException | URISyntaxException ex) {
            ex.printStackTrace();
        }
    }

    static String replace(String text, int stelle, String inhalt) throws UserException {
        int index1 = -1;
        for (int i = 0; i < stelle; i++) {
            index1 = text.indexOf('<', index1 + 1);
        }
        int index2 = text.indexOf('>', index1 + 1);
        int length = text.length();
        if (index1 < 0) {
            throw new UserException("Stelle " + stelle + " in " + text + " nicht auffindbar!");
        }
        return text.substring(0, index1) + "<" + inhalt + ">" + text.substring(index2 + 1, length);
    }

    static LinkedList<String> html_escape(LinkedList<String> a) {
        LinkedList<String> b = new LinkedList<>();
        for (String s : a) {
            s = s.replaceAll("Ä", "&Auml;");
            s = s.replaceAll("ä", "&auml;");
            s = s.replaceAll("Ö", "&Ouml;");
            s = s.replaceAll("ö", "&ouml;");
            s = s.replaceAll("Ü", "&Uuml;");
            s = s.replaceAll("ü", "&uuml;");
            s = s.replaceAll("ß", "&szlig;");
            b.addLast(s);
        }
        return b;
    }

    static String extract(String text, int stelle) {

        int index1 = -1;
        for (int i = 0; i < stelle; i++) {
            index1 = text.indexOf('<', index1 + 1);
        }
        int index2 = text.indexOf('>', index1 + 1);

        String result = text.substring(index1 + 1, index2);
        return result;
    }

    static String htmlEscapeSpitzeKlammern(String line) {
        line = line.replaceAll("<", "&lt;");
        line = line.replaceAll(">", "&gt;");
        return line;
    }

    static String htmlEscapeZeilenumbruchZuBr(String line) {
        line = line.replace("\n", "<br>").replace("\r", "<br>");
        return line;
    }

    static String html_escape_regeln(String line) {
        return htmlEscapeZeilenumbruchZuBr(htmlEscapeSpitzeKlammern(line));
    }

    static void toHtmlTable(String[][][] planS, LinkedList<String> html, String[] header) {
        int ROWS = planS.length;
        int COLS = planS[0].length;
        String styleTable = "style=\"width: 100%; table-layout:fixed; border: 1px solid black; border-collapse: collapse;\"";
        String sStyle1 = "style=\"width:9%; border: 1px solid black;border-collapse: collapse;\"";
        String sStyle2 = "style=\"width:18%; border: 1px solid black;border-collapse: collapse;\" colspan=\"2\"";

        String sTD1 = "<td " + sStyle1 + ">";
        String sTD2 = "<td " + sStyle2 + ">";

        String sTH1 = "<th " + sStyle1 + ">";
        String sTH2 = "<th " + sStyle2 + ">";
        String sTR = "<tr>";

        html.addLast("<table " + styleTable + ">");

        // table header
        html.addLast(sTR);
        String sHeader = "";
        for (int i = 0; i < header.length; i++) {
            sHeader = sHeader + (i == 0 ? sTH1 + header[i] : sTH2 + header[i]);
        }
        html.addLast(sHeader);
        html.addLast("</tr>");

        // rows
        for (int r = 0; r < ROWS; r++) {
            html.addLast(sTR);
            html.addLast(sTD1 + (r + 1) + ". </td>");
            for (int c = 0; c < COLS; c++) {
                if (planS[r][c][0].equals(planS[r][c][1])) {
                    html.addLast(sTD2 + planS[r][c][0] + "</td>");
                } else {
                    if (planS[r][c][0].trim().isEmpty()) {
                        planS[r][c][0] = "-----";
                    }
                    if (planS[r][c][1].trim().isEmpty()) {
                        planS[r][c][1] = "-----";
                    }
                    html.addLast(sTD1 + planS[r][c][0] + "</td>");
                    html.addLast(sTD1 + planS[r][c][1] + "</td>");
                }
            }
            html.addLast("</tr>");
        }

        html.addLast("</table>");
    }

    static String datum_naechster_tag(String datum) throws ParseException {
        Calendar c = Calendar.getInstance();
        SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT);
        Date date = sdf.parse(datum);
        c.setTime(date);
        c.add(Calendar.DATE, 1);
        Date date2 = c.getTime();
        return sdf.format(date2);
    }

    static String datum_vorheriger_tag(String datum) throws ParseException {
        Calendar c = Calendar.getInstance();
        SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT);
        Date date = sdf.parse(datum);
        c.setTime(date);
        c.add(Calendar.DATE, -1);
        Date date2 = c.getTime();
        return sdf.format(date2);
    }

    static int datum_compare(String datumVon, String datumBis) throws ParseException, UserException {
        String d1 = datumVon;
        String d2 = datumVon;

        if (d1.equals(datumBis)) {
            return 0;
        }

        LinkedList<String> debug = new LinkedList<>();
        debug.addLast("CALLED datum_compare(" + datumVon + ", " + datumBis + ")");
        for (int i = 0; i < 1000; i++) {
            d1 = datum_vorheriger_tag(d1);
            d2 = datum_naechster_tag(d2);
            debug.addLast("    " + d1 + " / " + d2);
            if (d1.equals(datumBis)) {
                return +1;
            }
            if (d2.equals(datumBis)) {
                return -1;
            }
        }

        for (String s : debug) {
            System.out.println("" + s);
        }
        throw new UserException("Datumsdifferenz zu groß!");
    }

    public static String html_add_space(int size, String s) {
        int missing = size - s.length();
        for (int i = 0; i < missing; i++) {
            s = "&nbsp;" + s;
        }
        return s;
    }

    static String datum_kalenderwoche(String sDatum) {
        try {
            SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT);
            Date date_ab = sdf.parse(sDatum);
            String sKW_ab = new SimpleDateFormat("w").format(date_ab);
            return sKW_ab;
        } catch (ParseException ex) {
            return "KW?";
        }
    }

    static int datum_wochentag_zu_index(String c) {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }

    static String convertZero(int n) {
        if (n == 0) {
            return "";
        }
        return "" + n;
    }

}
